//-----------------------------------------------------------------------------
// abon.c
// Adaboost evaluator plugin loader
// Jiri Havel, DCGM FIT BUT, Brno
// $Id: abon.cpp 165 2009-09-04 07:24:22Z ihavel $
//-----------------------------------------------------------------------------
#include <abon.h>
#include <autoarray.h>

#include <cassert>
#include <cstdio>
#include <errno.h>

#ifdef __cplusplus
extern "C" {
#endif

abonErrorCode trivialSetVerbosity(abonDetector *detector, unsigned verbosity)
{
    detector->verbosity = verbosity;
    return aboneOk;
}

abonErrorCode trivialSetTimer(abonDetector *detector, Timer *timer)
{
    detector->timer = timer;
    return aboneOk;
}

abonErrorCode trivialSetClassifier(abonDetector *detector, TClassifier *classifier)
{
    detector->classifier = classifier;
    return aboneOk;
}

abonErrorCode trivialSetStep(abonDetector *detector, unsigned hstep, unsigned vstep)
{
    detector->hstep = hstep;
    detector->vstep = vstep;
    return aboneOk;
}

abonErrorCode trivialSetSize(abonDetector *detector, unsigned width, unsigned height, unsigned pitch)
{
    detector->width = width;
    detector->height = height;
    detector->pitch = pitch;
    return aboneOk;
}

abonErrorCode trivialSetScale(abonDetector *detector, float start, float step, unsigned count)
{
    if(detector->verbosity) printf("Scale specification not supported!\n");
    return aboneOk;
}

abonErrorCode trivialSetRotation(abonDetector *detector, float start, float step, unsigned count)
{
    if(detector->verbosity) printf("Rotation specification not supported!\n");
    return aboneOk;
}

abonErrorCode trivialSetArguments(abonDetector *detector, unsigned argc, const char *argv[])
{
    if(detector->verbosity)
    {
        for(unsigned i = 1; i < argc; ++i)
        {
            printf("Argument \"%s\" ignored.\n", argv[i]);
        }
    }
    return aboneOk;
}

abonErrorCode trivialAttach(abonDetector *detector, abonDetector *attachment)
{
    if(detector->verbosity) printf("Detector attaching not supported!\n");
    return aboneOk;
}

// Emulation of abonProcess by abonStart

abonImageData *processByStartGet(void *param)
{
    abonImageData **imagePtr = (abonImageData**)param;
    abonImageData *image = *imagePtr;
    *imagePtr = NULL;
    return image;
}

abonErrorCode processByStartPut(void *param, abonImageData *image)
{
    return aboneOk;
}

abonErrorCode processByStart(abonDetector *detector, abonImageData *image)
{
    return abonStart(detector, processByStartGet, processByStartPut, (void*)&image);
}

// Emulation of abonStart by abonProcess

abonErrorCode startByProcess(abonDetector *detector, abonCbGetImage *get, abonCbPutImage *put, void *param)
{
    abonImageData *image;
    while((image = get(param)) != NULL)
    {
        abonErrorCode error = abonProcess(detector, image);
        if(error != aboneOk)
        {
            return error;
        }
        error = put(param, image);
        if(error != aboneOk)
        {
            return error;
        }
    }
    return aboneOk;
}

abonErrorCode abonValidateDetector(abonDetector *detector)
{
    abonErrorCode retval = aboneOk;
    if(detector == NULL)
    {
        fputs("abonCreate failed!\n", stderr);
        return aboneError;
    }
    if(detector->setVerbosity == NULL)
    {
        detector->setVerbosity = trivialSetVerbosity;
    }
    if(detector->setTimer == NULL)
    {
        detector->setTimer = trivialSetTimer;
    }
    if(detector->setClassifier == NULL)
    {
        detector->setClassifier = trivialSetClassifier;
    }
    if(detector->setStep == NULL)
    {
        detector->setStep = trivialSetStep;
    }
    if(detector->setSize == NULL)
    {
        detector->setSize = trivialSetSize;
    }
    if(detector->setScale == NULL)
    {
        detector->setScale = trivialSetScale;
    }
    if(detector->setRotation == NULL)
    {
        detector->setRotation = trivialSetRotation;
    }
    if(detector->setArguments == NULL)
    {
        detector->setArguments = trivialSetArguments;
    }
    if(detector->attach == NULL)
    {
        detector->attach = trivialAttach;
    }
    if((detector->process == NULL) && (detector->start == NULL))
    {
        fputs("both abonStart and abonProcess unitialized!\n", stderr);
        retval = aboneError;
    }
    else
    {
        if(detector->process == NULL)
        {
            detector->process = processByStart;
        }
        if(detector->start == NULL)
        {
            detector->start = startByProcess;
        }
    }
    if(detector->release == NULL)
    {
        fputs("abonRelease unitialized!\n", stderr);
        retval = aboneError;
    }
    return retval;
}

// Enum translation

static const char *const errorCodeStringTable[] = {
    "No error",
    "Invalid parameter",
    "Out of memory",
    "Invalid size",
    "Generic error"
};

const char *abonTranslateError(abonErrorCode error)
{
    assert(sizeof(errorCodeStringTable) == sizeof(const char*)*aboneNumErrorCodes);
    return error < aboneNumErrorCodes ? errorCodeStringTable[error] : "Unknown error";
}

static const char *const timeStampStringTable[] = {
    "Total time",
    "Pyramid time",
    "Convolve time",
    "Detect time",
    "Aux time 1",
    "Aux time 2"
};

const char *abonTranslateTimeStamp(abonTimeStamps timestamp)
{
    assert(sizeof(timeStampStringTable) == sizeof(const char*)*abontsNumTimeStamps);
    return timestamp < abontsNumTimeStamps ? timeStampStringTable[timestamp] : NULL;
}

// Plugin loading

#ifdef _WIN32

#include <windows.h>

void * libopen(const char * name)
{
    return (void*)LoadLibrary(name);
}

int libclose(void * lib)
{
    return FreeLibrary((HMODULE)lib) == TRUE ? 0 : 1;
}

void * libsymbol(void * lib, const char * name)
{
    return GetProcAddress((HMODULE)lib, name);
}

const char *liberror()
{
    // TODO : implement
    return "Not yet implemented!\n";
}

const char * const libext = "dll";

#else//_WIN32

#include <dlfcn.h>

void * libopen(const char * name)
{
    return dlopen(name, RTLD_NOW | RTLD_LOCAL);
}

int libclose(void * lib)
{
    return dlclose(lib);
}

void * libsymbol(void * lib, const char * name)
{
    return dlsym(lib, name);
}

const char *liberror()
{
    return dlerror();
}

const char * const libext = "so";

#endif//_WIN32

abonDetector *abonLoad(const char *lib)
{
    AUTO_ARRAY(char, libarr, strlen("lib.")+strlen(lib)+strlen(libext)+1);
    char * const libname = AUTO_ARRAY_PTR(libarr);
    sprintf(libname, "lib%s.%s", lib, libext);
    void * library = libopen(libname);
    if(library != NULL)
    {
        abonFunCreate *create = (abonFunCreate*)libsymbol(library, "abonCreate");
        if(create != NULL)
        {
            abonDetector *detector = create();
            if(abonValidateDetector(detector) == aboneOk)
            {
                detector->library = library;
                detector->create = create;
                return detector;
            }
        }
        else
        {
            fprintf(stderr, "'%s' is not ABON module.\n", libname);
        }
        if(libclose(library) != 0)
        {
            fprintf(stderr, "Cannot close '%s'.\n", libname);
        }
    }
    else
    {
        fprintf(stderr, "Cannot load '%s'.(%s)\n", libname, liberror());
    }
    return NULL;
}

void abonUnload(struct abonDetector *detector)
{
    if(detector != NULL)
    {
        void * library = detector->library;
        if(detector->release(detector) != aboneOk)
        {
            fputs("Cannot release detector.\n", stderr);
        }
        if((library != NULL) && (libclose(library) != 0))
        {
            fputs("Cannot close library.\n", stderr);
        }
    }
}

#ifdef __cplusplus
}//extern "C"
#endif
