//-----------------------------------------------------------------------------
// config.cpp
// Detector configuration
// Jiri Havel, DCGM FIT BUT, Brno
// $Id: config.cpp 172 2009-09-30 11:26:08Z ihavel $
//-----------------------------------------------------------------------------
#include "config.h"

// STL
#include <cassert>
#include <iterator>
#include <iostream>
#include <fstream>

// Command line parser
#include <argtable2.h>

// OpenCV
#include <highgui.h>

using namespace std;

namespace std {

ostream & operator<<(ostream & o, const Config::Input & i)
{
    switch(i.type)
    {
        case Config::Input::CAMERA :
            o << "camera " << i.number;
            break;
        case Config::Input::VIDEO :
            o << "video '" << i.name << "'";
            break;
        case Config::Input::IMAGE :
            o << "image '" << i.name << "'";
            break;
        case Config::Input::DATASET :
            o << "dataset";
            break;
        case Config::Input::SEQUENCE :
            o << "sequence " << i.number;
            break;
        default :
            assert(!"Invalid input type!");
    }
    return o;
}

ostream & operator<<(ostream & o, const Config & c)
{
    o   << "engine '" << c.engine << "'\n"
        << (c.disableWB ? "adaboost" : "waldboost") << " classifier '" << c.classifier << "', threshold " << c.threshold << ", " << c.stages << " stages\n"
        << "arguments : ";
    ostream_iterator<string> argumentOutput(o, ",");
    copy(c.arguments.begin(), c.arguments.end(), argumentOutput);
    o   << "\n"
        << "group threshold " << c.groupThreshold << " min " << c.groupMinCount << endl
        << "preprocess smooth " << c.smooth << " base " << c.scaleBase << endl
        << "rotation start " << c.rotationStart << " step " << c.rotationStep << " count " << c.rotationCount << endl
        << "output to '" << c.output << "' fourcc " << fourccToString(c.fourcc) << " "
        << (c.tag ? "tag, " : "") << (c.show ? "show, " : "") << (c.drawAll ? "draw all detections" : "") << endl
        << "db " << c.database << " dataset " << c.dataset << " in " << c.datapath << ", process " << c.process << endl
        << (c.verbose ? "verbose output\n" : "")
        << "video start " << c.start << " frames " << c.frames << " step " << c.step << endl
        << "Input :\n";
    ostream_iterator<Config::Input> outIterator(o, "\n");
    copy(c.inputs.begin(), c.inputs.end(), outIterator);
    return o;
}

}

bool parseCommandLine(Config & config, int argc, char ** argv)
{
    // configuration in file instead of command line
    arg_file *config_file = arg_filen("C", "config", "<file>", 0, argc, "configuration file");
    // engine
    arg_str  *engine  = arg_str0("E", "engine", "<engine>", "detection engine");
    // classifier
    arg_file *classifier = arg_file0("c", "classifier", "<xml>", "classifier xml file");
    arg_dbl  *threshold  = arg_dbl0("t", "threshold", "<double>", "detection threshold");
    arg_int  *stages     = arg_int0(NULL, "stages", "<unsigned>", "maximal evaluated stage count");
    arg_lit  *disableWB  = arg_lit0(NULL, "waldboost-disable", "disable early termination");
    arg_int  *hstep      = arg_int0(NULL, "hstep", "<unsigned>", "detection horizontal step");
    arg_int  *vstep      = arg_int0(NULL, "vstep", "<unsigned>", "detection vertical step");
    arg_str  *arguments  = arg_str0("A", "arguments", "<string>,<string>,...", "comma separated detector arguments");
    // nonmaxima supression
    arg_dbl  *groupThreshold = arg_dbl0("T", "group-threshold", "<double>", "detection overlap threshold");
    arg_int  *groupMinCount  = arg_int0("N", "goup-count", "<unsigned>", "group minimal detection count");
    // preprocessing
    arg_int  *smooth    = arg_int0(NULL, "smooth", "<unsigned>", "smoothing gaussian kernel size");
    arg_dbl  *scaleBase = arg_dbl0("b", "scale-base", "<double>", "starting image scale");
    arg_dbl  *rotationStart = arg_dbl0(NULL, "rot-start", "<double>", NULL);
    arg_dbl  *rotationStep  = arg_dbl0(NULL, "rot-step", "<double>", NULL);
    arg_int  *rotationCount = arg_int0(NULL, "rot-count", "<unsigned>", NULL);
    // output
    arg_str  *output  = arg_str0("o", "output", "<output>", "output file name");
    arg_str  *detections = arg_str0(NULL, "detections", "<detections>", "detections file name");
    arg_str  *timestamps = arg_str0(NULL, "timestamps", "<timestaps>", "timestamps file name");
    arg_str  *fourcc  = arg_str0(NULL, "fourcc", "<FOURCC>", "FOURCC code of output file format");
    arg_lit  *tag     = arg_lit0(NULL, "tag", "tag output images");
    arg_lit  *show    = arg_lit0("s", "show", "show output in highgui window");
    arg_lit  *fullscreen = arg_lit0(NULL, "fullscreen", "show output in fullscreen");
    arg_lit  *drawAll = arg_lit0("d", "draw-all", "draw all detections, not only groups");
    // database
    arg_str  *database    = arg_str0(NULL, "database", NULL, "database connection string");
    arg_str  *datapath    = arg_str0(NULL, "datapath", NULL, "data path prefix");
    arg_str  *dataset     = arg_str0(NULL, "dataset", NULL, "dataset name");
    arg_str  *process     = arg_str0(NULL, "process", NULL, "database output process name");
    arg_str  *description = arg_str0(NULL, "description", NULL, "database output process description");
    // misc
    arg_lit  *verbose = arg_lit0("v", "verbose", "verbose output");
    arg_lit  *help    = arg_lit0(NULL, "help", "print this help and exit");
    // input
    arg_int  *start  = arg_int0(NULL, "start", "<unsigned>", "starting frame");
    arg_int  *frames = arg_int0(NULL, "frames", "<unsigned>", "frame count");
    arg_int  *step   = arg_int0(NULL, "step", "<unsigned>", "frame step");
    arg_file *inputs = arg_filen(NULL, NULL, "{ <file> | CAM[n] }", 0, argc, "input files");
    // end
    struct arg_end *end = arg_end(20);
    //table
    void *argtable[] = {
        config_file,
        engine,
        classifier, threshold, stages, disableWB, hstep, vstep,
        arguments,
        groupThreshold, groupMinCount,
        smooth, scaleBase, rotationStart, rotationStep, rotationCount,
        output, detections, timestamps, fourcc, tag, show, fullscreen, drawAll,
        database, datapath, dataset, process, description,
        verbose, help, 
        start, frames, step, inputs,
        end
    };
    bool retval = true;
    if(arg_nullcheck(argtable) != 0)
    {
        cerr << "Insufficient memory to parse arguments!\n";
        retval = false;
    }
    else
    {
        int errors = arg_parse(argc, argv, argtable);
        //print help
        if(help->count > 0)
        {
            cout << "Usage: abon";
            arg_print_syntax(stdout,argtable,"\n");
            arg_print_glossary_gnu(stdout,argtable);
            retval = false;
        }
        //check for errors
        else if(errors > 0)
        {
            arg_print_errors(stderr, end, "detect");
            cerr << "Try 'detect --help' for more information.\n";
            retval = false;
        }
    }
    if(retval == false)
    {
        arg_freetable(argtable, sizeof(argtable)/sizeof(argtable[0]));
        return false;
    }
    //parse config files
    for(int i = 0; i < config_file->count; ++i)
    {
        if(verbose->count) cout << "Parsing '" << config_file->filename[i] << "'...\n";
        if(!parseCommandFile(config, config_file->filename[i]))
        {
            retval = false;
        }
    }
    //macro for checking unsigned argument value
#define CONFIG_UNSIGNED(name) do {\
    if(name->count) {\
        if(name->ival[0] < 0) {\
            cerr << "Value of " << name->hdr.longopts << " mustn't be negative!\n"; \
            retval = false; }\
        config.name = name->ival[0]; } } while(false)
#define CONFIG_POSINT(name) do {\
    if(name->count) {\
        if(name->ival[0] <= 0) {\
            cerr << "Value of " << name->hdr.longopts << " must be greater than zero!\n"; \
            retval = false; }\
        config.name = name->ival[0]; } } while(false)
#define CONFIG_DOUBLE(name) do {\
    if(name->count) {\
        config.name = name->dval[0]; } } while(false)
    //engine
    if(engine->count) config.engine  = engine->sval[0];
    //classifier
    if(classifier->count) config.classifier = classifier->filename[0];
    CONFIG_DOUBLE(threshold);
    CONFIG_UNSIGNED(stages);
    if(disableWB->count) config.disableWB = true;
    CONFIG_POSINT(hstep);
    CONFIG_POSINT(vstep);
    if(arguments->count)
    {
        splitCommandString(arguments->sval[0], config.arguments);
    }
    //group
    CONFIG_DOUBLE(groupThreshold);
    CONFIG_UNSIGNED(groupMinCount);
    //preprocess
    CONFIG_UNSIGNED(smooth);
    CONFIG_DOUBLE(scaleBase);
    CONFIG_DOUBLE(rotationStart);
    CONFIG_DOUBLE(rotationStep);
    CONFIG_UNSIGNED(rotationCount);
    //output
    if(output->count) config.output = output->sval[0];
    if(detections->count) config.detections = detections->sval[0];
    if(timestamps->count) config.timestamps = timestamps->sval[0];
    if(fourcc->count)
    {
        if(strlen(fourcc->sval[0]) != 4)
        {
            cerr << "FOURCC must have four characters!\n";
            retval = false;
        }
        config.fourcc = CV_FOURCC(fourcc->sval[0][0], fourcc->sval[0][1], fourcc->sval[0][2], fourcc->sval[0][3]);
    }
    if(tag->count) config.tag = true;
    if(show->count) config.show = true;
    if(fullscreen->count) config.fullscreen = true;
    if(drawAll->count) config.drawAll = true;
    //database
    if(database->count) config.database = database->sval[0];
    if(datapath->count) config.datapath = datapath->sval[0];
    if(dataset->count) config.dataset = dataset->sval[0];
    if(process->count) config.process = process->sval[0];
    if(description->count) config.description = description->sval[0];
    //misc
    if(verbose->count) config.verbose = true;
    //input
    CONFIG_UNSIGNED(start);
    CONFIG_UNSIGNED(frames);
    CONFIG_UNSIGNED(step);
#undef CONFIG_UNSIGNED
#undef CONFIG_POSINT
#undef CONFIG_DOUBLE
    //input files
    for(int i = 0; i < inputs->count; ++i)
    {
        //determine input type
        const char *filename = inputs->filename[i];
        const char *extension = inputs->extension[i];
        if(strncmp(filename, "CAM", 3) == 0)
        {
            char * rest;
            long n = strtol(filename + 3, &rest, 10);
            if((*rest == '\0') && (n >= 0))
            {
                config.inputs.push_back(Config::Input(Config::Input::CAMERA, n));
            }
            else
            {
                cerr << "Invalid camera number in '" << filename << "'\n";
                retval = false;
            }
        }
/*
        else if(strstr(filename, "DS") == filename)
        {
            config.inputs.push_back(Config::Input(Config::Input::DATASET));
        }
        else if(strstr(filename, "SEQ") == filename)
        {
            char * rest;
            long n = strtol(filename + 3, &rest, 10);
            if((*rest == '\0') && (n >= 0))
            {
                config.inputs.push_back(Config::Input(Config::Input::SEQUENCE, n));
            }
            else
            {
                cerr << "Invalid sequence number in '" << filename << "'\n";
                //return 0;
                goto error;
            }
        }
*/
        else if((strcmp(extension, ".avi") == 0) || (strcmp(extension, ".mpg") == 0))
        {
            config.inputs.push_back(Config::Input(Config::Input::VIDEO, filename));
        }
        else
        {
            config.inputs.push_back(Config::Input(Config::Input::IMAGE, filename));
        }
    }
    //cleanup
    arg_freetable(argtable, sizeof(argtable)/sizeof(argtable[0]));
    return retval;
}

// Split text file to lines, convert to argv-like array and parse like command line
bool parseCommandFile(Config & config, const char * filename)
{
    //open config file
    ifstream file(filename, ios::in|ios::ate);
    if(!file)
    {
        cerr << "Can't open config file '" << filename << "'!\n";
        return false;
    }
    //read complete contents
    vector<char> contents(file.tellg());
    file.seekg(0, ios::beg);
    file.read(&contents[0], contents.size());
    //replace \n with \0 and collect pointers
    vector<char*> argv(0); // Roman: puvodne bylo argv(1, ""), coz hazelo warning. Nechapu, proc by v tom vektoru mel byt jedem prvek s prazdnym retezcem
    unsigned first = 0;
    for(unsigned i = 0; i < contents.size(); ++i)
    {
        if(contents[i] == '\n')
        {
            contents[i] = '\0';
            //skip comments
            if(contents[first] != '#')
            {
                argv.push_back(&contents[first]);
            }
            first = i+1;
        }
    }
    argv.push_back(NULL);
    //parse as regular command line
    if(parseCommandLine(config, argv.size() - 1, &argv[0]))
    {
        return true;
    }
    else
    {
        cerr << "While parsing command file '" << filename << "'\n";
        return false;
    }
}

void splitCommandString(const char *str, vector<string> & arguments)
{
    const char *begin = str;
    for(const char *end = str; *end != 0; ++end)
    {
        if(*end == ',')
        {
            arguments.push_back(string(begin, end));
            begin = end+1;
        }
    }
    if(*begin != '\0')
    {
        arguments.push_back(string(begin));
    }
}
