//-----------------------------------------------------------------------------
// detect.cpp
// Command line parsing
// Jiri Havel, DCGM FIT BUT, Brno
// $Id: detect.cpp 189 2009-12-07 13:23:54Z ijuranek $
//-----------------------------------------------------------------------------

#include "io.h"

#include "window.h"

// STL
#include <iostream>

// Built in classifier
#include <facedetector.h>

using namespace std;

int main(int argc, char *argv[])
{
    try {
        // Set default configuration
        Config config;
        config.engine = "simple";
        config.threshold = 0.0;
        config.stages = 0;
        config.disableWB = false;
        config.hstep = config.vstep = 1;
        config.groupThreshold = 0.4;
        config.groupMinCount = 2;
        config.doNMS = true;
        config.smooth = 0;
        config.scaleBase = 1.0;
        config.rotationStart = 0.0;
        config.rotationStep = 0.0;
        config.rotationCount = 0;
        config.fourcc = CV_FOURCC('X', 'V', 'I', 'D');
        config.tag = config.show = config.fullscreen = config.drawAll = false;
        config.verbose = false;
        config.start = config.frames = 0;
        config.step = 1;

        // Parse command line arguments
        if(!parseCommandLine(config, argc, argv))
        {
            return EXIT_FAILURE;
        }

        // Optionally show config
        if(config.verbose)
        {
            cout << "Configuration:\n" << config;
        }

        // Create detector
        abonDetector *detector = abonLoad(config.engine.c_str());
        if(detector == NULL)
        {
            return EXIT_FAILURE;
        }
        DetectorPtr detector_unloader(detector);

        // Load classifier
        TClassifier *classifier = &facedetector;
        ClassifierPtr classifier_unloader(NULL);
        if(!config.classifier.empty())
        {
            if((classifier = loadClassifierXML(config.classifier.c_str())) == NULL)
            {
                cerr << "Can't load classifier file '" << config.classifier << "'!\n";
                return EXIT_FAILURE;
            }
            classifier_unloader.set(classifier);
        }

        // Initialize classifier
        if(initClassifier(classifier) == 0)
        {
            cerr << "Can't init classifier!" << endl;
            return EXIT_FAILURE;
        }
        classifier->threshold = static_cast<float>(config.threshold);//final threshold
        // Set evaluated stage count
        if(config.stages > 0)
        {
            if(config.stages > classifier->stageCount)
            {
                cerr << "Requested stage count is greater than classifier length!\n";
                return EXIT_FAILURE;
            }
            classifier->evaluatedStageCount = config.stages;
        }
        // Disable waldboost(early termination)
        if(config.disableWB)
        {
            if(config.verbose)
            {
                cout << "Disabling early termination.\n";
            }
            disableClassifierWaldBoost(classifier);
        }
    /*
        // Connect to database
        if(!config.database.empty())
        {
            if(config.verbose) cout << "Connecting... (" << config.database << ")\n";
            if(
                ((config.db = vplDbInit(NULL, config.database.c_str(), config.datapath.c_str())) == NULL) ||
                (config.db->conn == NULL))
            {
                cerr << "Can't connect to database!\n";
                goto error;
            }
            //list datasets
            if(config.verbose)
            {
                vplDbDatasetsPrint(config.db);
            }
            //open dataset
            if((config.seqs = vplDbSequencesList(config.db, config.dataset.c_str())) == NULL)
            {
                cerr << "Can't open dataset '" << config.dataset << "'!\n";
                goto error;
            }
            //list sequences
            if(config.verbose)
            {
                vplDbSequencesPrint(config.db, config.dataset.c_str());
            }
        }
    */

        // Init timer
        TimerPtr timer;
        if(timer.timer == NULL)
        {
            cerr << "Can't create timer!\n";
            return EXIT_FAILURE;
        }

        // Set detector parameters
        if(abonSetVerbosity(detector, config.verbose ? 1 : 0) != aboneOk)
        {
            cerr << "Can't set verbosity!\n";
            return EXIT_FAILURE;
        }
        if(abonSetTimer(detector, timer.timer) != aboneOk)
        {
            cerr << "Can't set timer!\n";
            return EXIT_FAILURE;
        }
        if(abonSetClassifier(detector, classifier) != aboneOk)
        {
            cerr << "Can't set classifier!\n";
            return EXIT_FAILURE;
        }
        if(abonSetStep(detector, config.hstep, config.vstep) != aboneOk)
        {
            cerr << "Can't set step!\n";
            return EXIT_FAILURE;
        }
        if(config.arguments.size() > 0)
        {
            vector<const char*> args(1, argv[0]);
            for(unsigned i = 0; i < config.arguments.size(); ++i)
            {
                args.push_back(config.arguments[i].c_str());
            }
            args.push_back(NULL);
            if(abonSetArguments(detector, args.size()-1, &args[0]) != aboneOk)
            {
                cerr << "Can't set arguments!\n";
                return EXIT_FAILURE;
            }
        }

        if(config.rotationCount > 0)
        {
            abonDetector *rot = abonLoad("rot");
            if(rot == NULL)
            {
                cerr << "Can't load librot!\n";
                return EXIT_FAILURE;
            }
            if(abonSetVerbosity(rot, config.verbose ? 1 : 0) != aboneOk)
            {
                cerr << "Can't set librot verbosity!\n";
                return EXIT_FAILURE;
            }
            if(abonSetRotation(rot, config.rotationStart, config.rotationStep, config.rotationCount) != aboneOk)
            {
                cerr << "Can't set rotation!\n";
                return EXIT_FAILURE;
            }
            if(abonAttach(rot, detector) != aboneOk)
            {
                cerr << "Can't attach detector!\n";
                return EXIT_FAILURE;
            }
            detector = rot;
            detector_unloader.set(rot);
        }

        Window window;

        if(config.show)
        {
            window.open();
            if(config.fullscreen)
            {
                window.setFullScreen();
            }
        }

        InputOutput io(config, detector, timer.timer, window);

        unsigned frames = 0;

        // Process inputs
        for(unsigned i = 0; i < config.inputs.size(); ++i)
        {
            if(config.verbose) cout << "Processing " << config.inputs[i] << " :\n";
            std::auto_ptr<Input> input(inputCreate(config, i));
            if(input.get() == NULL)
            {
                cerr << "Can't open '" << config.inputs[i] << "'!\n";
                return EXIT_FAILURE;
            }
            std::auto_ptr<Output> output(outputCreate(config, i, input.get()));
            pair<abonErrorCode, unsigned> o = io.runWithIO(input.get(), output.get());
            if(o.first != aboneOk)
            {
                cerr << "Error : " << abonTranslateError(o.first) << endl;
                return EXIT_FAILURE;
            }
            frames += o.second;
        }

        if(config.verbose)
        {
            float t = timer.time();
            cout << "Processed " << frames << " frames in " << t << " seconds == " << frames/t << " fps.\n";
        }

        return EXIT_SUCCESS;
    } catch(exception & ex) {
        cerr << "Error : " << ex.what() << endl;
        return EXIT_FAILURE;
    } catch(...) {
        cerr << "Unknown error\n";
        return EXIT_FAILURE;
    }
}
