#include "input.h"

#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>

using namespace std;

// Image ------------------------------------------------------------

void Image::init()
{
    detectionCount = 0;
    detections = detection_array;
    detectionsEnd = detection_array + DETECTION_COUNT;
    nmsDone = false;
}

void Image::init(IplImage *img)
{
    init();
    //check size and format of image
    if((image != NULL) && (
        (image->width != img->width) ||
        (image->height != img->height) ||
        (image->nChannels != img->nChannels) ||
        (image->depth != img->depth)))
    {
        cvReleaseImage(&image);
    }
    //check size of grayscale image
    if((grayscale != NULL) && (
        (grayscale->width != img->width) ||
        (grayscale->height != img->height)))
    {
        cvReleaseImage(&grayscale);
    }
    //clone or copy image
    if(image == NULL)
    {
        image = cvCloneImage(img);
    }
    else
    {
        cvCopy(img, image);
    }
    //alloc grayscale image
    if(grayscale == NULL)
    {
        grayscale = cvCreateImage(cvSize(img->width, img->height), IPL_DEPTH_8U, 1);
    }
    //set header data
    width = img->width;
    height = img->height;
    pitch = grayscale->widthStep;
    data = grayscale->imageData;
}

Image::Image()
{
    memset(this, 0, sizeof(*this));
}

Image *ImageManager::create()
{
    EnhancedImage *image = NULL;
    // Alloc image
    if(first_free == NULL)
    {//alloc new
        image = new EnhancedImage;
        memset(image, 0, sizeof(*image));
    }
    else
    {//retrieve from list
        image = first_free;
        first_free = image->next_free;
    }
    return image;
}

void ImageManager::release(Image *i)
{
    if(i != NULL)
    {
        EnhancedImage *image = (EnhancedImage*)i;
        image->next_free = first_free;
        first_free = image;
    }
}

ImageManager::ImageManager()
: first_free(NULL)
{
}

ImageManager::~ImageManager() throw()
{
    while(first_free != NULL)
    {
        EnhancedImage *image = first_free;
        first_free = image->next_free;
        delete image;
    }
}

// Input -------------------------------------------------------------------

Image *Input::getImage(ImageManager &manager, unsigned step)
{
    IplImage *frame = this->getFrame(step);
    if(frame == NULL)
        return NULL;
    Image *image = manager.create();
    image->init(frame);
    image->imageID = id++;
    return image;
}

Input::Input()
: id(0)
{
}

Input::~Input() throw()
{
}

// One image input

IplImage * ImageInput::getFrame(unsigned step)
{
    if(step > 0)
    {
        cvReleaseImage(&image);
    }
    return image;
}

ImageInput::ImageInput(IplImage * i)
: image(i)
{
}

ImageInput::~ImageInput() throw()
{
    cvReleaseImage(&image);
}

// Highgui capture input

IplImage * CaptureInput::getFrame(unsigned step)
{
    while(step-- > 0)
    {
        if(cvGrabFrame(capture) == 0)
        {
            return NULL;
        }
    }
    return cvRetrieveFrame(capture);
}

double CaptureInput::getProperty(int id)
{
    return cvGetCaptureProperty(capture, id);
}

CaptureInput::CaptureInput(CvCapture * c)
: capture(c)
{
    cvGrabFrame(capture);
}

CaptureInput::~CaptureInput() throw()
{
    cvReleaseCapture(&capture);
}

// Database sequence input
/*
IplImage * SequenceInput::getFrame(unsigned step)
{
	return (frameNumber + step) >= frames->length ? NULL :
		vplDbFrameGetImage(database, (VplDbFrame *)cvGetSeqElem(frames->frames, frameNumber += step));
}

SequenceInput::SequenceInput(VplDb * db, VplDbSequence * s)
: frameNumber(0), database(db), sequence(s), frames(vplDbSequenceFrames(db, s->id))
{
}

SequenceInput::~SequenceInput() throw()
{
	vplDbFramesListRelease(&frames);
}

// Database dataset input

IplImage *DatasetInput::getFrame(unsigned step)
{
	IplImage * frame = sequence->getFrame(step);
	if(
		(frame == NULL) && //end of current sequence
		((sequenceNumber + 1) < sequences->total)//not at the dataset end
	)
	{
		sequence.reset(new SequenceInput(database, (VplDbSequence *)cvGetSeqElem(sequences, ++sequenceNumber)));
		frame = sequence->getFrame(0);
	}
	return frame;
}

DatasetInput::DatasetInput(VplDb * db, CvSeq * seqs)
: sequenceNumber(0), database(db), sequences(seqs)
, sequence(new SequenceInput(db, (VplDbSequence *)cvGetSeqElem(seqs, 0)))
{
}
*/

// Output ---------------------------------------------------------------------

// Composite output

void CompositeOutput::add(Output *output)
{
    if(output != NULL)
    {
        outputs.push_back(output);
    }
}

bool CompositeOutput::putImage(abonImageData *image)
{
    bool retval = true;
    for(unsigned i = 0; i < outputs.size(); ++i)
    {
        retval = retval && outputs[i]->putImage(image);
    }
    return retval;
}

CompositeOutput::~CompositeOutput() throw()
{
    for(unsigned i = 0; i < outputs.size(); ++i)
    {
        delete outputs[i];
    }
}

// Detections output

bool DetectionsTextOutput::putImage(abonImageData *image)
{
    out << image->imageID << "; " << (image->nmsDone ? "nms; " : "; ");
    for(unsigned i = 0; i < image->detectionCount; ++i)
    {
        abonDetection *d = image->detections + i;
        out << d->x << "; " << d->y << "; " << d->width << "; " << d->height << "; "
            << d->angle << "; " << d->response << "; ";
    }
    out << endl;
    return out;
}

DetectionsTextOutput::DetectionsTextOutput(const char *filename)
: out(filename)
{
}

DetectionsTextOutput::~DetectionsTextOutput() throw()
{
}

// Timestamps output

bool TimestampsTextOutput::putImage(abonImageData *image)
{
    out << image->imageID << "; ";
    for(unsigned i = 0; i < abontsNumTimeStamps; ++i)
    {
        abonTimeStamp *s = image->timestamps + i;
        out << setprecision(12) << 1000 * s->start << "; " << 1000 * s->end << "; ";
    }
    out << image->detectionCount << "; ";
    out << endl;
    return out;
}

TimestampsTextOutput::TimestampsTextOutput(const char *filename)
: out(filename)
{
    out << "# ";
    for(unsigned i = 0; i < abontsNumTimeStamps; ++i)
    {
        out << abonTranslateTimeStamp(static_cast<abonTimeStamps>(i)) << "; ";
    }
    out << endl;
}

TimestampsTextOutput::~TimestampsTextOutput() throw()
{
}


// Image file output

bool ImageOutput::putImage(abonImageData * image)
{
    return cvSaveImage(name.c_str(), ((Image*)image)->image);
}

// Video file output

bool VideoOutput::putImage(abonImageData * image)
{
    return cvWriteFrame(writer, ((Image*)image)->image);
}

VideoOutput::~VideoOutput() throw()
{
    cvReleaseVideoWriter(&writer);
}

// Database process output
/*
bool ProcessOutput::putDetections(const TDetectionList & detections)
{
	VplDbEvent event = vplDbEventInit(input->getSequenceID(), process, input->getFrameNumber(), input->getFrameNumber(), 1, 0.0);
	for(unsigned i = 0; i < detections.size(); ++i)
	{
		event.conf = detections[i].response;
		CvBox2D face;
		face.angle = detections[i].angle;
		face.size.width = detections[i].area.width;
		face.size.height = detections[i].area.height;
		face.center.x = detections[i].area.x + face.size.width/2;
		face.center.y = detections[i].area.y + face.size.height/2;
		vplDbEventBox2dAdd(database, &event, &face);
	}
	return true;
}
*/

// Factory functions -----------------------------------------

Input *inputCreate(const Config &config, unsigned i)
{
    const Config::Input & input = config.inputs[i];
    switch(input.type)
    {
        case Config::Input::IMAGE :
        {
            IplImage * image = cvLoadImage(input.name.c_str());
            try {
                return (image == NULL) ? NULL : new ImageInput(image);
            } catch(...) {
                cvReleaseImage(&image);
                return NULL;
            }
        }
        case Config::Input::VIDEO :
        case Config::Input::CAMERA :
        {
            CvCapture * capture = (input.type == Config::Input::CAMERA)
                ? cvCaptureFromCAM(input.number) : cvCaptureFromFile(input.name.c_str());
            if(config.verbose)
            {
                cout << cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_WIDTH) << 'x' <<
                cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_HEIGHT) << ' ' <<
                cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_COUNT) << " frames " <<
                cvGetCaptureProperty(capture, CV_CAP_PROP_FPS) << " fps " <<
                fourccToString((int)cvGetCaptureProperty(capture, CV_CAP_PROP_FOURCC)) << endl;
            }
            try {
                return (capture == NULL) ? NULL : new CaptureInput(capture);
            } catch(...) {
                cvReleaseCapture(&capture);
                return NULL;
            }
        }
/*      
        case Config::Input::DATASET :
        {
            return new DatasetInput(c.db, c.seqs);
        }
        case Config::Input::SEQUENCE :
        {
            return new SequenceInput(c.db, (VplDbSequence *)cvGetSeqElem(c.seqs, input.number));
        }
*/
        default :
            assert(!"Invalid input type!");
    }
    return NULL;
}

Output *outputCreate(const Config &config, unsigned i, Input * input)
{
    unsigned output_count = 0;
    Output *output = NULL;
    Output *detections_output = NULL;
    if(!config.detections.empty())
    {
        output = detections_output = new DetectionsTextOutput(config.detections.c_str());
        ++output_count;
    }
    Output *timestamps_output = NULL;
    if(!config.timestamps.empty())
    {
        output = timestamps_output = new TimestampsTextOutput(config.timestamps.c_str());
        ++output_count;
    }
    Output *file_output = NULL;
    if(!config.output.empty())
    {
        ostringstream filename;
        filename << config.output;
        if(config.inputs.size() > 1)
        {
            filename << i;
        }
        switch(config.inputs[i].type)
        {
        case Config::Input::IMAGE :
            filename << ".jpg";
            output = file_output = new ImageOutput(filename.str());
            ++output_count;
            break;
        case Config::Input::VIDEO :
        case Config::Input::CAMERA :
                {
                CaptureInput * ci = (CaptureInput *)input;
                filename << ".avi";
                CvVideoWriter *writer = cvCreateVideoWriter(
                    filename.str().c_str(), config.fourcc,
                    ci->getProperty(CV_CAP_PROP_FPS)/config.step,
                    cvSize(ci->getProperty(CV_CAP_PROP_FRAME_WIDTH), ci->getProperty(CV_CAP_PROP_FRAME_HEIGHT))
                );
                output = file_output = new VideoOutput(writer);
                ++output_count;
                break;
            }
        default :
            assert(!"Invalid input type!");
        }
    }
    if(output_count > 1)
    {
        CompositeOutput *out = new CompositeOutput();
        out->add(detections_output);
        out->add(timestamps_output);
        out->add(file_output);
        output = out;
    }
    return output;
}
