#include "io.h"

#include <algorithm>
#include <cmath>

#include <cv.h>
#include <cxcore.h>

using namespace std;

template <typename T>
static inline T sqr(const T x)
{
	return x*x;
}

// Distance of centers of two rectangles
static inline float rectDist(CvRect & a, CvRect & b)
{
    return sqrt(static_cast<float>(sqr(a.x - b.x) + sqr(a.y - b.y)));
}

static inline float rectOverlap(CvRect & U, CvRect & V)
{
	float d = rectDist(U, V);

	int ru = (U.width + U.height);
	int rv = (V.width + V.height);

	float r1 = static_cast<float>(min(ru, rv))*0.5f;
	float r2 = static_cast<float>(max(ru, rv))*0.5f;

	return (d > r1+r2) ? 0 : (r1/r2) * (1.0f - (d / (r1+r2)));
}


static inline CvRect detectionRect(abonDetection * detection)
{
	return cvRect(detection->x, detection->y, detection->width, detection->height);
}

struct Bin
{
	//for detection
	unsigned bin;
	float overlap;
	//for bin with same index
	unsigned count;
	unsigned detection;
};

// Clustering and non-maxima supression
static inline unsigned groupDetections(abonDetection *detections, const unsigned count, const float threshold, const unsigned minGroup)
{
	if (count == 0)
		return 0;

	vector<Bin> bins(count);

	for (unsigned i = 0; i < count; ++i)
	{
        Bin bin = {i, 0, 0, i};
        bins[i] = bin;
	}
	
    // Assign a bin to each detection - actual clustering
    for (unsigned i = 0; i < count-1; ++i)
    {
        CvRect U = detectionRect(detections + i);

        for (unsigned j = i+1; j < count; ++j)
        {
            CvRect V = detectionRect(detections + j);

			float overlap = rectOverlap(U, V);

            if (overlap > threshold && overlap > bins[j].overlap)
            {
                bins[j].bin = bins[i].bin;
                bins[j].overlap = overlap;
            }
        }
    }

    // Select detection with max response from each cluster
    for (unsigned i = 0; i < count; ++i)
    {
        unsigned bin = bins[i].bin;
		unsigned det = bins[bin].detection;

		++bins[det].count;

		if (detections[i].response > detections[det].response)
        {
            bins[bin].detection = i;
			//move count to group maximum
			bins[i].count = bins[det].count;
			bins[det].count = 0;
        }
    }

	// Move group maximas to the front
	unsigned ngroup = count;//first nongroup detection
	for (unsigned i = 0; i < count; ++i)
	{
		if (bins[i].count > minGroup)//at i is group
		{
			if (i > ngroup)//some nongroups precede this
			{
				//move group to front
				swap(detections[i], detections[ngroup]);
				++ngroup;//all between ngroup and i are nongroups
			}
		}
		else//at i isn't group
		{
			if (i < ngroup)
			{
				ngroup = i;//set first nongroup
			}
		}
	}

	return ngroup;//first nongroup means group count
}


InputOutput::InputOutput(const Config &cfg, abonDetector *det, Timer *t, Window &w)
: config(cfg), detector(det), timer(t), window(w)
, quit(false)
, input(NULL), output(NULL), frames(0), image(NULL)
{
}

InputOutput::~InputOutput() throw()
{
}

pair<abonErrorCode, unsigned> InputOutput::runWithIO(Input *in, Output *out)
{
    memset(times, 0, sizeof(times));
    input = in;
    output = out;
    frames = 0;
    faces = 0;
    if((image = input->getImage(manager, config.start)) != NULL)
    {
		abonErrorCode error;
        //set detector parameters
        if((error = abonSetSize(detector, image->width, image->height, image->pitch)) != aboneOk)
		{
			return make_pair(error, 0u);
		}
        //set first expected id
        nextId = image->imageID;
        //run detector
        if((error = abonStart(detector, ioGetImage, ioPutImage, (void*)this)) != aboneOk)
		{
			return make_pair(error, 0u);
		}
        //print times
        if(detector->verbosity)
        {
            cout << "Times for " << frames << " frames : \n";
            for(unsigned i = 0; i < abontsNumTimeStamps; ++i)
            {
                cout << '\t' << abonTranslateTimeStamp(static_cast<abonTimeStamps>(i))
                    << " : " << times[i] << "s, " << times[i]/frames << "s/frame\n";
            }
        }
        return make_pair(aboneOk, frames);
    }
    else
    {
        return make_pair(aboneError, 0u);
    }
}

Image *InputOutput::getImage()
{
	Image *retval = NULL;
	if(!quit//q pressed
		&& (image != NULL)//end of input
		&& ((config.frames == 0) || (frames < config.frames))//frame count
		)
	{
		retval = image;
		//read next input image
		image = input->getImage(manager, config.step);
		//start total time
		if(timer != NULL)
		{
			retval->timestamps[abontsTotal].start = timerTime(timer);
		}
		//convert to grayscale
		cvCvtColor(retval->image, retval->grayscale, CV_RGB2GRAY);
        
        if (config.scaleBase != 1.0f)
        {
            // TODO: prescale the image

        }

		//smooth
		if(config.smooth > 0)
		{
			cvSmooth(retval->grayscale, retval->grayscale, CV_GAUSSIAN, 2*config.smooth+1);
/*          // Some fun stuff by Roman Juranek
            // just ignore it, or you will go mad!!! Seriously!
            static IplImage * tmp0 = 0;
            //static IplImage * tmp1 = 0;
            //static IplImage * tmp2 = 0;
            if (!tmp0)
            {
                tmp0 = cvCreateImage(cvGetSize(retval->grayscale), IPL_DEPTH_8U, 1);
                //tmp1 = cvCreateImage(cvGetSize(retval->grayscale), IPL_DEPTH_8U, 1);
                //tmp2 = cvCreateImage(cvGetSize(retval->grayscale), IPL_DEPTH_8U, 1);
            }
            
            //cvSmooth(retval->grayscale, tmp0, CV_GAUSSIAN, 21);
            //cvSmooth(retval->grayscale, tmp1, CV_GAUSSIAN, 11);
            //cvSmooth(retval->grayscale, tmp2, CV_GAUSSIAN, 5);

            unsigned char * image = (unsigned char*)(retval->grayscale->imageData);
            unsigned char * mask0 = (unsigned char*)(tmp0->imageData);
            //unsigned char * mask1 = (unsigned char*)(tmp1->imageData);
            //unsigned char * mask2 = (unsigned char*)(tmp2->imageData);

            int y = 0;
            while (image < (unsigned char*)(retval->grayscale->imageData + retval->grayscale->height * retval->grayscale->widthStep))
            {
                for (int x = 0; x < tmp0->width; ++x)
                {
                    //if (y < retval->grayscale->height / 2)
                    //{
                        //int diff0 = 128 - mask0[x];
                        //int diff1 = 128 - mask1[x];
                        //int diff2 = 128 - mask2[x];
                        //float r = image[x] + 0.3*diff0 + 0.2*diff1 + 0.1*diff2;
                        //r = (r < 0.0f) ? 0.0f : r;
                        //image[x] = r;
                        if (abs(mask0[x] - image[x]) < 10)
                        {
                            mask0[x] = int(0.8f * mask0[x] + 0.2f * image[x]);
                        }
                        else
                        {
                            mask0[x] = image[x];
                        }
                        image[x] = mask0[x];

                    //}
                    //else
                    //{
                    //    mask[x] = image[x];
                    //}
                }

                image += retval->grayscale->widthStep;
                mask0 += tmp0->widthStep;
                //mask1 += tmp1->widthStep;
                //mask2 += tmp2->widthStep;
                ++y;
            }


            cvNamedWindow("x", 1);
            cvShowImage("x", tmp0);
            cvWaitKey(50);

        */
		}
		//read next frame
		++frames;
	}
	return retval;
}

void drawDetections(IplImage * inputFrame, abonDetection *detections, unsigned detection_count,
        const CvScalar & color, bool outline, bool printResponse, const CvFont * font)
{
    for (unsigned i = 0; i != detection_count; ++i)
    {
        abonDetection & d = detections[i];
        CvRect r = detectionRect(&d);
        float angle = d.angle;
        float response = d.response;

        r.x += r.width/2;
        r.y += r.height/2;

        // Prepare rotation matrix
        float _negRot1M[6];
        CvMat negRot1M = cvMat(2, 3, CV_32FC1, _negRot1M); // Origin rotation matrix
        cv2DRotationMatrix( cvPoint2D32f(0,0), -angle, 1.0, &negRot1M);

        // Left and Down vectors
        float _right[3] = {float(r.width), 0, 1}, _right1[3];
        float _down[3] = {0, float(r.height), 1}, _down1[3];

        CvMat right = cvMat(3, 1, CV_32F, _right);
        CvMat right1 = cvMat(2, 1, CV_32F, _right1);
        CvMat down = cvMat(3, 1, CV_32F, _down);
        CvMat down1 = cvMat(2, 1, CV_32F, _down1);

        cvMatMul(&negRot1M, &right, &right1);
        cvMatMul(&negRot1M, &down, &down1);

        CvPoint TL = cvPoint(int(r.x - _right1[0]/2 - _down1[0]/2), int(r.y - _right1[1]/2 - _down1[1]/2));
        CvPoint TR = cvPoint(int(r.x + _right1[0]/2 - _down1[0]/2), int(r.y + _right1[1]/2 - _down1[1]/2));
        CvPoint BL = cvPoint(int(r.x - _right1[0]/2 + _down1[0]/2), int(r.y - _right1[1]/2 + _down1[1]/2));
        CvPoint BR = cvPoint(int(r.x + _right1[0]/2 + _down1[0]/2), int(r.y + _right1[1]/2 + _down1[1]/2));

        // Black outline
        if (outline)
        {
            cvLine(inputFrame, TL, TR, cvScalar(0,0,0), 2);
            cvLine(inputFrame, TR, BR, cvScalar(0,0,0), 2);
            cvLine(inputFrame, BR, BL, cvScalar(0,0,0), 2);
            cvLine(inputFrame, BL, TL, cvScalar(0,0,0), 2);
        }

        // line
        cvLine(inputFrame, TL, TR, color, 1);
        cvLine(inputFrame, TR, BR, color, 1);
        cvLine(inputFrame, BR, BL, color, 1);
        cvLine(inputFrame, BL, TL, color, 1);

        if (font && printResponse)
        {
            char responseStr[256];
            sprintf(responseStr, "%3.2f", response);
            CvPoint textPos = TL;
            textPos.y -= 5;
            cvPutText(inputFrame, responseStr, TL, font, color);
        }

    } // Result loop
}

abonErrorCode InputOutput::putImage(Image *image)
{
    if (config.drawAll && !image->nmsDone) // Draw all detections in case the NMS was not performed yet
    {
        drawDetections(image->image, image->detections, image->detectionCount, cvScalar(255, 255 , 255), false, false, 0);
    }

    if (!image->nmsDone && config.doNMS) // NMS was not performed by the detection engine
    {
        image->detectionCount = groupDetections(image->detections, image->detectionCount, config.groupThreshold, config.groupMinCount);
        image->nmsDone = true;
        //faces += image->detectionCount;
    }
    faces += image->detectionCount;

    if(timer != NULL)
    {
        image->timestamps[abontsTotal].end = timerTime(timer);
    }

    //draw detections
    drawDetections(image->image, image->detections, image->detectionCount, cvScalar(0, 255 , 0), true, false, 0);
    //

    ////////////////////////////////
    // Create bitmap with detections
    /*
    IplImage * img = image->image;
    cvConvertScaleAbs(img, img, 0.1, 128);

    for (abonDetection * d = image->detections; d != image->detections+image->detectionCount; ++d)
    {
        unsigned char * px = (unsigned char *)((img->imageData + (img->widthStep * (d->y+d->height/2))) + (3*(d->x+d->width/2)));
        *px = 255;
        *(px+1) = 255;
        *(px+2) = 255;

    }
    */
    ////////////////////////////////


    if(image->imageID != nextId)
    {
        //not expected image -> insert into queue
        outputQueue.push(image);
    }
    else
    {
        //expected image
        abonErrorCode error;
        //process it
        if((error = outputImage(image)) != aboneOk)
        {
            return error;
        }
        ++nextId;
        //look into queue for following images
        while(!outputQueue.empty()
            && outputQueue.top()->imageID == nextId)
        {
            if((error = outputImage(outputQueue.top())) != aboneOk)
            {
                return error;
            }
            outputQueue.pop();
            ++nextId;
        }
    }
    return aboneOk;
}

abonErrorCode InputOutput::outputImage(Image *image)
{
    //show final image
    if(config.show)
    {
        window.setSize(image->width, image->height);
        window.show(image->image);
        window.update();
        quit = window.isClosed();
    }
    //add time
    for(unsigned i = 0; i < abontsNumTimeStamps; ++i)
    {
        times[i] += (image->timestamps[i].end - image->timestamps[i].start);
    }
    //output
    if(output != NULL)
    {
        output->putImage(image);
    }
    //release image
    manager.release(image);
    return aboneOk;
}

// Wrappers ---------------------------------------------

abonImageData *InputOutput::ioGetImage(void *param)
{
    if(param == NULL)
    {
        return NULL;
    }
    return (abonImageData*)(((InputOutput*)param)->getImage());
}

abonErrorCode InputOutput::ioPutImage(void *param, abonImageData *image)
{
    if((param == NULL) || (image == NULL))
    {
        return aboneParam;
    }
    return ((InputOutput*)param)->putImage((Image*)image);
}
