/*
								+-----------------------------------+
								|                                   |
								|  *** Mk2 convolution manager ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2008   |
								|                                   |
								|           Convolve.cpp            |
								|                                   |
								+-----------------------------------+
*/

#include "./UberLame_src/CallStack.h"
#include <stdio.h>
#include <vector>
#include <algorithm>
#include <functional>
#include "./UberLame_src/OpenGL20.h"
#include "./UberLame_src/OpenGLState.h"
#include "./UberLame_src/Texture.h"
#include "./UberLame_src/RenderBuffer2.h"
#include "./UberLame_src/ShaderSys.h"
#include "./UberLame_src/GPUImgProc.h"
#include "./UberLame_src/StlUtils.h"
#include "./UberLame_src/VertexBufferObject.h"
#include "Load_XML.h"
#include "Convolve.h"

/*
 *								=== TKernelSize ===
 */

/**
 *	int TKernelSize::Dump() const
 *		- prints object information to stdout (debug)
 *		- returns number of printed characters (just so it can be used with std::mem_fun)
 */
int TKernelSize::Dump() const
{
	return printf("convolution: %d x %d\n", n_width, n_height);
	// returns int (for mem_fun_ref)
}

/**
 *	bool TKernelSize::operator ==(const TWeakClassifier &r_t_classify) const
 *		- returns true if this convolution size equals to block
 *		  size of weak LRD classifier r_t_classify, otherwise returns false
 */
bool TKernelSize::operator ==(const TWeakClassifier &r_t_classify) const
{
	return n_width == r_t_classify.n_block_width &&
		n_height == r_t_classify.n_block_height;
}

/*
 *								=== ~TKernelSize ===
 */

/*
 *								=== TConvolutionList ===
 */

/**
 *	bool TConvolutionList::Create(const TWaldBoostClassifier &r_t_classifier)
 *		- creates list of all convolution kernels,
 *		  required by a classifier cascade r_t_classifier
 *		- returns true on success, false on failure
 */
bool TConvolutionList::Create(const TWaldBoostClassifier &r_t_classifier)
{
	convolution_list.clear();
	// erase any contents of the list

	for(int i = 0; i < r_t_classifier.n_weak_classifier_num; ++ i) {
		std::vector<TKernelSize>::iterator p_conv_it =
			std::find(convolution_list.begin(), convolution_list.end(),
			r_t_classifier.p_weak_classifier[i]);
		// try to find convolution in the list

		if(p_conv_it == convolution_list.end()) {
			if(convolution_list.capacity() == convolution_list.size()) {
				convolution_list.reserve((convolution_list.size() * 2) | 1);
				if(convolution_list.capacity() == convolution_list.size())
					return false;
			}
			// make sure there's space in the list

			TKernelSize t_size;
			t_size.n_width = r_t_classifier.p_weak_classifier[i].n_block_width;
			t_size.n_height = r_t_classifier.p_weak_classifier[i].n_block_height;
			convolution_list.push_back(t_size);
			// insert required convolution 
		}
	}
	// create the list

	std::sort(convolution_list.begin(), convolution_list.end(), b_SmallerConvolution);
	// sort the list, smaller kernels first

	return true;
}

/**
 *	void TConvolutionList::Dump() const
 *		- prints object information to stdout (debug)
 */
void TConvolutionList::Dump() const
{
	std::for_each(convolution_list.begin(), convolution_list.end(),
		std::mem_fun_ref(&TKernelSize::Dump));
}

inline bool TConvolutionList::b_SmallerConvolution(const TKernelSize &r_t_a,
	const TKernelSize &r_t_b)
{
	return r_t_a.n_height < r_t_b.n_height ||
		(r_t_a.n_height == r_t_b.n_height && r_t_a.n_width < r_t_b.n_width);
}

/*
 *								=== ~TConvolutionList ===
 */

/*
 *								=== TImage ===
 */

/**
 *	TImage::TImage(Vector2i _v_size)
 *		- default constructor; creates image with size _v_size and places it to origin
 */
TImage::TImage(Vector2i _v_size)
	:v_position(0, 0), v_size(_v_size)
{}

/**
 *	Vector2i TImage::v_Min() const
 *		- returns coordinate of lower-left image corner
 */
Vector2i TImage::v_Min() const
{
	return v_position;
}

/**
 *	Vector2i TImage::v_Max(int n_padding) const
 *		- returns coordinate of upper-right image corner,
 *		  while considering padding n_padding units
 */
Vector2i TImage::v_Max(int n_padding) const
{
	return v_position + v_size + Vector2i(n_padding, n_padding); // padding
}

/**
 *	unsigned int TImage::n_Area(int n_padding) const
 *		- returns image area, along with padding n_padding units
 */
unsigned int TImage::n_Area(int n_padding) const
{
	return (v_size.x + n_padding) * (v_size.y + n_padding);
}

/**
 *	void TImage::Place(const Vector2i &r_v_min)
 *		- sets image position to r_v_min
 */
void TImage::Place(const Vector2i &r_v_min)
{
	v_position = r_v_min;
}

/**
 *	void TImage::Shift_Right(const TImage &r_t_other, int n_padding)
 *		- collision response
 *		- shifts this image right so minimal distance
 *		  to r_t_other is at least equal to n_padding
 */
void TImage::Shift_Right(const TImage &r_t_other, int n_padding)
{
	v_position.x = r_t_other.v_position.x +
		r_t_other.v_size.x + n_padding; // padding
}

/**
 *	bool TImage::b_Collide(const TImage &r_t_other, int n_padding)
 *		- returns true if minimal distance between this
 *		  image and r_t_other is smaller than n_padding,
 *		  otherwise returns false
 */
bool TImage::b_Collide(const TImage &r_t_other, int n_padding)
{
	return !b_Disjunct(v_position.y, v_Max(n_padding).y,
		r_t_other.v_position.y, r_t_other.v_Max(n_padding).y) &&
		!b_Disjunct(v_position.x, v_Max(n_padding).x,
		r_t_other.v_position.x, r_t_other.v_Max(n_padding).x);
}

/**
 *	static bool TImage::b_Disjunct(int n_min1, int n_max1, int n_min2, int n_max2)
 *		- utility function
 *		- returns true if two intervals [n_min1, n_max1] and
 *		  [n_min2, n_max2] are disjunct. otherwise returns false.
 */
bool TImage::b_Disjunct(int n_min1, int n_max1, int n_min2, int n_max2)
{
	return n_max1 <= n_min2 || n_min1 >= n_max2;
}

/*
 *								=== ~TImage ===
 */

/*
 *								=== TImagePage ===
 */

class TImagePage::CSolveCollisions {
protected:
	TImage &m_r_t_rect;
	const int m_n_width, m_n_height;
	const int m_n_padding;
	int &m_r_n_min_max_y;

public:
	CSolveCollisions(TImage &r_t_rect, int n_width, int n_height,
		int n_padding, int &r_n_temp)
		:m_r_t_rect(r_t_rect), m_n_width(n_width), m_n_height(n_height),
		m_n_padding(n_padding), m_r_n_min_max_y(r_n_temp)
	{}

	inline bool operator ()(const TImage *p_other)
	{
		if(m_r_t_rect.b_Collide(*p_other, m_n_padding)) {
			if(p_other->v_Max(m_n_padding).y < m_r_n_min_max_y)
				m_r_n_min_max_y = p_other->v_Max(m_n_padding).y;
			// track minimum of max-y

			m_r_t_rect.Shift_Right(*p_other, m_n_padding);
			// shift right

			if(m_r_t_rect.v_Max(0).x >= m_n_width) {
				m_r_t_rect.Place(Vector2i(0, m_r_n_min_max_y));
				m_r_n_min_max_y = m_n_height;
			}
			// restart position

			return true;
		}
		return false;
	}
};

class TImagePage::CSumArea {
protected:
	int m_n_area;

public:
	inline CSumArea()
		:m_n_area(0)
	{}

	inline void operator ()(const TImage *p_char_info)
	{
		m_n_area += p_char_info->n_Area(0);
	}

	inline operator int() const
	{
		return m_n_area;
	}
};

/**
 *	int TImagePage::n_InsertImage(TImage &r_t_image)
 *		- inserts image r_t_image into the page
 *		- returns 1 on success, 0 on failure (image wouldn't fit,
 *		  there is not enough continouos free space in the page)
 *		  or -1 on failure (not enough memory)
 *		- note this is very plain and cannot yield good efficiency
 *		  unless images are inserted with some strategy (but it is still
 *		  much better and faster than lightmap page packer, for example
 *		  the one in Half-Life 2)
 */
int TImagePage::n_InsertImage(TImage &r_t_image)
{
	r_t_image.Place(Vector2i(0, 0));
	// begin at origin

	bool b_implacable = false;
	int n_temp = n_height;
	while(std::find_if(image_list.begin(), image_list.end(),
	   CSolveCollisions(r_t_image, n_width, n_height,
	   n_padding, n_temp)) != image_list.end()) {
		if(r_t_image.v_Max(0).y >= n_height) {
			b_implacable = true;
			break;
		}
	}
	// solve collisions

	if(b_implacable || n_width < r_t_image.v_Max(0).x || n_height < r_t_image.v_Max(0).y)
		return 0;
	// couldn't fit to this page

	if(image_list.capacity() == image_list.size()) {
		image_list.reserve((image_list.size() * 2) | 1);
		if(image_list.capacity() == image_list.size())
			return -1;
	}
	image_list.push_back(&r_t_image);
	// add to the list

	return 1;
}

/**
 *	int TImagePage::n_Used_Area() const
 *		- returns area, used by images (sum of image areas with no padding,
 *		  not area of some bounding object arround all images)
 */
int TImagePage::n_Used_Area() const
{
	return std::for_each(image_list.begin(), image_list.end(), CSumArea());
}

/*
 *								=== ~TImagePage ===
 */

/*
 *								=== TDownscaleImageSet ===
 */

/**
 *	bool TDownscaleImageSet::Create(int n_width, int n_height, float f_factor = .9f,
 *		int n_min_width = 24, int n_min_height = 24,
 *		int n_initial_page_size = 256, int n_max_optimal_page_size = 512,
 *		int n_max_page_size = 1024, int n_padding = 1,
 *		int n_page_tighten_thresh = 32, int n_page_size_margin = 8)
 *		- creates a new image pyramid
 *		- n_width, n_height is size of the image in pixels,
 *		  pyramid contains images, scaled by factor f_factor (must be smaller
 *		  than 1, not checked) down to minimal size n_min_width per n_min_height
 *		  pixels, whichever dimension is reached first
 *		- n_initial_page_size is initial page size, n_max_optimal_page_size is optimal
 *		  page size (when page is containing small images, do not make it larger than this),
 *		  n_max_page_size is maximal page size (used only for pages, containing images larger
 *		  than n_max_optimal_page_size), n_padding is image padding
 *		- after all images have been inserted, pages with free space above
 *		  n_page_tighten_thresh arround any of it's sides are tightened with
 *		  margin n_page_size_margin (dimensions will be whole multiple of
 *		  n_page_size_margin)
 *		- returns true on success, false on failure
 */
bool TDownscaleImageSet::Create(int n_width, int n_height, float f_factor,
	int n_min_width, int n_min_height, int n_initial_page_size, int n_max_optimal_page_size,
	int n_max_page_size, int n_padding, int n_page_tighten_thresh, int n_page_size_margin)
{
	image_list.clear();
	page_list.clear();
	// clears image and page lists

	for(; n_width >= n_min_width && n_height >= n_min_height;
	   n_width = int(n_width * f_factor), n_height = int(n_height * f_factor)) {
		if(image_list.capacity() == image_list.size()) {
			image_list.reserve((image_list.size() * 2) | 1);
			if(image_list.capacity() == image_list.size())
				return false;
		}
		image_list.push_back(TImage(Vector2i(n_width, n_height)));
	}
	// generates downscaled versions of the image

	std::vector<TImage*> insert_queue;
	insert_queue.reserve(image_list.size());
	if(insert_queue.capacity() < image_list.size())
		return false;
	insert_queue.resize(image_list.size());
	for(int i = 0, j = image_list.size() - 1; j >= 0; ++ i, -- j)
		insert_queue[i] = &image_list[j];
	// create queue with images yet-to-be-inserted (reversed ordering)

	while(!insert_queue.empty()) {
		if(page_list.capacity() == page_list.size()) {
			page_list.reserve((page_list.size() * 2) | 1);
			if(page_list.capacity() == page_list.size())
				return false;
		}
		page_list.resize(page_list.size() + 1);
		TImagePage &r_t_page = page_list.back();
		// get blank, new page

		r_t_page.n_padding = n_padding;
		r_t_page.n_width = n_initial_page_size;
		r_t_page.n_height = n_initial_page_size;
		r_t_page.image_list.clear();
		// initialize page

		int n_next_image_index = insert_queue.size() - 1;
		for(int n_dim = 0; n_next_image_index >= 0;) {
			int n_result;
			if((n_result = r_t_page.n_InsertImage(
			   *insert_queue[n_next_image_index])) < 0)
				return false; // error
			else if(!n_result) {
				// image doesn't fit

				int n_max_size = (r_t_page.image_list.empty())?
					n_max_page_size : n_max_optimal_page_size;
				// in case there are already some images, use optimal
				// size. otherwise allow to go up to maximal size

				if(!(n_dim & 1)) {
					if(r_t_page.n_width * 2 > n_max_size)
						break;
					r_t_page.n_width *= 2;
				} else {
					if(r_t_page.n_height * 2 > n_max_size)
						break;
					r_t_page.n_height *= 2;
				}
				++ n_dim;
				// enlarge the page (alternate enlarge dimensions)

				r_t_page.image_list.clear();
				n_next_image_index = insert_queue.size() - 1;
				// restart inserting images

				continue;
			} else
				-- n_next_image_index; // image was inserted
		}
		// insert all images it could

		if(n_next_image_index == insert_queue.size() - 1)
			return false;
		// no images were inserted; maximal page size is too small
		
		insert_queue.erase(insert_queue.begin() +
			(n_next_image_index + 1), insert_queue.end());
		// remove inserted images from the queue

		int n_tight_width = 0, n_tight_height = 0;
		for(int i = 0, n = r_t_page.image_list.size(); i < n; ++ i) {
			n_tight_width = max(n_tight_width, r_t_page.image_list[i]->v_Max(0).x);
			n_tight_height = max(n_tight_height, r_t_page.image_list[i]->v_Max(0).y);
		}
		// find real required size

		if(r_t_page.n_width > n_tight_width + n_page_tighten_thresh) {
			r_t_page.n_width = n_tight_width + n_page_size_margin;
			r_t_page.n_width -= r_t_page.n_width % n_page_size_margin;
		}
		if(r_t_page.n_height > n_tight_height + n_page_tighten_thresh) {
			r_t_page.n_height = n_tight_height + n_page_size_margin;
			r_t_page.n_height -= r_t_page.n_height % n_page_size_margin;
		}
		// in case dimensions are close to power-of-two, don't do anything, otherwise
		// change size to required size. make it multiple of n_page_size_margin

		for(int i = 0, n = page_list.size(); i < n && !insert_queue.empty(); ++ i) {
			bool b_inserted = false;
			for(int j = insert_queue.size() - 1; j >= 0; -- j) {
				int n_result;
				if((n_result = r_t_page.n_InsertImage(*insert_queue[j])) < 0)
					return false; // error
				else if(!n_result)
					continue; // image doesn't fit
				else {
					insert_queue.erase(insert_queue.begin() + j); // image was inserted
					b_inserted = true;
				}
			}
			// insert all images it could

			if(b_inserted)
				i = -1;
			// restart the loop in case we added something
		}
		// try to insert remaining images (smaller ones) to the other pages
	}

	return true;
}

/**
 *	void TDownscaleImageSet::Dump() const
 *		- prints object information to stdout (debug)
 */
void TDownscaleImageSet::Dump() const
{
	printf("have pyramid of %d images:\n", image_list.size());
	char p_s_line[81] = {0};
	const int n_align = 80;
	for(int i = 0, n = image_list.size(); i < n; ++ i) {
		char p_s_size[16];
		sprintf(p_s_size, "%d x %d", image_list[i].v_size.x, image_list[i].v_size.y);
		if(strlen(p_s_line)) {
			int n_line_end = (i < n - 1)? 1 : 0;
			if(4 + strlen(p_s_line) + 2 + strlen(p_s_size) + n_line_end < n_align) {
				// 4 = indent (tab), 2 = ', ', n_line_end = ',' / '' on the end of the line
				strcat(p_s_line, ", ");
				strcat(p_s_line, p_s_size);
			} else {
				printf("\t%s%s\n", p_s_line, (i == n - 1)? "" : ",");
				strcpy(p_s_line, p_s_size);
			}
		} else
			strcpy(p_s_line, p_s_size);
	}
	if(strlen(p_s_line))
		printf("\t%s\n", p_s_line);
	// aligned list of image resolutions

	printf("\nhave %d pages:\n", page_list.size());
	for(int i = 0, n = page_list.size(); i < n; ++ i) {
		const TImagePage &r_t_page = page_list[i];
		printf("\tpage %d: %d x %d, %d images, efficiency %.2f\n",
			i, r_t_page.n_width, r_t_page.n_height, r_t_page.image_list.size(),
			float(r_t_page.n_Used_Area()) / (r_t_page.n_width * r_t_page.n_height));
	}
	// simple list
}

/*
 *								=== ~TDownscaleImageSet ===
 */

/*
 *								=== CImageDownscaler ===
 */

/**
 *	CImageDownscaler::CImageDownscaler(CGLState *p_state, TDownscaleImageSet &r_image_set)
 *		- default constructor; creates textures and vertex arrays
 *		  for rendering images into them
 *		- p_state is OpenGL state guard, r_image_set is reference
 *		  to downscaled image set description
 *		- note it is recommended to call b_Status() afterwards to check
 *		  wheter constructor succeeded
 */
CImageDownscaler::CImageDownscaler(CGLState *p_state, TDownscaleImageSet &r_image_set)
	:m_p_vertex_vbo(0), m_p_index_vbo(0)
{
	if(!CGLVertexBufferObject::b_Supported())
		return;
	// requires VBO's

	if(r_image_set.image_list.empty())
		return;
	int n_width = r_image_set.image_list.front().v_size.x;
	int n_height = r_image_set.image_list.front().v_size.y;
	// get size of original image

	int n_image_num = r_image_set.image_list.size();
	int n_vertex_num = 4 * n_image_num;
	int n_vertex_list_size = n_vertex_num * 5; // GL_T2F_V3F
	int n_index_num = 6 * n_image_num; // draw quads using GL_TRIANGLES
	// calculate buffer sizes

	unsigned int *p_index;
	if(!(p_index = new unsigned int[n_index_num]))
		return;
	unsigned int *p_index_ptr = p_index;
	for(int i = 0; i < n_image_num; ++ i, p_index_ptr += 6) {
		p_index_ptr[0] = (i * 4) + 0;
		p_index_ptr[1] = (i * 4) + 1;
		p_index_ptr[2] = (i * 4) + 2;
		p_index_ptr[3] = (i * 4) + 0;
		p_index_ptr[4] = (i * 4) + 2;
		p_index_ptr[5] = (i * 4) + 3;
	}
	// create index buffer contents (simple)

	if(!(m_p_index_vbo = new CGLVertexBufferObject(GL_ELEMENT_ARRAY_BUFFER_ARB)) ||
	   !m_p_index_vbo->Alloc(p_state, n_index_num * sizeof(unsigned int), p_index)) {
		delete[] p_index;
		return;
	}
	delete[] p_index;
	// alloc VBO and fill it with indices

	if(!stl_ut::Reserve_N(m_page_list, r_image_set.page_list.size()))
		return;
	// make sure there's enough space

	GLenum n_tex_fmt = (CGLExtensionHandler::b_SupportedExtension(
		"GL_ARB_texture_rg"))? GL_R8 : GL_RGB8;
	// use single-channel textures, if supported

	float *p_vertex;
	if(!(p_vertex = new float[n_vertex_list_size]))
		return;
	float *p_vertex_ptr = p_vertex;
	int n_first = 0;
	for(int i = 0, n = r_image_set.page_list.size(); i < n; ++ i) {
		const TImagePage &r_t_page = r_image_set.page_list[i];

		TPageGeometry t_geom;
		t_geom.n_count = 6 * r_t_page.image_list.size();
		t_geom.n_first = n_first;//((p_vertex_ptr - m_p_vertex) / (4 * 5)) * 6;
		n_first += t_geom.n_count;
		_ASSERTE(i + 1 == n || t_geom.n_first + t_geom.n_count < n_index_num);
		_ASSERTE(i + 1 < n || t_geom.n_first + t_geom.n_count == n_index_num);
		if(!(t_geom.p_fbo = CGLImageProc::p_GetRenderBuffer(r_t_page.n_width,
		   r_t_page.n_height, 1, true, 0, false, false, 0, false, false, 0, false))) {
			delete[] p_vertex;
			return;
		}
		if(!(t_geom.p_texture = new CGLTexture_Rect(p_state, r_t_page.n_width,
		   r_t_page.n_height, n_tex_fmt)) || !t_geom.p_texture->b_Status()) {
			if(t_geom.p_texture)
				delete t_geom.p_texture;
			delete[] p_vertex;
			return;
		}
		CGLTextureParams &r_params = t_geom.p_texture->r_Parameters(p_state);
		r_params.Set_Texture_Min_Filter(GL_NEAREST);
		r_params.Set_Texture_Mag_Filter(GL_NEAREST); // set nearest filtering
		m_page_list.push_back(t_geom);
		// create structure for drawing page geometry

		for(int j = 0, m = r_t_page.image_list.size(); j < m; ++ j) {
			const TImage &r_t_image = *r_t_page.image_list[j];
			float x = float(r_t_image.v_position.x) / r_t_page.n_width;
			float y = float(r_t_image.v_position.y) / r_t_page.n_height;
			float w = float(r_t_image.v_size.x) / r_t_page.n_width;
			float h = float(r_t_image.v_size.y) / r_t_page.n_height;
			// get relative downscaled image position and size

			for(int k = 0; k < 4; ++ k, p_vertex_ptr += 5) {
				int u = "0011"[k] - '0';
				int v = "0110"[k] - '0';
				// s, t coordinates but in [0, 1] range

				p_vertex_ptr[0] = float(u * n_width);
				p_vertex_ptr[1] = float(v * n_height); // texture rect texcoords are not normalized
				p_vertex_ptr[2] = (x + u * w) * 2 - 1;
				p_vertex_ptr[3] = (y + v * h) * 2 - 1; // normalized position
				p_vertex_ptr[4] = 0; // z = 0
			}
			// generate quad vertices
		}
		// generate geometry required to draw quads
	}
	// generate vertex coords

	if(!(m_p_vertex_vbo = new CGLVertexBufferObject(GL_ARRAY_BUFFER_ARB)) ||
	   !m_p_vertex_vbo->Alloc(p_state, n_vertex_list_size * sizeof(float), p_vertex)) {
		delete[] p_vertex;
		return;
	}
	delete[] p_vertex;
	// alloc VBO and fill it with vertices
}

/**
 *	CImageDownscaler::~CImageDownscaler()
 *		- destructor
 */
CImageDownscaler::~CImageDownscaler()
{
	if(m_p_vertex_vbo)
		delete m_p_vertex_vbo;
	if(m_p_index_vbo)
		delete m_p_index_vbo;
	std::for_each(m_page_list.begin(), m_page_list.end(), DeletePageTex);
}

/**
 *	bool CImageDownscaler::b_Status() const
 *		- returns true if constructor succeeded, otherwise returns false
 */
bool CImageDownscaler::b_Status() const
{
	return m_p_vertex_vbo != 0;
}

/**
 *	bool CImageDownscaler::BuildDownscaledImages(CGLState *p_state,
 *		CGLTexture_Rect *p_src_image)
 *		- renders downscaled images
 *		- p_state is OpenGL state guard, p_src_image is source image
 *		- returns true on success, false on failure
 *		- note p_src_image should have minification filter GL_LINEAR_MIPMAP_LINEAR
 *		  for the best image quality
 */
bool CImageDownscaler::BuildDownscaledImages(CGLState *p_state, CGLTexture_Rect *p_src_image)
{
	CGLImageProc::IdentityMatrices();
	CGLImageProc::DisableProgrammableShading(p_state);
	CGLImageProc::DisableFixedVertexProcessing(p_state);
	CGLImageProc::DisableFixedFragmentProcessing(p_state);
	// prepare for simple image downscaling (just image copying)

	p_state->ClearColor4f(0, 0, 0, 0);
	// clear black

	p_src_image->Bind_Enable(p_state);
	// bind source image

	//glColor3f(1, 1, 1);
	// not required here

	m_p_vertex_vbo->Bind(p_state);
	m_p_index_vbo->Bind(p_state);
	// bind VBO's

	glInterleavedArrays(GL_T2F_V3F, 0, m_p_vertex_vbo->p_OffsetPointer(0));
	// GL3 deprecated

	for(int i = 0, n = m_page_list.size(); i < n; ++ i) {
		TPageGeometry &r_t_page = m_page_list[i];
		if(!r_t_page.p_fbo->b_Status() || !r_t_page.p_fbo->Bind() ||
		   !r_t_page.p_fbo->BindColorTexture_2D(*r_t_page.p_texture,
		   GL_TEXTURE_RECTANGLE_ARB))
			return false;
		// bind render buffer, bind texture for rendering

		glViewport(0, 0, r_t_page.p_fbo->n_Width(), r_t_page.p_fbo->n_Height());
		glClear(GL_COLOR_BUFFER_BIT); // debugging - fill by color

		glDrawElements(GL_TRIANGLES, r_t_page.n_count, GL_UNSIGNED_INT,
			m_p_index_vbo->p_OffsetPointer(r_t_page.n_first * sizeof(unsigned int)));
		// draw images

		if(!r_t_page.p_fbo->BindColorTexture_2D(0, GL_TEXTURE_RECTANGLE_ARB) ||
		   !r_t_page.p_fbo->Release())
			return false;
		// release render buffer
	}

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	// GL3 deprecated

	p_state->BindVertexArrayBuffer(0);
	p_state->BindVertexElementArrayBuffer(0);
	// release VBO's

	return true;//glGetError() == GL_NO_ERROR;
}

/**
 *	int CImageDownscaler::n_Page_Num() const
 *		- returns number of pages (equals to number of pages in image set,
 *		  passed to constructor)
 */
int CImageDownscaler::n_Page_Num() const
{
	return m_page_list.size();
}

/**
 *	CGLTexture_Rect *CImageDownscaler::p_Page_Tex(int n_index)
 *		- returns texture, containing image page with zero-based index n_index
 */
CGLTexture_Rect *CImageDownscaler::p_Page_Tex(int n_index)
{
	return m_page_list[n_index].p_texture;
}

inline void CImageDownscaler::DeletePageTex(TPageGeometry &r_t_page)
{
	delete r_t_page.p_texture;
}

/*
 *								=== ~CImageDownscaler ===
 */

/*
 *								=== CImageConvolver ===
 */

class CImageConvolver::CWorkLoadStrategy {
protected:
	int m_n_pass_num, m_n_head_size, m_n_tail_size;

public:
	CWorkLoadStrategy(int n_op_num, int n_max_ops_pass)
		:m_n_pass_num((n_op_num + n_max_ops_pass - 1) / n_max_ops_pass),
		m_n_head_size((m_n_pass_num)? n_op_num / m_n_pass_num : 0), // avoid divide by zero
		m_n_tail_size(n_op_num - (m_n_head_size * (m_n_pass_num - 1)))
	{}

	inline int n_Pass_Num() const
	{
		return m_n_pass_num;
	}

	inline int n_Op_Num(int n_pass) const
	{
		return (n_pass < m_n_pass_num - 1)? m_n_head_size : m_n_tail_size;
	}

	inline int n_FirstOp_Index(int n_pass) const
	{
		return m_n_head_size * n_pass;
	}
};

typedef std::pair<TKernelSize, bool> TReferencedKernel;

static bool b_SmallerKernel(const TReferencedKernel &r_t_a, const TReferencedKernel &r_t_b)
{
	return r_t_a.first.n_width < r_t_b.first.n_width ||
		(r_t_a.first.n_width == r_t_b.first.n_width &&
		r_t_a.first.n_height < r_t_b.first.n_height);
}

static bool b_SmallerKernelArea(const TReferencedKernel &r_t_a, const TReferencedKernel &r_t_b)
{
	return r_t_a.first.n_width * r_t_a.first.n_height <
		r_t_b.first.n_width * r_t_b.first.n_height;
}

class CFindPrefferedBase {
protected:
	TKernelSize m_t_preferred;

public:
	CFindPrefferedBase(const TKernelSize &r_t_kernel)
		:m_t_preferred(r_t_kernel)
	{
		if(m_t_preferred.n_width > m_t_preferred.n_height)
			-- m_t_preferred.n_width;
		else if(m_t_preferred.n_height > m_t_preferred.n_width)
			-- m_t_preferred.n_height;
		else {
			-- m_t_preferred.n_width;
			-- m_t_preferred.n_height;
		}
		// prefer using square kernels
	}

	inline bool operator ()(const TReferencedKernel &r_t_kernel) const
	{
		return r_t_kernel.first.n_width == m_t_preferred.n_width &&
			r_t_kernel.first.n_height == m_t_preferred.n_height;
	}
};

class CFindAnyBase {
protected:
	TKernelSize m_t_kernel;

public:
	CFindAnyBase(const TKernelSize &r_t_kernel)
		:m_t_kernel(r_t_kernel)
	{}

	inline bool operator ()(const TReferencedKernel &r_t_kernel) const
	{
		return (r_t_kernel.first.n_width == m_t_kernel.n_width ||
			r_t_kernel.first.n_width + 1 == m_t_kernel.n_width) &&
			(r_t_kernel.first.n_height == m_t_kernel.n_height ||
			r_t_kernel.first.n_height + 1 == m_t_kernel.n_height);
	}

	inline bool operator ()(const CImageConvolver::TFinalConvolution &r_t_conv) const
	{
		int n_width = r_t_conv.n_kernel_size >> 8;
		int n_height = r_t_conv.n_kernel_size & 0xff;
		return (n_width == m_t_kernel.n_width || n_width + 1 == m_t_kernel.n_width) &&
			(n_height == m_t_kernel.n_height || n_height + 1 == m_t_kernel.n_height);
	}
};

class CFindKernel {
protected:
	TKernelSize m_t_kernel;

public:
	CFindKernel(const TKernelSize &r_t_kernel)
		:m_t_kernel(r_t_kernel)
	{}

	inline bool operator ()(const CImageConvolver::TFinalConvolution &r_t_conv) const
	{
		int n_width = r_t_conv.n_kernel_size >> 8;
		int n_height = r_t_conv.n_kernel_size & 0xff;
		return n_width == m_t_kernel.n_width && n_height == m_t_kernel.n_height;
	}
};

bool b_IsSmallerKernel(const CImageConvolver::TFinalConvolution &r_t_a,
	const CImageConvolver::TFinalConvolution &r_t_b)
{
	return r_t_a.n_kernel_size < r_t_b.n_kernel_size;
}

bool Copy_KernelList(std::vector<TReferencedKernel> &r_dest,
	const TConvolutionList &r_t_convolution_list)
{
	r_dest.reserve(r_t_convolution_list.convolution_list.size());
	if(r_dest.capacity() < r_t_convolution_list.convolution_list.size())
		return false;
	// make sure there's enough space (probably requires bit more than necessary)

	for(int i = 0, n = r_t_convolution_list.convolution_list.size(); i < n; ++ i) {
		TKernelSize t_kernel = r_t_convolution_list.convolution_list[i];
		if(t_kernel.n_width != 1 || t_kernel.n_height != 1)
			r_dest.push_back(TReferencedKernel(t_kernel, false));
	}
	// copy kernels (does not contain 1x1, we don't need to compute that)

	std::sort(r_dest.begin(), r_dest.end(), b_SmallerKernel);
	// sort kernels

	return true;
}

void Remove_BFilterAvailable_Kernels(std::vector<TReferencedKernel> &r_kernel_list)
{
	std::sort(r_kernel_list.begin(), r_kernel_list.end(), b_SmallerKernelArea);
	// sort kernels by area (then go from the end so bigger
	// kernels are avoided if possible)

#if 1
	for(int i = r_kernel_list.size() - 1; i >= 0; -- i) {
		if(r_kernel_list[i].second)
			continue;
		// in case kernel is already referenced by some other, don't do it

		TKernelSize t_kernel = r_kernel_list[i].first;
		// get kernel we need

		std::vector<TReferencedKernel>::iterator p_cur = r_kernel_list.begin() + i;
		// search between beginning and current kernel

		if(t_kernel.n_width <= 2 && t_kernel.n_height <= 2) {
			/*printf("kernel %dx%d can be obtained from %dx%d using bilibear filtering\n",
				t_kernel.n_width, t_kernel.n_height, 1, 1);*/
			// debug

			r_kernel_list.erase(p_cur); // erase this kernel from the list
			continue;
		}
		// kernels 2x1 1x2 and 2x2 can be created from 1x1 (which is not in the list
		// because it's not needed to be really calculated)

		std::vector<TReferencedKernel>::iterator p_source_it =
			std::find_if(r_kernel_list.begin(), p_cur, CFindPrefferedBase(t_kernel));
		if(p_source_it == r_kernel_list.end())
			p_source_it = std::find_if(r_kernel_list.begin(), p_cur, CFindAnyBase(t_kernel));
		// try to find source convolution

		if(p_source_it != p_cur) {
			printf("kernel %dx%d can be obtained from %dx%d using bilibear filtering\n",
				t_kernel.n_width, t_kernel.n_height,
				(*p_source_it).first.n_width, (*p_source_it).first.n_height);
			// debug

			(*p_source_it).second = true; // kernel is now referenced
			r_kernel_list.erase(p_cur); // erase this kernel from the list
		}
	}
	// walk trough kernels list, remove the ones that can be calculated
	// run-time in texturing units using bilinear filtering
#else
	fprintf(stderr, "!!!!!!!!! warning: fake bilinear-filter convolution disabled !!!!!!!!!\n");
	// debug
#endif

	std::sort(r_kernel_list.begin(), r_kernel_list.end(), b_SmallerKernel);
	// sort kernels

}

/**
 *	CImageConvolver::CImageConvolver(CGLState *p_state, int n_width, int n_height,
 *		const TConvolutionList &r_t_conv_list)
 *		- default constructor; creates convolution plan, compiles all shaders and
 *		  allocates all buffers, required to calculate given set of convolutions
 *		- p_state is OpenGL state guard, n_width and n_height is input image size
 *		  and r_t_conv_list is list of required convolutions
 *		- note convolutions of size up to 2x2 are implemented using bilinear
 *		  filtering when sampling textures, for that reason v_TextureOffset()
 *		  supplies sub-pixel texture offset
 *		- note it is recommended to call b_Status() afterwards to check
 *		  wheter constructor succeeded
 */
CImageConvolver::CImageConvolver(CGLState *p_state, int n_width, int n_height,
	const TConvolutionList &r_t_convolution_list)
	:m_b_status(false), m_b_src_image_needs_bilinear(false), m_p_vertex_vbo(0)
{
	if(!CGLVertexBufferObject::b_Supported())
		return;
	// requires VBO's

	m_convolution_list.reserve(r_t_convolution_list.convolution_list.size());
	if(m_convolution_list.capacity() < r_t_convolution_list.convolution_list.size())
		return;
	// make sure there's enough space in final convolution list

	std::vector<TReferencedKernel> required_conv_list;
	if(!Copy_KernelList(required_conv_list, r_t_convolution_list))
		return;
	// create sorted list of required kernels (doesn't contain 1x1)

	Remove_BFilterAvailable_Kernels(required_conv_list);
	// remove kernels available trough bilinear filtering of different ones

	std::vector<int> kernel_width_list;
	kernel_width_list.reserve(required_conv_list.size());
	if(kernel_width_list.capacity() < required_conv_list.size())
		return;
	for(int i = 0, n = required_conv_list.size(); i < n; ++ i) {
		TKernelSize t_size = required_conv_list[i].first;
		if(std::find(kernel_width_list.begin(), kernel_width_list.end(),
		   t_size.n_width) == kernel_width_list.end())
			kernel_width_list.push_back(t_size.n_width);
	}
	// build list of all kernel widths

	// todo - try to write second codepath for height-major convolution
	// (low priority hovewer, kernels do often come in reciprocal pairs)

	int n_max_draw_buffer_num = n_Max_DrawBuffer_Num();
	// maximal number of draw buffers - to divide parallel convolutions to passes

	GLenum n_tex_fmt = (CGLExtensionHandler::b_SupportedExtension(
		"GL_ARB_texture_rg"))? GL_R8 : GL_RGB8;
	// use single-channel textures, if supported

	CWorkLoadStrategy vert_work(kernel_width_list.size(), n_max_draw_buffer_num);
	for(int i = 0, n = vert_work.n_Pass_Num(); i < n; ++ i) {
		int n_pass_size = vert_work.n_Op_Num(i);
		int n_first_kernel = vert_work.n_FirstOp_Index(i);
		// calculate pass size (number of kernels) and index of first width in the pass

		int n_bypass = 0;
		if(kernel_width_list[0] == 1)
			n_bypass = 1;
		// avoid doing width 1 convolutions

		TPass t_pass;
		if(n_pass_size > n_bypass) {
			if(!Create_Pass(p_state, n_width, n_height, n_tex_fmt, 0,
			   n_pass_size - n_bypass, &kernel_width_list[n_first_kernel + n_bypass], pass_Horiz))
				return;
			// create a new pass, sourcing input image (0)

			t_pass = m_pass_list.back(); // use .back(), there might be more first passes!
			// create copy of the last pass (the one we just created)
			// do not use reference / pointer, adding more passes would invalidate it
		}

		for(int j = 0; j < n_pass_size; ++ j) {
			int n_kernel_width = kernel_width_list[j + n_first_kernel];
			// get kernel width

			CGLTexture_Rect *p_texture = (!j && n_bypass)? 0 : t_pass.p_texture[j - n_bypass];
			// get texture

			std::vector<int> kernel_height_list;
			kernel_height_list.reserve(required_conv_list.size());
			if(kernel_height_list.capacity() < required_conv_list.size())
				return;
			for(int k = 0, m = required_conv_list.size(); k < m; ++ k) {
				TKernelSize t_size = required_conv_list[k].first;
				if(t_size.n_width != n_kernel_width)
					continue;
				// select kernels with the right width

				if(t_size.n_height == 1) {
					_ASSERTE(n_kernel_width != 1); // 1x1 shouldn't be in the list
					TFinalConvolution t_fc;
					t_fc.n_kernel_size = n_KernelSize(n_kernel_width, 1);
					t_fc.p_texture = p_texture;
					t_fc.v_tex_offset = Vector2f(0, 0);
					_ASSERTE(m_convolution_list.capacity() > m_convolution_list.size());
					m_convolution_list.push_back(t_fc);
					// in case it's nx1 kernel, where n > 1, add it to final
					// convolution list; 1x1 convolution won't be there because it doesn't get
					// calculated and we don't know source texture in advance (there might
					// even be more of them)

					// skip height 1 kernels, image doesn't need second pass
				} else {
					if(std::find(kernel_height_list.begin(), kernel_height_list.end(),
					   t_size.n_height) == kernel_height_list.end())
						kernel_height_list.push_back(t_size.n_height);
					// add new kernel heights
				}
			}
			// build list of kernel heights

			CWorkLoadStrategy horiz_work(kernel_height_list.size(), n_max_draw_buffer_num);
			for(int k = 0, m = horiz_work.n_Pass_Num(); k < m; ++ k) {
				int n_pass_size = horiz_work.n_Op_Num(k);
				int n_first_kernel = horiz_work.n_FirstOp_Index(k);
				// calculate pass size (number of kernels) and index of first width in the pass

				if(!Create_Pass(p_state, n_width, n_height, n_tex_fmt, p_texture,
				   n_pass_size, &kernel_height_list[n_first_kernel], pass_Vert))
					return;
				// create a new pass, sourcing first pass image

				TPass &r_t_pass = m_pass_list.back();
				// get reference of second pass (won't be adding more passes in this scope)

				for(int l = 0; l < n_pass_size; ++ l) {
					TFinalConvolution t_fc;
					t_fc.n_kernel_size = n_KernelSize(n_kernel_width,
						kernel_height_list[n_first_kernel + l]);
					t_fc.p_texture = r_t_pass.p_texture[l];
					t_fc.v_tex_offset = Vector2f(0, 0);
					_ASSERTE(m_convolution_list.capacity() > m_convolution_list.size());
					m_convolution_list.push_back(t_fc);
				}
				// add final convolutions to the list (for easier searching)
			}
		}
	}
	// create rendering passes

	for(int i = 0, n = r_t_convolution_list.convolution_list.size(); i < n; ++ i) {
		TKernelSize t_kernel = r_t_convolution_list.convolution_list[i];
		if(t_kernel.n_width == 1 && t_kernel.n_height == 1)
			continue;
		// skip 1x1

		if(std::find_if(m_convolution_list.begin(), m_convolution_list.end(),
		   CFindKernel(t_kernel)) != m_convolution_list.end())
			continue;
		// skip convolutions we actually calculate

		TFinalConvolution t_fc;
		t_fc.p_texture = 0; // meaning the source image
		t_fc.n_kernel_size = n_KernelSize(t_kernel.n_width, t_kernel.n_height);
		t_fc.v_tex_offset = Vector2f(0, 0);
		// works for 1x1 convolution

		std::vector<TFinalConvolution>::iterator p_fc_it =
			std::find_if(m_convolution_list.begin(), m_convolution_list.end(),
			CFindAnyBase(t_kernel));
		// find base convolution image

		int n_base_width, n_base_height;
		if(p_fc_it == m_convolution_list.end() || !(*p_fc_it).p_texture) {
			_ASSERTE(t_kernel.n_width <= 2 && t_kernel.n_height <= 2);
			n_base_width = 1;
			n_base_height = 1;
			// this one is created from the source image

			m_b_src_image_needs_bilinear = true;
			// using source image, set bilinear filtering
		} else {
			n_base_width = (*p_fc_it).n_kernel_size >> 8;
			n_base_height = (*p_fc_it).n_kernel_size & 0xff;
			t_fc.p_texture = (*p_fc_it).p_texture;
			// this one is created from some convolution stage

			CGLTextureParams &r_params = t_fc.p_texture->r_Parameters(p_state);
			r_params.Set_Texture_Min_Filter(GL_LINEAR);
			r_params.Set_Texture_Mag_Filter(GL_LINEAR);
			// set bilinear filtering
		}
		// assign texture, determine 

		_ASSERTE(t_kernel.n_width == n_base_width || t_kernel.n_width == 1 + n_base_width);
		_ASSERTE(t_kernel.n_height == n_base_height || t_kernel.n_height == 1 + n_base_height);
		_ASSERTE(t_kernel.n_width != n_base_width || t_kernel.n_height != n_base_height);
		// make sure kernel size is up to twice base kernel size while not equal to it

		t_fc.v_tex_offset.x = (n_base_width == t_kernel.n_width)? 0 : .5f;
		t_fc.v_tex_offset.y = (n_base_height == t_kernel.n_height)? 0 : .5f;
		// simulate convolution by half-pixel offset and bilinear filtering

		_ASSERTE(m_convolution_list.capacity() > m_convolution_list.size());
		m_convolution_list.push_back(t_fc);
	}
	// add derived convolutions to the list (for easier searching)

	for(int i = 0, n = r_t_convolution_list.convolution_list.size(); i < n; ++ i) {
		TKernelSize t_kernel = r_t_convolution_list.convolution_list[i];
		if(t_kernel.n_width != 1 || t_kernel.n_height != 1)
			continue;
		// skip all but 1x1

		TFinalConvolution t_fc;
		t_fc.p_texture = 0; // meaning the source image
		t_fc.n_kernel_size = n_KernelSize(t_kernel.n_width, t_kernel.n_height);
		t_fc.v_tex_offset = Vector2f(0, 0);
		// works for 1x1 convolution

		_ASSERTE(m_convolution_list.capacity() > m_convolution_list.size());
		m_convolution_list.push_back(t_fc);

		break;
		// there is just one 1x1
	}
	// add 1x1 to the list (for easier searching)

	std::sort(m_convolution_list.begin(), m_convolution_list.end(), b_IsSmallerKernel);
	// sort (doesn't actually affect anything but dump)

	float p_vertex[4 * 5];
	float *p_vertex_ptr = p_vertex;
	for(int i = 0; i < 4; ++ i, p_vertex_ptr += 5) {
		int u = "0011"[i] - '0';
		int v = "0110"[i] - '0';
		// s, t coordinates but in [0, 1] range

		p_vertex_ptr[0] = float(u * n_width);
		p_vertex_ptr[1] = float(v * n_height);// texture rect texcoords are not normalized
		p_vertex_ptr[2] = float(u * 2 - 1);
		p_vertex_ptr[3] = float(v * 2 - 1); // normalized position
		p_vertex_ptr[4] = 0;
	}
	if(!(m_p_vertex_vbo = new CGLVertexBufferObject(GL_ARRAY_BUFFER_ARB)) ||
	   !m_p_vertex_vbo->Alloc(p_state, 4 * 5 * sizeof(float), p_vertex))
		return;
	// create fullscreen quad vertex list (can't use CGLImageProc's one,
	// this one has special texcoords)

	m_b_status = true;
	// constructor success
}

/**
 *	CImageConvolver::~CImageConvolver()
 *		- destructor
 */
CImageConvolver::~CImageConvolver()
{
	if(m_p_vertex_vbo)
		delete m_p_vertex_vbo;
	std::for_each(m_pass_list.begin(), m_pass_list.end(), DeletePass);
}

/**
 *	bool CImageConvolver::b_Status() const
 *		- returns true if constructor succeeded, otherwise returns false
 */
bool CImageConvolver::b_Status() const
{
	return m_p_vertex_vbo && m_b_status;
}

/**
 *	bool CImageConvolver::ConvolveImage(CGLState *p_state, CGLTexture_Rect *p_src_image)
 *		- calculates convolutions
 *		- p_state is OpenGL state guard and p_src_image is source image
 *		  (note it's dimensions are ignored and are expected to equal
 *		  to dimensions, passed to constructor)
 *		- returns true on success, false on failure
 */
bool CImageConvolver::ConvolveImage(CGLState *p_state, CGLTexture_Rect *p_src_image)
{
	p_src_image->Bind_Enable(p_state);
	// bind source image

	CGLImageProc::IdentityMatrices();
	CGLImageProc::DisableProgrammableShading(p_state);
	CGLImageProc::DisableFixedVertexProcessing(p_state);
	CGLImageProc::DisableFixedFragmentProcessing(p_state);
	// prepare for simple image downscaling (just image copying)

	m_p_vertex_vbo->Bind(p_state);
	// bind VBO

	glInterleavedArrays(GL_T2F_V3F, 0, m_p_vertex_vbo->p_OffsetPointer(0));
	// GL3 deprecated

	for(int i = 0, n = m_pass_list.size(); i < n; ++ i) {
		TPass &r_t_pass = m_pass_list[i];
		if(!r_t_pass.p_fbo->b_Status() || !r_t_pass.p_fbo->Bind())
			return false;
		for(int j = 0; j < r_t_pass.n_texture_num; ++ j) {
			if(!r_t_pass.p_fbo->BindColorTexture_2D(*r_t_pass.p_texture[j],
			   GL_TEXTURE_RECTANGLE_ARB, j))
				return false;
		}
		// bind render buffer, bind textures for rendering

		glViewport(0, 0, r_t_pass.p_fbo->n_Width(), r_t_pass.p_fbo->n_Height());
		if(!r_t_pass.p_source)
			p_src_image->Bind_Enable(p_state);
		else
			r_t_pass.p_source->Bind_Enable(p_state);
		r_t_pass.p_shader->Bind(p_state);
#if 0
		const unsigned int p_indices[6] = {0, 1, 2, 0, 2, 3};
		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, p_indices); // must use VBO (driver bug)
#else
		glDrawArrays(GL_QUADS, 4, 0); // note GL_QUADS is GL3 deprecated
#endif
		// draw images

		for(int j = 0; j < r_t_pass.n_texture_num; ++ j) {
			if(!r_t_pass.p_fbo->BindColorTexture_2D(0, GL_TEXTURE_RECTANGLE_ARB, j))
				return false;
		}
		if(!r_t_pass.p_fbo->Release())
			return false;
		// release render buffer
	}

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	// GL3 deprecated

	p_state->BindVertexArrayBuffer(0);
	// release VBO

	return true;//glGetError() == GL_NO_ERROR;
}

/**
 *	bool CImageConvolver::b_SourceImageNeedsBilinearFilter() const
 *		- returns true in case source image needs bilinear filter (GL_LINEAR),
 *		  otherwise returns false (meaning GL_NEAREST suffices)
 *		- note this happens when convolutions 1x2, 2x1 or 2x2 are required
 */
bool CImageConvolver::b_SourceImageNeedsBilinearFilter() const
{
	return m_b_src_image_needs_bilinear;
}

/**
 *	CGLTexture_Rect *CImageConvolver::p_Image(int n_index)
 *		- returns texture, containing convolution with zero-based index n_index
 *		- note this returns 0 when convolution is in source image
 *		  (for example convolution with kernel 1x1)
 */
CGLTexture_Rect *CImageConvolver::p_Image(int n_index)
{
	return m_convolution_list[n_index].p_texture;
}

/**
 *	Vector2f CImageConvolver::v_TextureOffset(int n_index) const
 *		- returns sub-texel offset to calculate convolution with zero-based
 *		  index n_index using bilinear filter in texturing unit
 */
Vector2f CImageConvolver::v_TextureOffset(int n_index) const
{
	return m_convolution_list[n_index].v_tex_offset;
}

/**
 *	Vector2i CImageConvolver::v_Kernel(int n_index) const
 *		- returns size of kernel with zero-based index n_index
 */
Vector2i CImageConvolver::v_Kernel(int n_index) const
{
	uint16_t n_size = m_convolution_list[n_index].n_kernel_size;
	return Vector2i(n_size >> 8, n_size & 0xff);
}

/**
 *	int CImageConvolver::n_Kernel_Num() const
 *		- returns number of calculated kernels (equals to number of kernels
 *		  in convolution list, passed to constructor)
 */
int CImageConvolver::n_Kernel_Num() const
{
	return m_convolution_list.size();
}

/**
 *	void CImageConvolver::Dump() const
 *		- prints object information to stdout (debug)
 */
void CImageConvolver::Dump() const
{
	printf("calculate %d convolutions in %d passes:\n",
		m_convolution_list.size(), m_pass_list.size());
	for(int i = 0, n = m_convolution_list.size(); i < n; ++ i) {
		Vector2f v_off = m_convolution_list[i].v_tex_offset;
		if(v_off.x != 0 || v_off.y != 0) {
			// print source convolution and bilinear filter shape

			uint16_t n_size = m_convolution_list[i].n_kernel_size;
			printf("\tconvolution %d: %d x %d, dataflow: ", i, n_size >> 8, n_size & 0xff);
			if(!m_convolution_list[i].p_texture)
				printf("src-image");
			else {
				const CGLTexture_Rect *p_trace_tex = m_convolution_list[i].p_texture;
				for(int j = 0; j < n; ++ j) {
					if(m_convolution_list[j].v_tex_offset.x == 0 &&
					   m_convolution_list[j].v_tex_offset.y == 0 &&
					   m_convolution_list[j].p_texture == p_trace_tex) {
						uint16_t n_size = m_convolution_list[j].n_kernel_size;
						printf("convolution %d (%d x %d)", i, n_size >> 8, n_size & 0xff);
						break;
					}
					_ASSERTE(j + 1 != n); // must find it
				}
				// find convolution source
			}
			printf(" + bilinear filter convolution %d x %d\n", (v_off.x != 0)? 2 : 1, (v_off.y != 0)? 2 : 1);
		} else {
			// print pipeline schema

			int n_pass_num = 0, p_pass[2], p_draw_buffer[2];
			for(const CGLTexture_Rect *p_trace_tex = m_convolution_list[i].p_texture;
			   p_trace_tex; ++ n_pass_num) {
				if(n_pass_num == 2) {
					fprintf(stderr, "error: unable to trace convolution data flow (>2 passes)\n");
					return;
				}
				p_pass[n_pass_num] = -1;
				for(int j = 0, m = m_pass_list.size(); j < m; ++ j) {
					const TPass &r_pass = m_pass_list[j];
					for(int k = 0; k < r_pass.n_texture_num; ++ k) {
						if(r_pass.p_texture[k] == p_trace_tex) {
							p_pass[n_pass_num] = j;
							p_draw_buffer[n_pass_num] = k;
							p_trace_tex = r_pass.p_source; // trace next texture in the chain
							break;
						}
					}
					if(p_pass[n_pass_num] != -1)
						break;
				}
				if(p_pass[n_pass_num] == -1) {
					fprintf(stderr, "error: unable to trace convolution data flow\n");
					return;
				}
				// find pass and source
			}

			uint16_t n_size = m_convolution_list[i].n_kernel_size;
			printf("\tconvolution %d: %d x %d, dataflow: src-image",
				i, n_size >> 8, n_size & 0xff);
			for(int j = n_pass_num - 1; j >= 0; -- j)
				printf(" -> pass %d to draw-buffer %d", p_pass[j], p_draw_buffer[j]);
			printf("\n");
		}
		// print convolution props and data
	}
}

bool CImageConvolver::Create_Pass(CGLState *p_state, int n_width, int n_height,
	GLenum n_tex_format, CGLTexture_Rect *p_src_tex, int n_kernel_num,
	const int *p_kernel_list, int n_type) // do not access p_src_tex, nulls are legal and expected
{
	if(m_pass_list.capacity() == m_pass_list.size()) {
		m_pass_list.reserve((m_pass_list.size() * 2) | 1);
		if(m_pass_list.capacity() == m_pass_list.size())
			return false;
	}
	// make sure there's enough space in the pass list

	const char *p_s_shader_head =
		"#extension ARB_texture_rectangle : enable\n"
#ifdef DISABLE_TEXTURE_RECT_NV_WARNING
		"#define samplerRect sampler2DRect\n"
		"#define textureRect texture2DRect\n" // remove those lines on ati
#endif //DISABLE_TEXTURE_RECT_NV_WARNING
		"uniform samplerRect n_tex;\n"
		"void main()\n"
		"{\n"
		"    float f_accum = textureRect(n_tex, gl_TexCoord[0].xy).x;\n";
	std::string s_shader_source;
	if(!stl_ut::AppendCStr(s_shader_source, p_s_shader_head))
		return false;
	// begin with shader head

	for(int i = 0, n_kernel_size = 1; i < n_kernel_num; ++ i) {
		_ASSERTE(!i || p_kernel_list[i - 1] < p_kernel_list[i]);
		// make sure kernels are in ascending order

		for(int n_required_size = p_kernel_list[i];
		   n_kernel_size < n_required_size; ++ n_kernel_size) {
			int n_dx = (n_type == pass_Horiz)? n_kernel_size : 0;
			int n_dy = (n_type == pass_Horiz)? 0 : n_kernel_size;
			// new sample position

			char p_s_buffer[80];
			sprintf(p_s_buffer, "    f_accum += textureRect(n_tex, "
				"gl_TexCoord[0].xy + vec2(%d.0, %d.0)).x;\n", n_dx, n_dy);
			if(!stl_ut::AppendCStr(s_shader_source, p_s_buffer))
				return false;
			// add line, reading in the new sample
		}
		// add more samples to meet required kernel size

		char p_s_buffer[80];
		sprintf(p_s_buffer, "    gl_FragData[%d] = vec4(f_accum * %f);\n",
			i, 1.0f / n_kernel_size); // todo - see if it's big perf impact when writing to red component only, while using rgb textures / vice versa
		if(!stl_ut::AppendCStr(s_shader_source, p_s_buffer))
			return false;
		// add line, writing output to fragment
	}
	// concat shader body

	const char *p_s_shader_tail = "}\n";
	if(!stl_ut::AppendCStr(s_shader_source, p_s_shader_tail))
		return false;
	// add shader tail

	/*printf("-------------------------------------------------------------"
		"-------------------\nconvolution shader:\n%s\n------------------"
		"--------------------------------------------------------------\n",
		s_shader_source.c_str());*/
	// dump shader source so we know

#ifdef _DEBUG
	std::string s_shader_name;
	if(!stl_ut::AppendCStr(s_shader_name, (n_type == pass_Horiz)? "conv_horiz_pass" : "conv_vert_pass"))
		return false;
	for(int i = 0; i < n_kernel_num; ++ i) {
		char p_s_number[16];
		sprintf(p_s_number, "_%d", p_kernel_list[i]);
		if(!stl_ut::AppendCStr(s_shader_name, p_s_number))
			return false;
	}
	const char *p_s_shader_name = s_shader_name.c_str();
	//printf("building %s ...\n", p_s_shader_name);
	// debug
#else //_DEBUG
	const char *p_s_shader_name = (n_type == pass_Horiz)? "conv_horiz_pass" : "conv_vert_pass";
#endif //_DEBUG

	TShaderInfo t_fs(p_s_shader_name, TShaderInfo::proc_Fragment,
		s_shader_source.c_str(), "n_tex");
	if(!t_fs.p_high)
		return false;
	// create shader info; do *not* include pass index in shader name, turns out some
	// passes have the same shaders (they are compiled once, thanks to hash system)

	TPass t_pass = {0};
	if(!(t_pass.p_shader = CGLImageProc::p_CompileLinkReport(p_state, 0, 0, &t_fs, false)))
		return false;
	// compile shader

	t_pass.n_texture_num = n_kernel_num;
	for(int i = 0; i < n_kernel_num; ++ i) {
		if(!(t_pass.p_texture[i] = new CGLTexture_Rect(p_state, n_width,
		   n_height, n_tex_format)) || !t_pass.p_texture[i]->b_Status()) {
			SafeDeletePass(t_pass);
			return false;
		}
		// alloc, error check

		CGLTextureParams &r_params = t_pass.p_texture[i]->r_Parameters(p_state);
#if 1
		r_params.Set_Texture_Min_Filter(GL_NEAREST);
		r_params.Set_Texture_Mag_Filter(GL_NEAREST);
#else
		r_params.Set_Texture_Min_Filter(GL_LINEAR);
		r_params.Set_Texture_Mag_Filter(GL_LINEAR);
		bool _b_ = true;
		if(_b_) {
			_b_ = false;
			fprintf(stderr, "!!!!!!!!! warning: forcing linear filtering for convolution images !!!!!!!!!\n");
		}
		// debug
#endif
		// set nearest filtering
	}
	// alloc textures

	if(!(t_pass.p_fbo = CGLImageProc::p_GetRenderBuffer(n_width, n_height, n_kernel_num,
	   true, 0, false, false, 0, false, false, 0, false))) {
		SafeDeletePass(t_pass);
		return false;
	}
	// alloc renderbuffer

	t_pass.p_source = p_src_tex;
	// set source image

	m_pass_list.push_back(t_pass);
	// add pass to the list

	return true;
}

inline void CImageConvolver::SafeDeletePass(TPass &r_t_pass)
{
	if(r_t_pass.p_shader)
		CGLImageProc::ReturnShaderObject(r_t_pass.p_shader);
	for(int i = 0, n = r_t_pass.n_texture_num; i < n; ++ i) {
		if(r_t_pass.p_texture[i])
			delete r_t_pass.p_texture[i];
	}
}

inline void CImageConvolver::DeletePass(TPass &r_t_pass)
{
	CGLImageProc::ReturnShaderObject(r_t_pass.p_shader);
	for(int i = 0, n = r_t_pass.n_texture_num; i < n; ++ i)
		delete r_t_pass.p_texture[i];
}

inline uint16_t CImageConvolver::n_KernelSize(int n_width, int n_height)
{
	_ASSERTE(n_width > 0 && n_width < 256);
	_ASSERTE(n_height > 0 && n_height < 256);
	return (n_width << 8) | n_height;
}

int CImageConvolver::n_Max_DrawBuffer_Num()
{
	int n_max_draw_buffer_num;
	if(CGLExtensionHandler::b_SupportedExtension("GL_ARB_draw_buffers") ||
	   CGLExtensionHandler::b_Support_OpenGL(2, 0))
		glGetIntegerv(GL_MAX_DRAW_BUFFERS, &n_max_draw_buffer_num);
	else
		n_max_draw_buffer_num = 1;
	// get number of draw buffers

	_ASSERTE(n_max_draw_buffer_num <= max_DrawBuffers);
	// some arrays have max_DrawBuffers elements

	return n_max_draw_buffer_num;
}

/*
 *								=== ~CImageConvolver ===
 */
