/*
								+-----------------------------------+
								|                                   |
								|  *** Mk2 convolution manager ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2008   |
								|                                   |
								|            Convolve.h             |
								|                                   |
								+-----------------------------------+
*/

#ifndef __WALDBOOST_CONVOLUTION_MANAGER_INCLUDED
#define __WALDBOOST_CONVOLUTION_MANAGER_INCLUDED

#define DISABLE_TEXTURE_RECT_NV_WARNING
// disables warning about non-standard samplerRect and textureRect
// on nvidia machines, however might not work anywhere else

#include "./UberLame_src/Vector.h"
#include "./UberLame_src/MinMax.h"
#include "Load_XML.h"

/**
 *	struct TKernelSize
 *		- convolution kernel size (integer width and height)
 */
struct TKernelSize {
	int n_width;
	int n_height;

	/**
	 *	int TKernelSize::Dump() const
	 *		- prints object information to stdout (debug)
	 *		- returns number of printed characters (just so it can be used with std::mem_fun)
	 */
	int Dump() const;

	/**
	 *	bool TKernelSize::operator ==(const TWeakClassifier &r_t_classify) const
	 *		- returns true if this convolution size equals to block
	 *		  size of weak LRD classifier r_t_classify, otherwise returns false
	 */
	bool operator ==(const TWeakClassifier &r_t_classify) const;
};

/**
 *	struct TConvolutionList
 *		- list of all convolution kernels, required by a classifier cascade
 */
struct TConvolutionList {
	std::vector<TKernelSize> convolution_list;

	/**
	 *	bool TConvolutionList::Create(const TWaldBoostClassifier &r_t_classifier)
	 *		- creates list of all convolution kernels,
	 *		  required by a classifier cascade r_t_classifier
	 *		- returns true on success, false on failure
	 */
	bool Create(const TWaldBoostClassifier &r_t_classifier);

	/**
	 *	void TConvolutionList::Dump() const
	 *		- prints object information to stdout (debug)
	 */
	void Dump() const;

protected:
	static inline bool b_SmallerConvolution(const TKernelSize &r_t_a, const TKernelSize &r_t_b);
};

/**
 *	struct TImage
 *		- information about image placement inside image atlas
 */
struct TImage {
	Vector2i v_position, v_size;

	/**
	 *	TImage::TImage(Vector2i _v_size)
	 *		- default constructor; creates image with size _v_size and places it to origin
	 */
	TImage(Vector2i _v_size);

	/**
	 *	Vector2i TImage::v_Min() const
	 *		- returns coordinate of lower-left image corner
	 */
	Vector2i v_Min() const;

	/**
	 *	Vector2i TImage::v_Max(int n_padding) const
	 *		- returns coordinate of upper-right image corner,
	 *		  while considering padding n_padding units
	 */
	Vector2i v_Max(int n_padding) const;

	/**
	 *	unsigned int TImage::n_Area(int n_padding) const
	 *		- returns image area, along with padding n_padding units
	 */
	unsigned int n_Area(int n_padding) const;

	/**
	 *	void TImage::Place(const Vector2i &r_v_min)
	 *		- sets image position to r_v_min
	 */
	void Place(const Vector2i &r_v_min);

	/**
	 *	bool TImage::b_Collide(const TImage &r_t_other, int n_padding)
	 *		- returns true if minimal distance between this
	 *		  image and r_t_other is smaller than n_padding,
	 *		  otherwise returns false
	 */
	bool b_Collide(const TImage &r_t_other, int n_padding);

	/**
	 *	void TImage::Shift_Right(const TImage &r_t_other, int n_padding)
	 *		- collision response
	 *		- shifts this image right so minimal distance
	 *		  to r_t_other is at least equal to n_padding
	 */
	void Shift_Right(const TImage &r_t_other, int n_padding);

	/**
	 *	static bool TImage::b_Disjunct(int n_min1, int n_max1, int n_min2, int n_max2)
	 *		- utility function
	 *		- returns true if two intervals [n_min1, n_max1] and
	 *		  [n_min2, n_max2] are disjunct. otherwise returns false.
	 */
	static bool b_Disjunct(int n_min1, int n_max1, int n_min2, int n_max2);
};

/**
 *	struct TImagePage
 *		- a single page of image atlas
 *		- has ability to insert and automatically place images
 */
struct TImagePage {
	int n_width, n_height;
	int n_padding;
	std::vector<TImage*> image_list;

	/**
	 *	int TImagePage::n_InsertImage(TImage &r_t_image)
	 *		- inserts image r_t_image into the page
	 *		- returns 1 on success, 0 on failure (image wouldn't fit,
	 *		  there is not enough continouos free space in the page)
	 *		  or -1 on failure (not enough memory)
	 *		- note r_t_image is stored only by reference, so it's important
	 *		  it doesn't change address while using this image page
	 *		- note this is very plain and cannot yield good efficiency
	 *		  unless images are inserted with some strategy (but it is still
	 *		  much better and faster than lightmap page packer, for example
	 *		  the one in Half-Life 2)
	 */
	int n_InsertImage(TImage &r_t_image);

	/**
	 *	int TImagePage::n_Used_Area() const
	 *		- returns area, used by images (sum of image areas with no padding,
	 *		  not area of some bounding object arround all images)
	 */
	int n_Used_Area() const;

protected:
	class CSolveCollisions;
	class CSumArea;
};

/**
 *	struct TDownscaleImageSet
 *		- image atlas, containing pyramid of downscaled images
 *		- note this employs simple image stacking strategy which yields
 *		  average space utilization above 90% (but isn't proven to be optimal)
 */
struct TDownscaleImageSet {
	std::vector<TImage> image_list;
	std::vector<TImagePage> page_list;

	/**
	 *	bool TDownscaleImageSet::Create(int n_width, int n_height, float f_factor = .9f,
	 *		int n_min_width = 24, int n_min_height = 24,
	 *		int n_initial_page_size = 256, int n_max_optimal_page_size = 512,
	 *		int n_max_page_size = 1024, int n_padding = 1,
	 *		int n_page_tighten_thresh = 32, int n_page_size_margin = 8)
	 *		- creates a new image pyramid
	 *		- n_width, n_height is size of the image in pixels,
	 *		  pyramid contains images, scaled by factor f_factor (must be smaller
	 *		  than 1, not checked) down to minimal size n_min_width per n_min_height
	 *		  pixels, whichever dimension is reached first
	 *		- n_initial_page_size is initial page size, n_max_optimal_page_size is optimal
	 *		  page size (when page is containing small images, do not make it larger than this),
	 *		  n_max_page_size is maximal page size (used only for pages, containing images larger
	 *		  than n_max_optimal_page_size), n_padding is image padding
	 *		- after all images have been inserted, pages with free space above
	 *		  n_page_tighten_thresh arround any of it's sides are tightened with
	 *		  margin n_page_size_margin (dimensions will be whole multiple of
	 *		  n_page_size_margin)
	 *		- returns true on success, false on failure
	 */
	bool Create(int n_width, int n_height, float f_factor = .9f,
		int n_min_width = 24, int n_min_height = 24,
		int n_initial_page_size = 256, int n_max_optimal_page_size = 512,
		int n_max_page_size = 1024, int n_padding = 1,
		int n_page_tighten_thresh = 32, int n_page_size_margin = 8);

	/**
	 *	void TDownscaleImageSet::Dump() const
	 *		- prints object information to stdout (debug)
	 */
	void Dump() const;
};

/**
 *	class CImageDownscaler
 *		- simple OpenGL renderer of downscaled images
 *		- note this doesn't use any display lists (uses VBO instead)
 */
class CImageDownscaler {
protected:
	struct TPageGeometry {
		int n_first, n_count;
		CGLFrameBuffer_FBO *p_fbo;
		CGLTexture_Rect *p_texture;
	};

	CGLVertexBufferObject *m_p_vertex_vbo, *m_p_index_vbo;

	std::vector<TPageGeometry> m_page_list; // list of render passes

public:
	/**
	 *	CImageDownscaler::CImageDownscaler(CGLState *p_state, TDownscaleImageSet &r_image_set)
	 *		- default constructor; creates textures and vertex arrays
	 *		  for rendering images into them
	 *		- p_state is OpenGL state guard, r_image_set is reference
	 *		  to downscaled image set description
	 *		- note it is recommended to call b_Status() afterwards to check
	 *		  wheter constructor succeeded
	 */
	CImageDownscaler(CGLState *p_state, TDownscaleImageSet &r_image_set);

	/**
	 *	CImageDownscaler::~CImageDownscaler()
	 *		- destructor
	 */
	~CImageDownscaler();

	/**
	 *	bool CImageDownscaler::b_Status() const
	 *		- returns true if constructor succeeded, otherwise returns false
	 */
	bool b_Status() const;

	/**
	 *	bool CImageDownscaler::BuildDownscaledImages(CGLState *p_state,
	 *		CGLTexture_Rect *p_src_image)
	 *		- renders downscaled images
	 *		- p_state is OpenGL state guard, p_src_image is source image
	 *		- returns true on success, false on failure
	 *		- note p_src_image should have minification filter GL_LINEAR_MIPMAP_LINEAR
	 *		  for the best image quality
	 */
	bool BuildDownscaledImages(CGLState *p_state, CGLTexture_Rect *p_src_image);

	/**
	 *	int CImageDownscaler::n_Page_Num() const
	 *		- returns number of pages (equals to number of pages in image set,
	 *		  passed to constructor)
	 */
	int n_Page_Num() const;

	/**
	 *	CGLTexture_Rect *CImageDownscaler::p_Page_Tex(int n_index)
	 *		- returns texture, containing image page with zero-based index n_index
	 */
	CGLTexture_Rect *p_Page_Tex(int n_index);

protected:
	static inline void DeletePageTex(TPageGeometry &r_t_page);
};

/**
 *	class CImageConvolver
 *		- state-of-art simple convolution manager
 *		- creates convolution plan for effectively calculating
 *		  set of convolutions from the same image
 */
class CImageConvolver {
public:
	enum {
		max_DrawBuffers = 16,
		pass_Horiz = 0,
		pass_Vert
	};

	struct TFinalConvolution {
		CGLTexture_Rect *p_texture; // reference to TPass::p_texture
		uint16_t n_kernel_size; // width, height as 8:8 HI:LO (width in HI)
		Vector2f v_tex_offset; // for convolution using bilinear filtering
	};

protected:
	struct TPass {
		CGLTexture_Rect *p_source; // 0 means source image, otherwise it's reference to earlier pass texture
		CGLFrameBuffer_FBO *p_fbo; // framebuffer (can't use global, might require different number of texture attachments)
		CGLShaderObject *p_shader; // shader
		int n_texture_num;
		CGLTexture_Rect *p_texture[max_DrawBuffers]; // dest textures (allocated here)
	};

	std::vector<TPass> m_pass_list; // list of rendering passes
	std::vector<TFinalConvolution> m_convolution_list; // list of final textures
	CGLVertexBufferObject *m_p_vertex_vbo;
	bool m_b_status;

	bool m_b_src_image_needs_bilinear;

	class CWorkLoadStrategy;

public:
	/**
	 *	CImageConvolver::CImageConvolver(CGLState *p_state, int n_width, int n_height,
	 *		const TConvolutionList &r_t_conv_list)
	 *		- default constructor; creates convolution plan, compiles all shaders and
	 *		  allocates all buffers, required to calculate given set of convolutions
	 *		- p_state is OpenGL state guard, n_width and n_height is input image size
	 *		  and r_t_conv_list is list of required convolutions
	 *		- note convolutions of size up to 2x2 are implemented using bilinear
	 *		  filtering when sampling textures, for that reason v_TextureOffset()
	 *		  supplies sub-pixel texture offset
	 *		- note it is recommended to call b_Status() afterwards to check
	 *		  wheter constructor succeeded
	 */
	CImageConvolver(CGLState *p_state, int n_width, int n_height,
		const TConvolutionList &r_t_conv_list);

	/**
	 *	CImageConvolver::~CImageConvolver()
	 *		- destructor
	 */
	~CImageConvolver();

	/**
	 *	bool CImageConvolver::b_Status() const
	 *		- returns true if constructor succeeded, otherwise returns false
	 */
	bool b_Status() const;

	/**
	 *	bool CImageConvolver::ConvolveImage(CGLState *p_state, CGLTexture_Rect *p_src_image)
	 *		- calculates convolutions
	 *		- p_state is OpenGL state guard and p_src_image is source image
	 *		  (note it's dimensions are ignored and are expected to equal
	 *		  to dimensions, passed to constructor)
	 *		- returns true on success, false on failure
	 */
	bool ConvolveImage(CGLState *p_state, CGLTexture_Rect *p_src_image);

	/**
	 *	int CImageConvolver::n_Kernel_Num() const
	 *		- returns number of calculated kernels (equals to number of kernels
	 *		  in convolution list, passed to constructor)
	 */
	int n_Kernel_Num() const;

	/**
	 *	Vector2i CImageConvolver::v_Kernel(int n_index) const
	 *		- returns size of kernel with zero-based index n_index
	 */
	Vector2i v_Kernel(int n_index) const;

	/**
	 *	CGLTexture_Rect *CImageConvolver::p_Image(int n_index)
	 *		- returns texture, containing convolution with zero-based index n_index
	 *		- note this returns 0 when convolution is in source image
	 *		  (for example convolution with kernel 1x1)
	 */
	CGLTexture_Rect *p_Image(int n_index);

	/**
	 *	Vector2f CImageConvolver::v_TextureOffset(int n_index) const
	 *		- returns sub-texel offset to calculate convolution with zero-based
	 *		  index n_index using bilinear filter in texturing unit
	 */
	Vector2f v_TextureOffset(int n_index) const;

	/**
	 *	bool CImageConvolver::b_SourceImageNeedsBilinearFilter() const
	 *		- returns true in case source image needs bilinear filter (GL_LINEAR),
	 *		  otherwise returns false (meaning GL_NEAREST suffices)
	 *		- note this happens when convolutions 1x2, 2x1 or 2x2 are required
	 */
	bool b_SourceImageNeedsBilinearFilter() const; // for fake convolution at computation-time

	/**
	 *	void CImageConvolver::Dump() const
	 *		- prints object information to stdout (debug)
	 */
	void Dump() const;

protected:
	bool Create_Pass(CGLState *p_state, int n_width, int n_height, GLenum n_tex_format,
		CGLTexture_Rect *p_src_tex, int n_kernel_num, const int *p_kernel_list, int n_type);
	static inline void SafeDeletePass(TPass &r_t_pass);
	static inline void DeletePass(TPass &r_t_pass);
	static inline uint16_t n_KernelSize(int n_width, int n_height);
	static int n_Max_DrawBuffer_Num();
};

#endif //__WALDBOOST_CONVOLUTION_MANAGER_INCLUDED
