/*
								+----------------------------------+
								|                                  |
								|    ***  Bit array class   ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2005   |
								|                                  |
								|           BitArray.cpp           |
								|                                  |
								+----------------------------------+
*/

/*
 *	passed code revision at 2007-03-15
 *
 *	improved code to meet berLame quality standards
 *	now using unsigned integers for lengths instead of signed
 *	made CBit owned by CBitArray (as was logical)
 *	fixed hack in constant bit access using operator []
 *	renamed CBitArray::Free() to CBitArray::Erase()
 *	CBitArray::operator ==() now returns wheter arrays are equal (bool)
 *	instead of integer equality sign which is now returned by CBitArray::n_Compare()
 *	added inequality comparison operators
 *	removed most of non-operator functions (Set_Bit_High() and Set_Bit_Low()
 *	remains only for preformance purposes)
 *	most stuff is now inline, the whole thing is significantly faster
 *
 *	2007-06-26
 *
 *	fixed some bugs in array reallocation using CBitArray::Extend and
 *	in array copy operator CBitArray::operator =
 *
 *	2007-03-04
 *
 *	changed integer types for more safety (now using uint32_t type as array
 *	base unit and unsigned long to store array length)
 *
 *	2008-12-28
 *
 *	exposed internal data buffer for easier and faster serialization
 *	fixed error which occured when copying bits from one array to another
 *	(CBit instances were copied, no array writes occured)
 *
 *	2009-04-23
 *
 *	motivation for writing this file:
 *		std::vector<bool> is troublesome and generally two considered options are
 *			to erase (not deprecate) it, or rename it
 *		std::vector<bool> doesn't offer access to it's data buffer
 *		std::vector<bool> doesn't offer Set_Bit_Low() and Set_Bit_High() functions,
 *			therefore theoretically offers lower reachable performance
 *
 *	added CBitArray::n_Capacity(), CBitArray::Reserve() functions for more efficient
 *	(re)allocation
 *
 *	renamed CBitArray::CBit to CBitArray::CBitReference
 *
 *	added CBitArray::CBitReference::Raise() and CBitArray::CBitReference::Clear()
 *	functions for more efficient writes to individual bits (note
 *	CBitArray::CBitReference::operator =(bool) contains branch, not present
 *	in either of above functions)
 *
 */

#include <vector>
#include "CallStack.h"
#include <string.h>
#include <stdio.h>
#include "BitArray.h"

/*
 *								=== CBitArray ===
 */

/*
 *	CBitArray::CBitArray(unsigned long n_length = 0)
 *		- default constructor with array length specifier
 *		- see CBitArray::n_Length() to find out if there was enough memory
 */
CBitArray::CBitArray(unsigned long n_length /*= 0*/)
	:m_p_array(0), m_n_array_size(0), m_n_array_used(0)
{
	Alloc(n_length);
	// set array empty and alloc
}

/*
 *	CBitArray::~CBitArray()
 *		- destructor
 */
CBitArray::~CBitArray()
{
	Erase();
}

/*
 *	bool CBitArray::Alloc(unsigned long n_length)
 *		- allocates array to lenght n_length bits
 *		- original contents are erased
 *		- returns true on success, false on failure (not enough memory)
 */
bool CBitArray::Alloc(unsigned long n_length)
{
	Erase();

	m_n_array_size = (n_length + 31) / 32;
	// calc necessary buffer length

	if(!(m_p_array = new uint32_t[m_n_array_size])) {
		m_n_array_size = 0;
		m_n_array_used = 0;
		return false;
	}
#ifdef _DEBUG
	memset(m_p_array, 0, m_n_array_size * sizeof(uint32_t));
#endif //_DEBUG
	// alloc

	m_n_array_used = n_length;
	// set length in bits

	return true;
}

/*
 *	bool CBitArray::Extend(unsigned long n_length)
 *		- extends array to lenght n_length bits
 *		- returns true on success, false on failure (not enough memory)
 */
bool CBitArray::Extend(unsigned long n_length)
{
	if(!m_n_array_size)
		return Alloc(n_length);
	// use Alloc() for empty array

	unsigned long n_new_size = (n_length + 31) / 32;
	// calc new length

	if(n_new_size <= m_n_array_size) {
		m_n_array_used = n_length;

		return true;
	}
	// if it's equally long (or even shorter), we don't have to do anything

	uint32_t *p_array;
	if(!(p_array = new uint32_t[n_new_size]))
		return false;
	// alloc new buffer

#ifdef _DEBUG
	memset(p_array, 0, n_new_size * sizeof(uint32_t));
#endif //_DEBUG
	memcpy(p_array, m_p_array, m_n_array_size * sizeof(uint32_t));
	// copy contents

	delete[] m_p_array;
	m_p_array = p_array;
	m_n_array_size = n_new_size;
	m_n_array_used = n_length;
	// delete old buffer

	return true;
}

/*
 *	bool CBitArray::Reserve(unsigned long n_length)
 *		- extends bufer to total lenght n_length bits,
 *		  but does not change array length (changes capacity)
 *		- returns true on success, false on failure (not enough memory)
 */
bool CBitArray::Reserve(unsigned long n_length)
{
	if(!m_n_array_size) {
		bool b_result = Alloc(n_length);
		m_n_array_size = 0;
		return b_result;
	}
	// use Alloc() for empty array

	unsigned long n_new_size = (n_length + 31) / 32;
	// calc new length

	if(n_new_size <= m_n_array_size)
		return true;
	// if it's equally long (or even shorter), we don't have to do anything

	uint32_t *p_array;
	if(!(p_array = new uint32_t[n_new_size]))
		return false;
	// alloc new buffer

#ifdef _DEBUG
	memset(p_array, 0, n_new_size * sizeof(uint32_t));
#endif //_DEBUG
	memcpy(p_array, m_p_array, m_n_array_size * sizeof(uint32_t));
	// copy contents

	delete[] m_p_array;
	m_p_array = p_array;
	m_n_array_used = n_length;
	// delete old buffer

	return true;
}

/*
 *	void CBitArray::Erase()
 *		- erase array, free it's memory
 */
void CBitArray::Erase()
{
	m_n_array_size = 0;
	m_n_array_used = 0;
	if(m_p_array) {
		delete[] m_p_array;
		m_p_array = 0;
	}
	// delete array, set to zero
}

/*
 *	void CBitArray::Swap(CBitArray &r_other)
 *		- swaps this array with r_t_other
 */
void CBitArray::Swap(CBitArray &r_other)
{
	std::swap(m_p_array, r_other.m_p_array);
	std::swap(m_n_array_size, r_other.m_n_array_size);
	std::swap(m_n_array_used, r_other.m_n_array_used);
}

/*
 *	bool CBitArray::operator =(bool b_value)
 *		- sets all bits in array to b_value
 *		- returns b_value
 */
bool CBitArray::operator =(bool b_value)
{
	memset(m_p_array, (b_value)? 0xff : 0x00, m_n_array_size * sizeof(uint32_t));
	// set whole array to 0-s or 1-s, depending on b_value

	return b_value;
}

/*
 *	void CBitArray::Invert()
 *		- inverts all bits in the array
 */
void CBitArray::Invert()
{
	for(uint32_t *p_dest = m_p_array, *p_end = m_p_array + m_n_array_size; p_dest != p_end; ++ p_dest)
		*p_dest = ~*p_dest;
	// invert the whole bit array
}

/*
 *	bool CBitArray::operator =(const CBitArray &r_array)
 *		- copies r_array to this bit array
 *		- returns true if there was enough memory and array was successfully copied, otherwise false
 */
bool CBitArray::operator =(const CBitArray &r_array)
{
	if(!Extend(r_array.n_Length()))
		return false;
	memcpy(m_p_array, r_array.m_p_array, m_n_array_size * sizeof(uint32_t));
	// alloc and copy

	return true;
}

/*
 *	bool CBitArray::operator +=(const CBitArray &r_array)
 *		- array concatenation
 *		- return true in case there was enough memory and arrays
 *		  were concatenated, otherwise false
 */
bool CBitArray::operator +=(const CBitArray &r_array)
{
	unsigned long n_last_bit;
	if(!Extend((n_last_bit = n_Length()) + r_array.n_Length()))
		return false;
	// alloc

	for(unsigned long i = 0; i < r_array.m_n_array_used;) {
		if(!(n_last_bit & 31)) { // dest array aligned
			if(!(i & 31)) { // source array aligned
				m_p_array[n_last_bit >> 5] = r_array.m_p_array[i >> 5];
				i += 32;
				n_last_bit += 32;
			} else { // source array un-aligned
				unsigned long n_bits = 32 - (i & 31);
				// number of bits to copy

				m_p_array[n_last_bit >> 5] = r_array.m_p_array[i >> 5] >> (i & 31);
				i += n_bits;
				n_last_bit += n_bits;
			}
		} else { // dest array un-aligned
			if(!(i & 31)) { // source array aligned
				unsigned long n_bits = /*32 -*/ (n_last_bit & 31);
				// number of bits to copy

				m_p_array[n_last_bit >> 5] &= (1 << (/*32 -*/ n_bits)) - 1;
				m_p_array[n_last_bit >> 5] |= r_array.m_p_array[i >> 5] << (n_last_bit & 31);

				i += n_bits;
				n_last_bit += 32 - n_bits; // save 1 sub by reordering "32 - " part here
			} else { // source array un-aligned
				unsigned long n_bits = 32 - (((i & 31) < (n_last_bit & 31))?
					(n_last_bit & 31) : (i & 31));
				// number of bits to copy

				m_p_array[n_last_bit >> 5] &= (1 << (n_last_bit & 31)) - 1;
				m_p_array[n_last_bit >> 5] |= (r_array.m_p_array[i >> 5] >> (i & 31)) <<
					(n_last_bit & 31);

				i += n_bits;
				n_last_bit += n_bits;
			}
		}
	}

	return true;
}

/*
 *	bool CBitArray::operator ==(bool b_value) const
 *		- return true in case all bits in array are equal to b_value, otherwise false
 *		- always returns true for an empty array
 */
bool CBitArray::operator ==(bool b_value) const
{
	for(unsigned long i = 0; i < (m_n_array_used >> 5); i ++) {
		if(m_p_array[i] != (unsigned)((b_value)? -1 : 0)) // -1 = all ones in binary
			return false;
	}
	if(m_n_array_used & 31) {
		int n_mask = (1 << (m_n_array_used & 31)) - 1; // mask for remaining bits
		return (m_p_array[m_n_array_used >> 5] & n_mask) == (unsigned)((-1 * b_value) & n_mask);
	}

	return true;
}

/*
 *	bool CBitArray::operator ==(const CBitArray &r_array) const
 *		- returns true in case array lengths and contents are equal, otherwise false
 */
bool CBitArray::operator ==(const CBitArray &r_array) const
{
	if(r_array.m_n_array_used != m_n_array_used)
		return false;
	unsigned long n_lesser_size = (m_n_array_used > r_array.m_n_array_used)?
		r_array.m_n_array_used : m_n_array_used;
	for(unsigned long i = 0; i < (n_lesser_size >> 5); i ++) {
		if(m_p_array[i] != r_array.m_p_array[i])
			return false;
	}
	if(n_lesser_size & 31) {
		int n_mask = (1 << (n_lesser_size & 31)) - 1; // mask for remaining bits
		return (m_p_array[n_lesser_size >> 5] & n_mask) ==
			(r_array.m_p_array[n_lesser_size >> 5] & n_mask);
	}

	return true;
}

/*
 *	int CBitArray::n_Compare(const CBitArray &r_array) const
 *		- aray comparison in strcmp() like fashion
 *		- compares bits in natural order, first bit decides
 *		- in case all bits are equal, compare lengths
 *		- returns 1 if this > r_array, 0 if this == r_array or -1 if this < r_array
 */
int CBitArray::n_Compare(const CBitArray &r_array) const
{
	unsigned long n_lesser_size = (m_n_array_used > r_array.m_n_array_used)?
		r_array.m_n_array_used : m_n_array_used;
	for(unsigned long i = 0; i < (n_lesser_size >> 5); i ++) {
		if(m_p_array[i] != r_array.m_p_array[i])
			return (r_array.m_p_array[i] > m_p_array[i])? -1 : 1;
	}
	if(n_lesser_size & 31) {
		unsigned long n_mask = (1 << (n_lesser_size & 31)) - 1; // mask for remaining bits
		if((m_p_array[n_lesser_size >> 5] & n_mask) !=
		   (r_array.m_p_array[n_lesser_size >> 5] & n_mask)) {
			return ((r_array.m_p_array[n_lesser_size >> 5] & n_mask) >
				(m_p_array[n_lesser_size >> 5] & n_mask))? -1 : 1;
		}
	}

	return (r_array.m_n_array_used > m_n_array_used)? -1 :
		((m_n_array_used == r_array.m_n_array_used)? 0 : 1);
}

/*
 *	void CBitArray::FillBuffer(const void *p_data)
 *		- fills data buffer with data from p_data
 *		- note p_data must contain n_Buffer_Size() * sizeof(uint32_t) bytes
 */
void CBitArray::FillBuffer(const void *p_data)
{
	memcpy(m_p_array, p_data, m_n_array_size * sizeof(uint32_t));
}

/*
 *	bool CBitArray::Store(FILE *p_fw) const
 *		- writes array to file p_fw (should be opened for writing in binary mode)
 *		- returns true if successful, otherwise false
 */
bool CBitArray::Store(FILE *p_fw) const
{
	return fwrite(&m_n_array_used, sizeof(size_t), 1, p_fw) == 1 &&
	   fwrite(m_p_array, sizeof(uint32_t), m_n_array_size, p_fw) == m_n_array_size;
}

/*
 *	bool CBitArray::Load(FILE *p_fr)
 *		- loads an array from file p_fr (should be opened for reading in binary mode)
 *		- returns true if successful, otherwise false
 */
bool CBitArray::Load(FILE *p_fr)
{
	unsigned long n_array_used;
	return fread(&n_array_used, sizeof(size_t), 1, p_fr) == 1 && Extend(n_array_used) &&
		fread(m_p_array, sizeof(uint32_t), m_n_array_size, p_fr) == m_n_array_size;
}

/*
 *								=== ~CBitArray ===
 */
