/*
								+----------------------------------+
								|                                  |
								|    ***  Bit array class   ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2005   |
								|                                  |
								|            BitArray.h            |
								|                                  |
								+----------------------------------+
*/

/*
 *	passed code revision at 2007-03-15
 *
 *	improved code to meet berLame quality standards
 *	now using unsigned integers for lengths instead of signed
 *	made CBit owned by CBitArray (as was logical)
 *	fixed hack in constant bit access using operator []
 *	renamed CBitArray::Free() to CBitArray::Erase()
 *	CBitArray::operator ==() now returns wheter arrays are equal (bool)
 *	instead of integer equality sign which is now returned by CBitArray::n_Compare()
 *	added inequality comparison operators
 *	removed most of non-operator functions (Set_Bit_High() and Set_Bit_Low()
 *	remains only for preformance purposes)
 *	most stuff is now inline, the whole thing is significantly faster
 *
 *	2007-06-26
 *
 *	fixed some bugs in array reallocation using CBitArray::Extend and
 *	in array copy operator CBitArray::operator =
 *
 *	2007-03-04
 *
 *	changed integer types for more safety (now using uint32_t type as array
 *	base unit and unsigned long to store array length)
 *
 *	2008-12-28
 *
 *	exposed internal data buffer for easier and faster serialization
 *	fixed error which occured when copying bits from one array to another
 *	(CBit instances were copied, no array writes occured)
 *
 *	2009-04-23
 *
 *	motivation for writing this file:
 *		std::vector<bool> is troublesome and generally two considered options are
 *			to erase (not deprecate) it, or rename it
 *		std::vector<bool> doesn't offer access to it's data buffer
 *		std::vector<bool> doesn't offer Set_Bit_Low() and Set_Bit_High() functions,
 *			therefore theoretically offers lower reachable performance
 *
 *	added CBitArray::n_Capacity(), CBitArray::Reserve() functions for more efficient
 *	(re)allocation
 *
 *	renamed CBitArray::CBit to CBitArray::CBitReference
 *
 *	added CBitArray::CBitReference::Raise() and CBitArray::CBitReference::Clear()
 *	functions for more efficient writes to individual bits (note
 *	CBitArray::CBitReference::operator =(bool) contains branch, not present
 *	in either of above functions)
 *
 */

#ifndef __BIT_ARRAY_INCLUDED
#define __BIT_ARRAY_INCLUDED

#include "Integer.h"

class CBitArray {
public:
	typedef unsigned long size_t;
	typedef unsigned long offset_t;

	/*
	 *	class CBitArray::CBitReference
	 *		- write-able bit class
	 */
	class CBitReference {
	protected:
		CBitArray *m_p_array;
		offset_t m_n_my_index;

	public:
		inline CBitReference(CBitArray *p_array, offset_t n_index)
			:m_p_array(p_array), m_n_my_index(n_index)
		{}

		/*
		 *	inline bool CBitArray::CBitReference::operator =(bool b_value)
		 *		- writes b_value to parent array
		 *		- returns b_value
		 */
		inline bool operator =(bool b_value)
		{
			if(b_value)
				m_p_array->Set_Bit_High(m_n_my_index);
			else
				m_p_array->Set_Bit_Low(m_n_my_index);
			return b_value;
		}

		/*
		 *	inline void CBitArray::CBitReference::Raise()
		 *		- writes 1 to this bit
		 */
		inline void Raise()
		{
			m_p_array->Set_Bit_High(m_n_my_index);
		}

		/*
		 *	inline void CBitArray::CBitReference::Clear()
		 *		- writes 0 to this bit
		 */
		inline void Clear()
		{
			m_p_array->Set_Bit_High(m_n_my_index);
		}

		/*
		 *	inline bool CBitArray::CBitReference::operator =(const CBitReference &r_bit)
		 *		- writes value of r_bit to parent array
		 *		- returns b_value
		 */
		inline bool operator =(const CBitReference &r_bit)
		{
			return *this = (bool)r_bit;
		}

		/*
		 *	inline bool CBitArray::CBitReference::operator ==(const CBitReference &r_bit)
		 *		- comparison operator for two bits (do not compare
		 *		  array and index, compare referenced value)
		 */
		inline bool operator ==(const CBitReference &r_bit)
		{
			return (bool)*this == (bool)r_bit;
		}

		/*
		 *	inline CBitArray::CBitReference::operator bool() const
		 *		- returns current contents of bit in parent array
		 */
		inline operator bool() const
		{
			return (*(const CBitArray*)m_p_array)[m_n_my_index];
		}
	};

protected:
	uint32_t *m_p_array;
	size_t m_n_array_size;
	size_t m_n_array_used;

public:
	/*
	 *	CBitArray::CBitArray(size_t n_length = 0)
	 *		- default constructor with array length specifier
	 *		- see CBitArray::n_Length() to find out if there was enough memory
	 */
	CBitArray(size_t n_length = 0);

	/*
	 *	CBitArray::~CBitArray()
	 *		- destructor
	 */
	~CBitArray();

	/*
	 *	bool CBitArray::Alloc(size_t n_length)
	 *		- allocates array to lenght n_length bits
	 *		- original contents are erased (use Extend if required)
	 *		- returns true on success, false on failure (not enough memory)
	 */
	bool Alloc(size_t n_length);

	/*
	 *	bool CBitArray::Extend(size_t n_length)
	 *		- extends array to total lenght n_length bits
	 *		- returns true on success, false on failure (not enough memory)
	 */
	bool Extend(size_t n_length);

	/*
	 *	bool CBitArray::Reserve(size_t n_length)
	 *		- extends bufer to total lenght n_length bits,
	 *		  but does not change array length (changes capacity)
	 *		- returns true on success, false on failure (not enough memory)
	 */
	bool Reserve(size_t n_length);

	/*
	 *	void CBitArray::Erase()
	 *		- erase array, free it's memory
	 */
	void Erase();

	/*
	 *	void CBitArray::Swap(CBitArray &r_other)
	 *		- swaps this array with r_t_other
	 */
	void Swap(CBitArray &r_other);

	/*
	 *	inline size_t CBitArray::n_Length() const
	 *		- returns length of array in bits
	 */
	inline size_t n_Length() const
	{
		return m_n_array_used;
	}

	/*
	 *	inline size_t CBitArray::n_Capacity() const
	 *		- returns maximal length of array in bits without reallocating
	 */
	inline size_t n_Capacity() const
	{
		return m_n_array_size * 32;
	}

	/*
	 *	inline CBitReference CBitArray::operator [](offset_t n_index)
	 *		- returns write-able n_index'th bit
	 */
	inline CBitReference operator [](offset_t n_index)
	{
		return CBitReference(this, n_index);
	}

	/*
	 *	inline bool CBitArray::operator [](offset_t n_index) const
	 *		- returns contents of n_index'th bit
	 */
	inline bool operator [](offset_t n_index) const
	{
		return (m_p_array[n_index >> 5] >> (n_index & 31)) & 1;
	}

	/*
	 *	bool CBitArray::operator =(bool b_value)
	 *		- sets all bits in array to b_value
	 *		- returns b_value
	 */
	bool operator =(bool b_value);

	/*
	 *	void CBitArray::Invert()
	 *		- inverts all bits in the array
	 */
	void Invert();

	/*
	 *	bool CBitArray::operator =(const CBitArray &r_array)
	 *		- copies r_array to this bit array
	 *		- returns true if there was enough memory and array was
	 *		  successfully copied, otherwise false
	 */
	bool operator =(const CBitArray &r_array);

	/*
	 *	bool CBitArray::operator +=(const CBitArray &r_array)
	 *		- array concatenation
	 *		- return true in case there was enough memory and arrays
	 *		  were concatenated, otherwise false
	 */
	bool operator +=(const CBitArray &r_array);

	/*
	 *	bool CBitArray::operator ==(bool b_value) const
	 *		- return true in case all bits in array are equal to b_value, otherwise false
	 *		- always returns true for an empty array
	 */
	bool operator ==(bool b_value) const;

	/*
	 *	bool CBitArray::operator ==(const CBitArray &r_array) const
	 *		- returns true in case array lengths and contents are equal, otherwise false
	 */
	bool operator ==(const CBitArray &r_array) const;

	/*
	 *	int CBitArray::n_Compare(const CBitArray &r_array) const
	 *		- aray comparison in strcmp() like fashion
	 *		- compares bits in natural order, first bit decides
	 *		- in case all bits are equal, compare lengths
	 *		- returns 1 if this > r_array, 0 if this == r_array or -1 if this < r_array
	 */
	int n_Compare(const CBitArray &r_array) const;

	/*
	 *	inline bool CBitArray::operator >(const CBitArray &r_array) const
	 *		- returns true in case this array is greater than r_array, otherwise false
	 */
	inline bool operator >(const CBitArray &r_array) const
	{
		return n_Compare(r_array) > 0;
	}

	/*
	 *	inline bool CBitArray::operator >=(const CBitArray &r_array) const
	 *		- returns true in case this array is greater or equal than r_array, otherwise false
	 */
	inline bool operator >=(const CBitArray &r_array) const
	{
		return n_Compare(r_array) >= 0;
	}

	/*
	 *	inline bool CBitArray::operator >(const CBitArray &r_array) const
	 *		- returns true in case this array is less than r_array, otherwise false
	 */
	inline bool operator <(const CBitArray &r_array) const
	{
		return n_Compare(r_array) < 0;
	}

	/*
	 *	inline bool CBitArray::operator <=(const CBitArray &r_array) const
	 *		- returns true in case this array is less or equal than r_array, otherwise false
	 */
	inline bool operator <=(const CBitArray &r_array) const
	{
		return n_Compare(r_array) <= 0;
	}

	/*
	 *	inline bool CBitArray::operator !=(const CBitArray &r_array) const
	 *		- returns true in case this array is not equal to r_array, otherwise false
	 */
	inline bool operator !=(const CBitArray &r_array) const
	{
		return !(*this == r_array); // operator == is faster than n_Compare
	}

	/*
	 *	inline void CBitArray::Set_Bit_High(offset_t n_index)
	 *		- sets n_index'th bit to 1
	 */
	inline void Set_Bit_High(offset_t n_index)
	{
		m_p_array[n_index >> 5] |= 1 << (n_index & 31);
	}

	/*
	 *	inline void CBitArray::Set_Bit_Low(offset_t n_index)
	 *		- sets n_index'th bit to 0
	 */
	inline void Set_Bit_Low(offset_t n_index)
	{
		m_p_array[n_index >> 5] &= ~(1 << (n_index & 31));
	}

	/*
	 *	inline size_t CBitArray::n_Buffer_Size() const
	 *		- return size of internal data buffer, in dwords
	 */
	inline size_t n_Buffer_Size() const
	{
		return m_n_array_size;
	}

	/*
	 *	inline const uint32_t *CBitArray::p_Buffer() const
	 *		- returns const pointer to data buffer
	 *		- note the buffer is n_Buffer_Size() long
	 */
	inline const uint32_t *p_Buffer() const
	{
		return m_p_array;
	}

	/*
	 *	inline uint32_t *CBitArray::p_Buffer()
	 *		- returns pointer to data buffer (do not delete it)
	 *		- note the buffer is n_Buffer_Size() long
	 */
	inline uint32_t *p_Buffer()
	{
		return m_p_array;
	}

	/*
	 *	void CBitArray::FillBuffer(const void *p_data)
	 *		- fills data buffer with data from p_data
	 *		- note p_data must contain n_Buffer_Size() * sizeof(uint32_t) bytes
	 */
	void FillBuffer(const void *p_data);

	/*
	 *	bool CBitArray::Store(FILE *p_fw) const
	 *		- writes array to file p_fw (should be opened for writing in binary mode)
	 *		- returns true if successful, otherwise false
	 */
	bool Store(FILE *p_fw) const;

	/*
	 *	bool CBitArray::Load(FILE *p_fr)
	 *		- loads an array from file p_fr (should be opened for reading in binary mode)
	 *		- returns true if successful, otherwise false
	 */
	bool Load(FILE *p_fr);
};

#endif //__BIT_ARRAY_INCLUDED
