/*
								+---------------------------------+
								|                                 |
								|  ***  Standard hash funcs  ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|             Hash.h              |
								|                                 |
								+---------------------------------+
*/

/*
 *
 *	2007-12-24
 *
 *	improved linux compatibility by adding posix integer types
 *
 *	2008-03-04
 *
 *	now using Integer.h header
 *
 *	2008-07-04
 *
 *	replaced all occurences of unsigned char with uint8_t
 *	changed CStreamHash::Process_Data(const unit8_t*, int) to
 *	CStreamHash::Process_Data(const void*, int) so there's no need
 *	to type cast data anymore (convenience improvement)
 *
 *	2009-05-04
 *
 *	fixed mixed windows / linux line endings
 *
 */

#ifndef __HASH_INCLUDED
#define __HASH_INCLUDED

#include "Integer.h"

/*
 *	struct TMD5
 *		- class, implementing message digest 5 algorithm by Ronald Rivest
 *		  ("RSA Data Security, Inc. MD5 Message-Digest Algorithm", RFC1321)
 *		- cooperates with CStreamHash<TMD5, 512, false>
 */
struct TMD5 {
protected:
	uint32_t m_n_hash[4];

public:
	TMD5();

	TMD5 operator =(const TMD5 &r_t_md5);

	bool operator ==(const TMD5 &r_t_md5) const;
	bool operator !=(const TMD5 &r_t_md5) const;
	bool operator >(const TMD5 &r_t_md5) const;
	bool operator <(const TMD5 &r_t_md5) const;
	bool operator >=(const TMD5 &r_t_md5) const;
	bool operator <=(const TMD5 &r_t_md5) const;

	inline uint32_t &operator [](int n_index)
	{
		return m_n_hash[n_index];
	}

	inline uint32_t operator [](int n_index) const
	{
		return m_n_hash[n_index];
	}

	void Process_Data(const uint32_t *p_data);

protected:
	inline uint32_t n_leftrot(uint32_t x, int y);
};

/*
 *	struct TSHA1
 *		- this class implements the Secure Hashing Algorithm 1 as
 *		  defined in FIPS PUB 180-1 published April 17, 1995
 *		- cooperates with CStreamHash<TSHA1, 512, true>
 */
struct TSHA1 {
protected:
	uint32_t m_n_hash[5];

public:
	TSHA1();

	TSHA1 operator =(const TSHA1 &r_t_sha1);

	bool operator ==(const TSHA1 &r_t_sha1) const;
	bool operator !=(const TSHA1 &r_t_sha1) const;
	bool operator >(const TSHA1 &r_t_sha1) const;
	bool operator <(const TSHA1 &r_t_sha1) const;
	bool operator >=(const TSHA1 &r_t_sha1) const;
	bool operator <=(const TSHA1 &r_t_sha1) const;

	inline uint32_t &operator [](int n_index)
	{
		return m_n_hash[n_index];
	}

	inline uint32_t operator [](int n_index) const
	{
		return m_n_hash[n_index];
	}

	void Process_Data(const uint32_t *p_data);

protected:
	inline uint32_t n_leftrot(uint32_t x, int y);
	inline uint32_t n_EndianConversion(uint32_t x);
};

/*
 *	struct TTinyMD5
 *		- class, implementing modified MD5 algorithm
 *		- the modification lies in reducing result to a single 32-bit integer
 *		  (purpose - to have better hashing function for strings, instead of
 *		  widely used character sum)
 *		- cooperates with CStreamHash<TTinyMD5, 64, false>
 */
struct TTinyMD5 {
protected:
	uint8_t m_n_hash[4];

public:
	TTinyMD5();

	TTinyMD5 operator =(const TTinyMD5 &r_t_md5);

	bool operator ==(const TTinyMD5 &r_t_md5) const;
	bool operator !=(const TTinyMD5 &r_t_md5) const;
	bool operator >(const TTinyMD5 &r_t_md5) const;
	bool operator <(const TTinyMD5 &r_t_md5) const;
	bool operator >=(const TTinyMD5 &r_t_md5) const;
	bool operator <=(const TTinyMD5 &r_t_md5) const;

	inline uint8_t &operator [](int n_index)
	{
		return m_n_hash[n_index];
	}

	inline uint8_t operator [](int n_index) const
	{
		return m_n_hash[n_index];
	}

	inline operator uint32_t() const
	{
		return m_n_hash[0] | (m_n_hash[1] << 8) | (m_n_hash[2] << 16) | (m_n_hash[3] << 24);
	}

	void Process_Data(const uint32_t *p_data);

protected:
	inline uint8_t n_leftrot(uint8_t x, int y);
};

/*
 *	template <class THash, const int n_length_bits, const bool b_length_big_endian>
 *		- THash is class with public member function
 *			void Process_Data(const uint32_t *p_data);
 *		- n_length_bits is number of bits (divisible by 32 and greater or equal 64!)
 *		  that will be passed to Process_Data()
 *		- b_length_big_endian enables conversion of length to big endian (source is little endian)
 */
template <class THash, const int n_length_bits, const bool b_length_big_endian>
class CStreamHash {
protected:
	uint32_t m_n_length[2]; // lo, hi

	uint8_t m_n_prev_data[n_length_bits / 8];
	int m_n_prev_length; // % n_length_bits / 8

	THash m_t_hash;

	bool m_b_processing;

public:
	/*
	 *	CStreamHash::CStreamHash()
	 *		- default constructor
	 *		- inits hash and enables data processing
	 */
	CStreamHash()
		:m_n_prev_length(0), m_t_hash(), m_b_processing(true)
	{
		m_n_length[0] = 0;
		m_n_length[1] = 0;
	}

	/*
	 *	void CStreamHash::Reset()
	 *		- reset hash and reenables new data processing
	 */
	void Reset()
	{
		m_t_hash = THash();
		m_b_processing = true;
	}

	/*
	 *	bool CStreamHash::Process_Data(const void *p_data, int n_length)
	 *		- process data p_data with length n_length bytes
	 *		- return true if processing is enabled (i.e. t_Hash() hasn't still
	 *		  called yet or Reset() has been called already) otherwise false
	 */
	bool Process_Data(const void *p_data, int n_length)
	{
		const uint8_t *p_byte_data = (const uint8_t*)p_data;

		if(!m_b_processing)
			return false;

		if(m_n_length[0] > 0xfffffffful - n_length)
			m_n_length[1] ++;
		m_n_length[0] += n_length;

		if(m_n_prev_length) {
			if(m_n_prev_length + n_length >= n_length_bits / 8) {
				memcpy(&m_n_prev_data[m_n_prev_length], p_byte_data,
					(n_length_bits / 8 - m_n_prev_length) * sizeof(uint8_t));
				p_byte_data += n_length_bits / 8 - m_n_prev_length;
				n_length -= n_length_bits / 8 - m_n_prev_length;
				m_n_prev_length = n_length_bits / 8;
				// there is enough data to form 512 bit chunk
			} else {
				memcpy(&m_n_prev_data[m_n_prev_length], p_byte_data, n_length * sizeof(uint8_t));
				m_n_prev_length += n_length;
				// just add what we got

				return true;
			}
		}
		// if there wasnt enough data last time to form 512 bit chunk, padd it

		for(const uint32_t *p_data_ptr = (const uint32_t*)p_byte_data - m_n_prev_length / 4,
		   *p_in_data = (const uint32_t*)((m_n_prev_length)? m_n_prev_data : p_byte_data),
		   *p_end = (const uint32_t*)p_byte_data + n_length / (n_length_bits / 8) * ((n_length_bits / 8) / 4);
		   p_data_ptr < p_end; p_in_data = (p_data_ptr += n_length_bits / 32)) {
			m_t_hash.Process_Data(p_in_data);
		}
		// main processing loop

		if(m_n_prev_length = n_length % (n_length_bits / 8)) {
			memcpy(m_n_prev_data, (uint32_t*)(p_byte_data + n_length - m_n_prev_length),
				m_n_prev_length * sizeof(uint8_t));
		}
		// store any data we didn't use

		return true;
	}

	/*
	 *	THash CStreamHash::t_Result()
	 *		- sends last piece of data padded with a sinble 1 bit, 512 - 1 - 64 - len % 512 zeroes
	 *		  and 64bit integer containing length to processing (len is length of all data received
	 *		  since last Reset())
	 *		- disables data processing
	 *		- returns value of hash
	 */
	THash t_Result()
	{
		if(m_b_processing) {
			uint8_t n_tmp = 0x80;
			memcpy(&m_n_prev_data[m_n_prev_length], &n_tmp, sizeof(uint8_t));
			m_n_prev_length ++;
			// add padding 1

			if(m_n_prev_length > n_length_bits / 8 - 8) {
				int n_zero_num = n_length_bits / 8 - m_n_prev_length;
				memset(&m_n_prev_data[m_n_prev_length], 0, n_zero_num * sizeof(uint8_t));
				m_n_prev_length = n_length_bits / 8;
				// padd with zeros

				Process_Data(m_n_prev_data, 0);
			}
			// n_length_bits / 8 bit integer with length wouldn't fit there, padd with zeros and process

			m_n_length[1] <<= 3;
			m_n_length[1] |= m_n_length[0] >> 29;
			m_n_length[0] <<= 3;
			// length is in bits -> multiply by 8

			if(b_length_big_endian) {
				int n_tmp = m_n_length[0];
				m_n_length[0] = n_EndianConversion(m_n_length[1]);
				m_n_length[1] = n_EndianConversion(n_tmp);
			}
			// convert length endianness

			int n_zero_num = n_length_bits / 8 - 8 - m_n_prev_length;
			memset(&m_n_prev_data[m_n_prev_length], 0, n_zero_num * sizeof(uint8_t));
			m_n_prev_length += n_zero_num;
			memcpy(&m_n_prev_data[m_n_prev_length], m_n_length, sizeof(m_n_length));
			m_n_prev_length = n_length_bits / 8;
			// append more zeros and length

			Process_Data(m_n_prev_data, 0);

			m_b_processing = false;
		}
		// process padding

		return m_t_hash;
	}

protected:
	/*
	 *	uint32_t CStreamHash::n_EndianConversion(uint32_t x)
	 *		- endian conversion
	 */
	uint32_t n_EndianConversion(uint32_t x)
	{
		return (x >> 24) | ((x >> 8) & 0xff00) | ((x & 0xff00) << 8) | (x << 24);
	}
};

#endif //__HASH_INCLUDED
