/*
								+----------------------------------+
								|                                  |
								|     ***  STL utilities  ***      |
								|                                  |
								|   Copyright  -tHE SWINe- 2008   |
								|                                  |
								|           StlUtils.cpp           |
								|                                  |
								+----------------------------------+
*/

/*
 *
 *	2008-12-22
 *
 *	added typename keyword in Reserve_NMore() so it now compiles with g++
 *
 *	2009-01-13
 *
 *	added Format() function and StlUtils.cpp to keep it's body (it's not template)
 *
 *	2009-03-24
 *
 *	added Assign and Append, accepting std::string as second argument,
 *	should allow assigning and appending strings with different character
 *	types as well (no conversion or error handling though).
 *
 *	(there was just const char* before; it was possible to use these versions
 *	to append std::string to std::string using c_str() function, but it
 *	would expectably yield worse performance)
 *
 *	2009-05-04
 *
 *	stl_ut::Swap() was removed in favor of std::swap
 *
 *	added __STL_UTILS_ENABLE_EXCEPTIONS macro, controlling between exception-
 *	based and null-pointer based errors on operator new()
 *
 *	written documentation comments
 *
 */

#include "CallStack.h"
#if defined(WIN32) || defined (_WIN64)
#include <windows.h>
#endif
#include <stdarg.h>
#include <stdio.h>
#include <vector>
#include <algorithm>
#include "StlUtils.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

/*
 *								=== STL Utilities ===
 */

namespace stl_ut {

/*
 *	bool stl_ut::Format(std::string &r_s_result, const char *p_s_fmt, ...)
 *		- equivalent of standard "c" library sprintf function,
 *		  working with std::string output
 *		- returns true on success, false on failure (not enough memory)
 *		- note this is safe against output buffer overrun, but is
 *		  still susceptible to ill-formated format string (p_s_fmt)
 *		  and bad arguments (not enough of them / not corresponding
 *		  to tags in the format string). but as the first one being run-time
 *		  error and the latter one compile-time error, this brings
 *		  enought improvement.
 */
bool Format(std::string &r_s_result, const char *p_s_fmt, ...)
{
	va_list ap;

	int n_size = strlen(p_s_fmt) + 1;
	if(!Reserve_N(r_s_result, n_size))
		return false;
	r_s_result.resize(n_size);
	// alloc str as long as formatting str (should be enough in some cases)

	for(;;) {
		va_start(ap, p_s_fmt);
#if defined(_MSC_VER) && !defined(__MWERKS__) // msvc
		int n = _vsnprintf(&r_s_result[0], n_size - 1, p_s_fmt, ap);
		// maximum characters to write, not the size of buffer
#else
		int n = vsnprintf(&r_s_result[0], n_size, p_s_fmt, ap);
#endif
		va_end(ap);
		// try to sprintf

		if(n > -1 && n < n_size) {
#if defined(_MSC_VER) && !defined(__MWERKS__) // msvc
			r_s_result.resize(n);
#else
			r_s_result.resize(strlen(r_s_result.c_str()));
#endif
			_ASSERTE(r_s_result.length() == strlen(r_s_result.c_str()));
			return true;
		}
		// see if we made it

		if(n > -1) // glibc 2.1
			n_size = n + 1; // precisely what is needed
		else // glibc 2.0, msvc
			n_size *= 2;

		if(!Reserve_N(r_s_result, n_size))
			return false;
		r_s_result.resize(n_size);
		// realloc to try again
	}
}

}

/*
 *								=== ~STL Utilities ===
 */
