/*
								+-----------------------------------+
								|                                   |
								|  ***  LRD wald-boost engine  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2008   |
								|                                   |
								|            WaldBoost.h            |
								|                                   |
								+-----------------------------------+
*/

#ifndef __LRD_WALD_BOOST_ENGINE_INCLUDED
#define __LRD_WALD_BOOST_ENGINE_INCLUDED

#define WBOOST_LAST_OCCLUSION_QUERY_PASS 4
#define WBOOST_SUBDIVIDE_PASS_NUM 0
#define WBOOST_SUBDIVIDE_PASS_SIZE 20
//#define WBOOST_USE_TRISTRIPS
#define WBOOST_DISABLE_OCCLUSION_QUERIES
//#define WBOOST_UNBIND_FBO_EVERY_WBOOST_PASS
#define DISABLE_TEXTURE_RECT_NV_WARNING
#define WBOOST_TILE_RENDERER_USE_RGB8

#include "./UberLame_src/BitArray.h"
#include "./UberLame_src/StlUtils.h"
#include "Load_XML.h"
#include "Convolve.h"

#include "AsciiHist.h" // debug

/**
 *	class CAlphaTextureAtlas
 *		- simple texture atlas builder for classifier alpha tables
 */
class CAlphaTextureAtlas {
protected:
	int n_alpha_num; // number of alphas per classifier
	struct TPage {
		int n_first_classifier; // index of first classifier's alpha on this page
		int n_classifier_num; // number of classifiers on this page
		CGLTexture_Rect *p_texture; // texture
	};
	std::vector<TPage> texture_list; // might take up more pages

public:
	/**
	 *	CAlphaTextureAtlas::CAlphaTextureAtlas(CGLState *p_state,
	 *		const TWaldBoostClassifier &r_t_classifier,
	 *		int n_max_texture_size = 512, bool b_allow_more_columns = false)
	 *		- default constructor; creates texture atlas for classifier r_t_classifier
	 *		  (ie. allocates row of pixels to store alpha table for each classifier
	 *		  and creates those textures (uses GL_FLOAT_R16_NV or GL_RGB16F_ARB format))
	 *		- n_max_texture_size is maximal texture size
	 *		- setting b_allow_more_columns enables storing alphas in more than one column
	 *		  (makes it possible to create wider textures, containing more alphas,
	 *		  but requires x-coordinate offset for alpha texture fetch)
	 *		- it's recommended to call b_Status() afterwards to check constructor success
	 */
	CAlphaTextureAtlas(CGLState *p_state, const TWaldBoostClassifier &r_t_classifier,
		int n_max_texture_size = 512, bool b_allow_more_columns = false);

	/**
	 *	CAlphaTextureAtlas::~CAlphaTextureAtlas()
	 *		- destructor
	 */
	~CAlphaTextureAtlas();

	/**
	 *	bool CAlphaTextureAtlas::b_Status() const
	 *		- returns true if constructor succeeded and the atlas is ready to use,
	 *		  otherwise returns false
	 */
	bool b_Status() const;

	/**
	 *	int CAlphaTextureAtlas::n_Texture(int n_classifier_index, int &r_n_x, int &r_n_y) const
	 *		- returns index of texture page (not OpenGL id) containing alphas of
	 *		  weak classifier n_classifier_index
	 *		- r_n_x and r_n_y are set to offset to the first pixel of alpha table
	 *		  (table is oriented horizontally and next alphas are adressed by
	 *		  increasing x-offset)
	 *		- returns -1 in case n_classifier_index is greater or equal to number of
	 *		  weak classifiers contained in classifier, passed to constructor
	 */
	int n_Texture(int n_classifier_index, int &r_n_x, int &r_n_y) const;

	/**
	 *	CGLTexture_Rect *CAlphaTextureAtlas::p_Texture(int n_index)
	 *		- returns texture with index n_index (zero-based texture page index,
	 *		  returned by n_Texture())
	 */
	CGLTexture_Rect *p_Texture(int n_index);

	/**
	 *	void CAlphaTextureAtlas::Dump() const
	 *		- prints information about texture atlas to stdout (debug)
	 */
	void Dump() const;

protected:
	static inline void DeletePageTexture(TPage &r_t_page);
};

/**
 *	class CTileRenderer
 *		- tile-based renderer for wald-boost
 *		- creates vertex arrays, containing tile geometry,
 *		  manages drawing tiles and occlusion queries, as well
 *		  as downloading active regions and finding detections (on CPU)
 */
class CTileRenderer {
public:
	/**
	 *	struct CTileRenderer::TDetection
	 *		- position of object detection
	 *		- n_scale is index in image pyramid (0 = original image, 1 = first downscale, ...)
	 */
	struct TDetection {
		int x, y;
		int n_scale;
		float f_response;
	};

protected:
	struct TTileData {
		Vector2i v_pos, v_size;
		const TImage *p_image;
	};

protected:
	int m_n_tile_num; // number of subdivision tiles
#ifndef WBOOST_DISABLE_OCCLUSION_QUERIES
	GLuint *m_p_occlusion_query; // occlusion queries, one per tile
#endif //WBOOST_DISABLE_OCCLUSION_QUERIES

	CGLTexture_Rect *m_p_texture[2]; // texture, pong
	CGLFrameBuffer *m_p_fbo; // framebuffer
	CGLVertexBufferObject m_vbo;

	CBitArray m_tile_flags;

	int m_n_iteration;
	int m_n_active_tile_num;

	bool m_b_status;

	float *m_p_tmp_buffer;
	TTileData *m_p_tile_data;
	int m_n_downscale;

	bool m_b_histogram_triggered;
	CHistogram m_t_hist;

public:
	/**
	 *	CTileRenderer::CTileRenderer(CGLState *p_state, const TImagePage &r_t_page,
	 *		int n_pass_num, int n_tile_size = 32, int n_downscale = 1)
	 *		- default constructor; creates tile renderer
	 *		- p_state is OpenGL state guard, r_t_page is downscaled image pyramid page,
	 *		  n_pass_num is number of rendering passes for drawing, n_tile_size is
	 *		  size of tile in pixels and n_downscale is downscale factor of detection
	 *		  raster (recommended using powers of two only)
	 *		- it's recommended to check b_Status() to see constructor outcome
	 */
	CTileRenderer(CGLState *p_state, const TImagePage &r_t_page, int n_pass_num,
		int n_tile_size = 32, int n_downscale = 1);

	/**
	 *	CTileRenderer::~CTileRenderer()
	 *		- destructor
	 */
	~CTileRenderer();

	/**
	 *	bool CTileRenderer::b_Status() const
	 *		- returns true if constructor succeeded, otherwise returns false
	 */
	bool b_Status() const;

	/**
	 *	void CTileRenderer::Trigger_Histogram()
	 *		- debug function
	 *		- triggers tile execution histogram (it will be printed after the next pass)
	 */
	void Trigger_Histogram();

	/**
	 *	bool CTileRenderer::Download_Detections(CGLState *p_state,
	 *		std::vector<TDetection> &r_detection_list, float f_thresh)
	 *		- downloads detections from FBO
	 *		- p_state is OpenGL state guard, r_detection_list (output) is list of
	 *		  detected objects and f_thresh is detection threshold
	 *		- returns true on success, false on failure
	 */
	bool Download_Detections(CGLState *p_state,
		std::vector<TDetection> &r_detection_list, float f_thresh);

	/**
	 *	int CTileRenderer::n_Tile_Num() const
	 *		- returns number of rendering tiles
	 */
	int n_Tile_Num() const;

	/**
	 *	int CTileRenderer::n_Active_Tile_Num() const
	 *		- returns number of active rendering tiles (such tiles where occlusion
	 *		  query returned nonzero in the last pass, or all tiles before first pass)
	 */
	int n_Active_Tile_Num() const;

	/**
	 *	bool CTileRenderer::Render_Begin(CGLState *p_state)
	 *		- begins rendering (sets up FBO, ...)
	 *		- p_state is OpenGL state guard
	 *		- returns true on success, false on failure
	 */
	bool Render_Begin(CGLState *p_state);

	/**
	 *	bool CTileRenderer::Render_Iteration(CGLState *p_state)
	 *		- renders one iteration (draws active tiles with occlusion query
	 *		  while reading results of occlusion query from the previous pass
	 *		  (except in the first pass))
	 *		- returns true on success, false on failure
	 */
	bool Render_Iteration(CGLState *p_state);

	/**
	 *	bool CTileRenderer::Render_Finish(CGLState *p_state)
	 *		- finalizes rendering (release FBO, ...)
	 *		- returns true on success, false on failure
	 */
	bool Render_Finish(CGLState *p_state);

	/**
	 *	CGLTexture_Rect *CTileRenderer::p_Result_Texture()
	 *		- returns texture, containing rendering result
	 */
	CGLTexture_Rect *p_Result_Texture();

protected:
	CGLTexture_Rect *p_AllocTexture(CGLState *p_state, int n_width, int n_height);
};

/**
 *	class CWaldBoostShaderGenerator
 *		- class, generating and managing wald-boost shaders
 *		- cooperates with CTileRenderer
 */
class CWaldBoostShaderGenerator {
public:
	enum {
		max_TextureUnits = 8
	};

protected:
	struct TPass {
		//CGLTexture_Rect *p_pong_texture; // CTileRenderer binds it
		CGLTexture_Rect *p_alpha_texture;
		int n_tex_num;
		int p_tex_binding[max_TextureUnits - 2]; // source image (-1) / convolution texture indices
		CGLShaderObject *p_shader;
	};
	std::vector<TPass> m_pass_list;
	int m_n_texture_unit_num;
	bool m_b_status;

public:
	/**
	 *	CWaldBoostShaderGenerator::CWaldBoostShaderGenerator(CGLState *p_state,
	 *		CImageConvolver *p_convolver, CAlphaTextureAtlas *p_alpha_tex_atlas,
	 *		const TWaldBoostClassifier &r_t_classifier, int n_downscale = 1,
	 *		int n_max_shaders_pass = 20)
	 *		- default constructor; creates data textures and assembles + compiles shader
	 *		- p_state is OpenGL state guard, p_convolver is convolution manager,
	 *		  p_alpha_tex_atlas is alpha texture atlas, r_t_classifier contains classifier
	 *		  information, n_downscale is detection raster downscale factor (factors greater
	 *		  than 1 yield less precise, but faster detections) and n_max_shaders_pass
	 *		  is limit of weak classifiers in a single pass (limited by maximal shader length)
	 *		- note this allocates memory and creates OpenGL objects. it's recommended to
	 *		  call b_Status() afterwards
	 */
	CWaldBoostShaderGenerator(CGLState *p_state,
		CImageConvolver *p_convolver, CAlphaTextureAtlas *p_alpha_tex_atlas,
		const TWaldBoostClassifier &r_t_classifier, int n_downscale = 1,
		int n_max_shaders_pass = 20);

	/**
	 *	bool CWaldBoostShaderGenerator::b_Status() const
	 *		- returns true if constructor succeeded and shader generator
	 *		  is ready for use, otherwise returns false
	 */
	bool b_Status() const;

	/**
	 *	int CWaldBoostShaderGenerator::n_Pass_Num() const
	 *		- returns number of passes (a lot)
	 */
	int n_Pass_Num() const;

	/**
	 *	bool CWaldBoostShaderGenerator::Render(CGLState *p_state,
	 *		CGLTexture_Rect *p_base_texture, CImageConvolver *p_convolver,
	 *		CTileRenderer *p_renderer)
	 *		- renders a single frame (detects faces in a single frame)
	 *		- p_state is OpenGL state guard, p_base_texture is base texture
	 *		  (source image, or downscaled image pyramid), p_convolver is image
	 *		  convolver (unused) and p_renderer is tile renderer
	 *		- returns true on success, false on failure
	 */
	bool Render(CGLState *p_state, CGLTexture_Rect *p_base_texture,
		CImageConvolver *p_convolver, CTileRenderer *p_renderer);

protected:
	bool GenerateShaderSection(std::string &r_s_body, const TWeakClassifier &r_t_cl,
		int n_conv_texture_slot, Vector2f v_texture_off, int n_alpha_x, int n_alpha_y);
	const char *p_s_ConvSampler_Name(int n_conv_texture_slot);
	inline bool Append(std::string &r_s_str, const char *p_s_string);
	Vector2f v_GetTextureOffset(int n_block_width, int n_block_height, CImageConvolver *p_convolver);
	int n_GetTexture(int n_block_width, int n_block_height, CImageConvolver *p_convolver);
};

/**
 *	class CWaldBoost_AllInOne_ShaderGenerator
 *		- shader generator for all weak classifiers in a single pas
 *		- cooperates with CTileRenderer
 */
class CWaldBoost_AllInOne_ShaderGenerator {
protected:
	CGLTexture_Rect *p_alpha_texture, *p_data_texture;
	CGLShaderObject *p_shader;
	bool m_b_status;

public:
	/**
	 *	CWaldBoost_AllInOne_ShaderGenerator::CWaldBoost_AllInOne_ShaderGenerator(CGLState *p_state,
	 *		CImageConvolver *p_convolver, CAlphaTextureAtlas *p_alpha_tex_atlas,
	 *		const TWaldBoostClassifier &r_t_classifier)
	 *		- default constructor; creates data textures and assembles + compiles shader
	 *		- p_state is OpenGL state guard, p_convolver is convolution unit (it isn't
	 *		  really required here, it is only intended to have interface similar to
	 *		  CWaldBoostShaderGenerator) and p_alpha_tex_atlas is alpha texture atlas,
	 *		  r_t_classifier is classifier
	 *		- note there are some restrictions, all convolutions must be calculated using
	 *		  bilinear filtering (ie. 1x1, 1x2, 2x1 or 2x2 only) and all alphas must be
	 *		  in a single texture page, in a single column (both restrictions could be removed
	 *		  if necessary, the aim was to keep the code as simple and fast as possible)
	 *		- note this allocates memory and creates OpenGL objects. it's recommended to
	 *		  call b_Status() afterwards
	 */
	CWaldBoost_AllInOne_ShaderGenerator(CGLState *p_state,
		CImageConvolver *p_convolver, CAlphaTextureAtlas *p_alpha_tex_atlas,
		const TWaldBoostClassifier &r_t_classifier);

	/**
	 *	CWaldBoost_AllInOne_ShaderGenerator::~CWaldBoost_AllInOne_ShaderGenerator()
	 *		- destructor
	 */
	~CWaldBoost_AllInOne_ShaderGenerator();

	/**
	 *	bool CWaldBoost_AllInOne_ShaderGenerator::b_Status() const
	 *		- returns true if constructor succeeded and shader generator
	 *		  is ready for use, otherwise returns false
	 */
	bool b_Status() const;

	/**
	 *	int CWaldBoost_AllInOne_ShaderGenerator::n_Pass_Num() const
	 *		- returns number of passes (1)
	 */
	int n_Pass_Num() const;

	/**
	 *	void CWaldBoost_AllInOne_ShaderGenerator::Set_FinalThreshold(float f_thresh)
	 *		- sets final threshold to f_thresh (default 0.0)
	 */
	void Set_FinalThreshold(float f_thresh);

	/**
	 *	bool CWaldBoost_AllInOne_ShaderGenerator::Render(CGLState *p_state,
	 *		CGLTexture_Rect *p_base_texture, CImageConvolver *p_convolver,
	 *		CTileRenderer *p_renderer)
	 *		- renders a single frame (detects faces in a single frame)
	 *		- p_state is OpenGL state guard, p_base_texture is base texture
	 *		  (source image, or downscaled image pyramid), p_convolver is image
	 *		  convolver (unused) and p_renderer is tile renderer
	 *		- returns true on success, false on failure
	 *		- note this is only below 20 OpenGL commands long
	 *		  (24 including the rest to display the results)
	 */
	bool Render(CGLState *p_state, CGLTexture_Rect *p_base_texture,
		CImageConvolver *p_convolver, CTileRenderer *p_renderer);

protected:
	static Vector2f v_GetTextureOffset(Vector2i v_texcoord_step, CImageConvolver *p_convolver);
	static int n_Transpose3x3Index(int i);
	static Vector4f lessThan(Vector4f v, float f); // glsl lookalike
	static Vector4f greaterThanEqual(Vector4f v, float f); // glsl lookalike
	static void test_AB_multiplexing(); // this is more of a programming check, not runtime check
};

#endif //__LRD_WALD_BOOST_ENGINE_INCLUDED
