//-----------------------------------------------------------------------------
// librot.cpp
// Generic image rotation plugin
// Ji Havel, DCGM FIT BUT, Brno
// $Id: librot.cpp 161 2009-08-28 09:06:15Z ihavel $
//-----------------------------------------------------------------------------

#include <abon.h>

#include <cxcore.h>
#include <cv.h>

#include <list>

using namespace std;

#define LIBNAME "librot"
#define LIBVERSION "$Id: librot.cpp 161 2009-08-28 09:06:15Z ihavel $"

struct RotDetector : public abonDetector
{
    //following detector
    abonDetector *detector;

    //parameters
    float rotationStart, rotationStep;
    unsigned rotationCount;

    //image source
    abonCbGetImage *ioGet;
    abonCbPutImage *ioPut;
    void *ioParam;

    //processed image
    IplImage *imageHeader;
    abonImageData *imageData;
    unsigned rotation;

    //translation and size of output image
    float tx, ty;
    unsigned sx, sy;

    //image list for merging
    struct ParentImage
    {
        abonImageData *image;
        unsigned count;
        ParentImage(abonImageData *i, unsigned c) throw() : image(i), count(c) {}
    };
    typedef list<ParentImage> ImageList;
    ImageList imageList;

    //child image
    struct Image : public abonImageData
    {
        IplImage *image;
        ParentImage *parentImage;
        float angle;

        Image(unsigned w, unsigned h, unsigned dc) throw(abonErrorCode)
        {
            if((image = cvCreateImage(cvSize(w, h), IPL_DEPTH_8U, 1)) == NULL)
            {
                throw aboneMemory;
            }
            if((detections = new(nothrow) abonDetection[dc]) == NULL)
            {
                cvReleaseImage(&image);
                throw aboneMemory;
            }
            width = w;
            height = h;
            pitch = image->widthStep;
            data = image->imageData;
            classifiers = 0;
            weakClassifiers = 0;
            detectionCount = 0;
            detectionsEnd = detections + dc;
        }

        ~Image() throw ()
        {
            cvReleaseImage(&image);
            delete [] detections;
        }
    };
    
    bool getNewImage();

    RotDetector() throw();
    ~RotDetector() throw();

    static abonImageData *rotGet(void *param) throw();
    static abonErrorCode rotPut(void *param, abonImageData *image) throw();
};

extern "C" { // Plugin interface --------------------------------------------------

abonDetector *abonCreate()
{
    printf("["LIBNAME"]: "LIBVERSION"\n");
    return new(nothrow) RotDetector;
}

abonErrorCode rotSetSize(abonDetector *det, unsigned width, unsigned height, unsigned pitch)
{
    RotDetector *detector = (RotDetector*)det;
    if(detector == NULL) return aboneError;
    if(detector->verbosity) printf("["LIBNAME"]: Setting size...\n");

    assert(width <= pitch);

    detector->width = width;
    detector->height = height;
    detector->pitch = pitch;

    //free image header if size doesn't match
    if(
        (detector->imageHeader != NULL) &&
        ((detector->imageHeader->width != width) || (detector->imageHeader->height != height))
        )
    {
        cvReleaseImageHeader(&detector->imageHeader);
        detector->imageHeader = NULL;
    }

    //allocate new image header
    if(
        (detector->imageHeader == NULL) &&
        ((detector->imageHeader = cvCreateImageHeader(cvSize(width, height), IPL_DEPTH_8U, 1)) == NULL)
        )
    {
        return aboneMemory;
    }

    detector->imageHeader->widthStep = pitch;

    float nx = 0, px = width, ny = 0, py = height;

    //calculate rotated image size
    for(unsigned i = 0; i < detector->rotationCount; ++i)
    {
        float angle = detector->rotationStart + detector->rotationStep*i;

        float rotCoefsPos[6];
        CvMat rotMatPos = cvMat(2, 3, CV_32FC1, rotCoefsPos);
        cv2DRotationMatrix(cvPoint2D32f(width/2,height/2), angle, 1.0, &rotMatPos);

        float corners[4][3] = {
            { 0    , 0     , 1 },
            { width, 0     , 1 },
            { width, height, 1 },
            { 0    , height, 1 }
        };

        for(unsigned j = 0; j < 4; ++j)
        {
            float corners_out[3];
            CvMat ctr = cvMat(3, 1, CV_32F, corners[j]);
            CvMat ctr1 = cvMat(2, 1, CV_32F, corners_out);
            cvMatMul(&rotMatPos, &ctr, &ctr1);

            nx = min(nx, corners_out[0]);
            px = max(px, corners_out[0]);
            ny = min(ny, corners_out[1]);
            py = max(py, corners_out[1]);
        };
    }

    detector->tx = -nx;
    detector->ty = -ny;
    detector->sx = px - nx;
    detector->sy = py - ny;

    IplImage tmp;
    if(cvInitImageHeader(&tmp, cvSize(detector->sx, detector->sy), IPL_DEPTH_8U, 1) == NULL)
    {
        return aboneError;
    }

    if(detector->detector != NULL)
    {
        return abonSetSize(detector->detector, tmp.width, tmp.height, tmp.widthStep);
    }
    else
    {
        return aboneOk;
    }
}

abonErrorCode rotSetRotation(abonDetector *detector, float start, float step, unsigned count)
{
    if(detector == NULL) return aboneError;
    if(detector->verbosity) printf("["LIBNAME"]: Setting rotation...\n");

    assert(count > 0);

    ((RotDetector*)detector)->rotationStart = start;
    ((RotDetector*)detector)->rotationStep = step;
    ((RotDetector*)detector)->rotationCount = count;

    return aboneOk;
}

abonErrorCode rotAttach(abonDetector *detector, abonDetector *attachment)
{
    if(detector == NULL) return aboneError;
    if(detector->verbosity) printf("["LIBNAME"]: Attaching...\n");

    ((RotDetector*)detector)->detector = attachment;

    return aboneOk;
}

abonErrorCode rotStart(abonDetector *det, abonCbGetImage *get, abonCbPutImage *put, void *param)
{
    try {
        RotDetector *detector = (RotDetector*)det;
        if(detector == NULL) return aboneError;

        assert(get != NULL);
        assert(put != NULL);
        assert(detector->detector != NULL);
        assert(detector->imageHeader != NULL);

        if(detector->verbosity) printf("["LIBNAME"]: Starting...\n");

        //init image input
        detector->ioGet = get;
        detector->ioPut = put;
        detector->ioParam = param;
        if(!detector->getNewImage())
        {
            return aboneOk;
        }
                    
        return abonStart(detector->detector, RotDetector::rotGet, RotDetector::rotPut, detector);
    } catch(bad_alloc & ex) {
        return aboneMemory;
    } catch(abonErrorCode ec) {
        return ec;
    } catch(...) {
        return aboneError;
    }
}

abonErrorCode rotRelease(abonDetector *detector)
{
    if(detector == NULL) return aboneOk;
    if(detector->verbosity) printf("["LIBNAME"]: Releasing...\n");

    delete (RotDetector*)detector;

    return aboneOk;
}
    
} // Plugin interface --------------------------------------------------

bool RotDetector::getNewImage()
{
    if((imageData = ioGet(ioParam)) == NULL)
    {
        return false;
    }
    imageHeader->imageData = imageData->data;
    rotation = 0;
    imageList.push_back(ParentImage(imageData, rotationCount));
    imageData->detectionCount = 0;
    imageData->classifiers = 0;
    imageData->weakClassifiers = 0;
    return true;
}

RotDetector::RotDetector() throw()
    : abonDetector(abonDetectorInitializer)
    , detector(NULL), imageHeader(NULL), imageData(NULL)
    , rotationStart(0), rotationStep(0), rotationCount(1)
{
    abonDetector::setSize = rotSetSize;
    abonDetector::setRotation = rotSetRotation;
    abonDetector::attach = rotAttach;
    abonDetector::start = rotStart;
    abonDetector::release = rotRelease;
}

RotDetector::~RotDetector() throw()
{
    abonUnload(detector);
    if(imageHeader != NULL)
    {
        cvReleaseImageHeader(&imageHeader);
    }
}

abonImageData *RotDetector::rotGet(void *param) throw()
{
    try {
        RotDetector *detector = (RotDetector*)param;

        if((detector->rotation >= detector->rotationCount) && !detector->getNewImage())
        {
            return NULL;
        }

        float angle  = detector->rotationStart + detector->rotationStep*detector->rotation;

        //create rotated image
        Image *image = new RotDetector::Image(detector->sx, detector->sy, detector->imageData->detectionsEnd - detector->imageData->detections);
        image->imageID = detector->imageData->imageID*detector->rotationCount + detector->rotation;
        image->parentImage = &detector->imageList.back();
        image->angle = angle;

        //fill with data
        float rotCoefsPos[6];
        float transCoefsPos[9] = {1, 0, detector->tx, 0, 1, detector->ty, 0, 0, 1};
        CvMat rotMatPos = cvMat(2, 3, CV_32FC1, rotCoefsPos);
        CvMat transMatPos = cvMat(3, 3, CV_32FC1, transCoefsPos);
        cv2DRotationMatrix(cvPoint2D32f(image->width/2,image->height/2), angle, 1.0, &rotMatPos);
        cvMatMul(&rotMatPos, &transMatPos, &rotMatPos);
        cvWarpAffine(detector->imageHeader, image->image, &rotMatPos);

        ++detector->rotation;

        return image;
    } catch(...) {
        return NULL;
    }
}

abonErrorCode RotDetector::rotPut(void *param, abonImageData *img) throw()
{
    try {
        RotDetector *detector = (RotDetector*)param;
        Image *image = (Image*)img;

        --image->parentImage->count;
        abonImageData *parent = image->parentImage->image;

        //matrix for detection rotation
        float rotCoefsNeg[6];
        CvMat rotMatNeg = cvMat(2, 3, CV_32FC1, rotCoefsNeg);
        cv2DRotationMatrix(cvPoint2D32f(image->width/2,image->height/2), -image->angle, 1.0, &rotMatNeg);
        rotCoefsNeg[2] -= detector->tx;
        rotCoefsNeg[5] -= detector->ty;

        //use only detections, that fit into detection array
        image->detectionCount = min(image->detectionCount, (parent->detectionsEnd - parent->detections - parent->detectionCount));

        //rotate detections to match original image space
        for(unsigned i = 0; i < image->detectionCount; ++i)
        {
            abonDetection *d = image->detections + i;
            float _ctr[3] = {float(d->x) + d->width/2, float(d->y) + d->height/2, 1}, _ctr1[3];
            CvMat ctr = cvMat(3, 1, CV_32F, _ctr);
            CvMat ctr1 = cvMat(2, 1, CV_32F, _ctr1);
            cvMatMul(&rotMatNeg, &ctr, &ctr1);
            d->x = int(_ctr1[0]) - d->width/2;
            d->y = int(_ctr1[1]) - d->height/2;
            d->angle = image->angle;
        }

        //add detections to parent image
        copy(image->detections, image->detections + image->detectionCount, parent->detections + parent->detectionCount);
        parent->detectionCount += image->detectionCount;
        parent->classifiers += image->classifiers;
        parent->weakClassifiers += image->weakClassifiers;

        delete image;

        //return all finished images
        while(!detector->imageList.empty() && (detector->imageList.front().count == 0))
        {
            abonErrorCode ec = detector->ioPut(detector->ioParam, detector->imageList.front().image);
            if(ec != aboneOk)
            {
                return ec;
            }
            detector->imageList.pop_front();
        }

        return aboneOk;
    } catch(bad_alloc & ex) {
        return aboneMemory;
    } catch(abonErrorCode ec) {
        return ec;
    } catch(...) {
        return aboneError;
    }
}
