//-----------------------------------------------------------------------------
// libsimple.cpp
// Unoptimized detection with WaldBoost classifier
// Roman Juranek, DCGM FIT BUT, Brno
// $Id: libsimple.cpp 171 2009-09-30 10:32:35Z ijuranek $
//-----------------------------------------------------------------------------

#include <aboninterface.h>
#include <timer.h>
#include "lrd_engine.h"
#include "image.h"

#include <iostream>
#include <cstdlib>
#include <cstring>
#include <vector>

using namespace std;

#define LIBNAME "libsimple"
#define LIBVERSION "$Id: libsimple.cpp 171 2009-09-30 10:32:35Z ijuranek $"

// Detector parameters
struct SimpleDetector
{
	abonDetector header; // Header
    
    TClassifier * c;
    
    float scale;
    unsigned   levels;

    int scanStepH, scanStepV;

    // Dynamically created stuff
    vector<TImage*> * pyramid;
};


/// Release pyramid structure in the detector.
/// The detector->pyramid will be 0 as a consequence.
/// \param detector The detector
static void releasePyramid(SimpleDetector * detector)
{
    if (!detector || !detector->pyramid) return;
    
    if (((abonDetector*)detector)->verbosity)
    {
        printf("["LIBNAME"]: Releasing pyramid\t");
    }

    vector<TImage*> & pyr = *detector->pyramid;

    for (size_t i = 0; i < pyr.size(); ++i)
    {
        releaseImage(&pyr[i]);
    }

    delete detector->pyramid;
    detector->pyramid = 0;

    if (((abonDetector*)detector)->verbosity)
    {
        printf("OK\n");
    }

    return;
}

/// Build pyramidal structure.
/// Create images according to parameters set in the detector.
/// \param detector The detector
/// \param width Width of input image
/// \param height Height of input image
/// \param pitch Alignment of image rows
/// \returns Ptr to vector of IplImage*.
static vector<TImage*> * createPyramid(SimpleDetector * detector, unsigned width, unsigned height)
{
    if (!detector) return 0;
    SimpleDetector * det = (SimpleDetector*) detector;

    if (((abonDetector*)detector)->verbosity)
    {
        printf("["LIBNAME"]: Building pyramid\t");
    }

    TClassifier * classifier = det->c;

    vector<TImage*> * pyr = new vector<TImage*>();

    // Prepare image pyramid
    float levelWidth = width;
    float levelHeight = height;
    
    while (levelWidth > classifier->width && levelHeight > classifier->height && pyr->size() < det->levels)
    {
        // Calculate scale
        int w = int(levelWidth);
        int h = int(levelHeight);

        //float scale = float(width) / w;
        //if (scale > det->maxScale) break;
        
        TImage * tmp = createImage(w, h);
        pyr->push_back(tmp);
        
        levelWidth *= det->scale;
        levelHeight *= det->scale;
    }

    if (((abonDetector*)detector)->verbosity)
    {
        printf("OK (%d levels)\n", pyr->size());
    }

    return pyr;
}


extern "C" {
abonFunSetVerbosity simpleSetVerposity;
abonFunSetTimer simpleSetTimer;
abonFunSetClassifier simpleSetClassifier;
abonFunSetStep simpleSetStep;
abonFunSetSize simpleSetSize;
abonFunSetScale simpleSetScale;
abonFunStart simpleStart;
abonFunRelease simpleRelease;
abonFunTimestampDescription simpleTimestampDescription;
};


abonErrorCode simpleSetVerbosity(abonDetector *detector, unsigned verbosity)
{
    if (!detector) return aboneOk;
    detector->verbosity = verbosity;
	return aboneOk;
}


abonErrorCode simpleSetTimer(abonDetector *detector, Timer * timer)
{
    if (!detector) return aboneOk;
    detector->timer = timer;
	return aboneOk;
}


abonErrorCode simpleSetClassifier(abonDetector * detector, TClassifier * classifier)
{
    if (!detector) return aboneOk;
    if (classifier)
    {
        if (classifier->tp != LRD && classifier->tp != LRP)
        {
            printf("["LIBNAME"]: ERROR: Cannot evaluate non-LRD classifiers\n");
            return aboneError;
        }
        SimpleDetector * det = (SimpleDetector*) detector;
        det->c = classifier;
    }
	return aboneOk;
}


abonErrorCode simpleSetStep(abonDetector *detector, unsigned hstep, unsigned vstep)
{
    if (!detector) return aboneOk;
    else
    {
        SimpleDetector * det = (SimpleDetector*) detector;
        det->scanStepH = hstep;
        det->scanStepV = vstep;
        return aboneOk;
    }
}


abonErrorCode simpleSetSize(abonDetector * detector, unsigned width, unsigned height, unsigned pitch)
{
    if (!detector) return aboneOk;
    SimpleDetector * det = (SimpleDetector*) detector;

    if (detector->verbosity)
    {
        printf("["LIBNAME"]: Setting size to %dx%dpx with %dB row alignment\n", width, height, pitch);
    }

    releasePyramid(det);
    det->pyramid = createPyramid(det, width, height);
    
    if (!det->pyramid) return aboneError;

	return aboneOk;
}

// TODO
// - jaka je semantika tech parametru?
// - o 'start' by se mel postarat prescale (?)
abonErrorCode simpleSetScale(abonDetector *detector, float start, float step, unsigned count)
{
    if (!detector) return aboneOk;
    SimpleDetector * det = (SimpleDetector*) detector;

    det->scale = step; // XXX predpokladam, ze step je scale factor!!!
    det->levels = count;

    return aboneOk;
}


abonErrorCode simpleStart(abonDetector * detector, abonCbGetImage *get, abonCbPutImage *put, void *param)
{
    if (!detector) return aboneError;

    SimpleDetector * det = (SimpleDetector*) detector;

    // get the classifier
    TClassifier * classifier = det->c;
    if (!classifier)
    {
        fprintf(stderr, "["LIBNAME"]: ERROR: No classifier speified!\n");
        return aboneError;
    }

    if (!det->pyramid) // or assert()? This would be fail of framework not an user
    {
        fprintf(stderr, "["LIBNAME"]: ERROR: Call 'setSize()' first!\n");
        return aboneError;
    }
    
    vector<TImage*> & pyramid = *det->pyramid;

	if (detector->verbosity)
    {
        printf("["LIBNAME"]: Starting detection\n");
    }

    // Image loop
    abonImageData * image = 0;
    while((image = get(param)) != NULL)
	{
        // Convert the image to IPL (only header is created!)
        TImage * cvImage = createImageHeader(image->width, image->height, image->pitch, image->data);

        // 
        // preprocess
        //
        
        double t0 = (detector->timer) ? timerTime(detector->timer) : 0.0 ;
        for (size_t i = 0; i < pyramid.size(); ++i)
        {
            resize(cvImage, pyramid[i]);
        }
        double t1 = (detector->timer) ? timerTime(detector->timer) : 0.0 ;
        
        //
        // detection
        //

        weakClassifiers = 0;

        // ptr to results
        abonDetection * cursor = image->detections;
        int total = 0;
        double t2 = (detector->timer) ? timerTime(detector->timer) : 0.0 ;
        
        for (size_t i = 0; i < pyramid.size(); ++i)
        {
            float scale = float(cvImage->width) / pyramid[i]->width;
            // Scan the image
            int n = scanImage(pyramid[i], classifier, cursor, image->detectionsEnd, det->scanStepH, det->scanStepV);
            // Scan the image

            // Adjust the size and position of the detections according to the pyramid level
            for (abonDetection * d = cursor; d < cursor + n; ++d)
            {
                d->x *= scale;
                d->y *= scale;
                d->width *= scale;
                d->height *= scale;
            }

            // Update ptr to results and result counter
            cursor += n;
            total += n;
        } // Pyramid level loop
        
        double t3 = (detector->timer) ? timerTime(detector->timer) : 0.0 ;
        
        // Fill the timestamps
        image->timestamps[abontsPyramid].start = t0;
        image->timestamps[abontsPyramid].end = t1;
        image->timestamps[abontsDetection].start = t2;
        image->timestamps[abontsDetection].end = t3;
        
        // Set the reslut counter for the image
        image->detectionCount = total;

        // Release the header
        releaseImage(&cvImage);

        image->weakClassifiers = weakClassifiers;

        // Write the image to the output stream
        put(param, image);
	}
	
    if (detector->verbosity)
    {
        printf("["LIBNAME"]: Finish\n");
    }

    return aboneOk;
}


abonErrorCode simpleRelease(abonDetector *detector)
{
    if (detector)
    {
        SimpleDetector * det = (SimpleDetector*) detector;
        releasePyramid(det);
        free(detector);
        if (detector->verbosity)
        {
            printf("["LIBNAME"]: Released\n");
        }
    }
	return aboneOk;
}


extern "C" abonDetector * abonCreate()
{
	SimpleDetector *detector = (SimpleDetector*)malloc(sizeof(SimpleDetector));
	if(detector != NULL)
    {
        detector->header = abonDetectorInitializer; // Initialize the base structure

        // Fill in the simple detector functions
        detector->header.setVerbosity = simpleSetVerbosity;
        detector->header.setTimer = simpleSetTimer;
		detector->header.setClassifier = simpleSetClassifier;
		detector->header.setStep = simpleSetStep;
		detector->header.setSize = simpleSetSize;
        detector->header.setScale = simpleSetScale;
		detector->header.start = simpleStart;
		detector->header.release = simpleRelease;
        
        // Default values of internal parameters
        detector->c = 0;
        detector->scale = 0.875;
        detector->levels = 100;
        detector->scanStepH = 1;
        detector->scanStepV = 1;
        detector->pyramid = 0;
        
        printf("["LIBNAME"]: "LIBVERSION"\n");
	}

	return (abonDetector*)detector;
}

