//-----------------------------------------------------------------------------
// lrd_engine.cpp
// Naive implementation of LRD evaluation
// Roman Juranek, DCGM FIT BUT, Brno
// $Id: lrd_engine.cpp 157 2009-08-25 08:31:24Z ihavel $
//-----------------------------------------------------------------------------


#include "lrd_engine.h"
#include <iostream>
#include <cxcore.h>

using namespace std;

unsigned weakClassifiers;

/// Sum of regions in 3x3 grid.
/// Sums values in regions and stores the results in a vector.
static void sumRegions3x3(unsigned char * data, int w, int h, unsigned widthStep, int * v)
{
	unsigned blockStep = h * widthStep;
    widthStep -= w;

	// Prepare pointer array
    unsigned char * base[9] = {
		data, data+w, data+2*w,
		data+blockStep,data+blockStep+w,data+blockStep+2*w,
		data+2*blockStep,data+2*blockStep+w,data+2*blockStep+2*w,
	};

    for (int y = 0; y < h; ++y)
    {
        // go through all pixels in row and accumulate
		int x = 0;
        while (x < w)
        {
			for (int i = 0; i < 9; ++i)
			{
				v[i] += *base[i];
				++base[i];
			}
			++x;
        }
        // set pointers to next line 
		for (int i = 0; i < 9; ++i)
			base[i] += widthStep;
    }
}


/// Simple version of LRD evaluation.
/// Sums the pixels in the feature grid and calculates ranks of selected pixels.

static inline float evalLRDStageSimple(TImage * image, unsigned smpOffset, TStage * stg)
{
	int values[9] = {0,0,0,0,0,0,0,0,0};
    int featureOffset = stg->y * image->widthStep + stg->x;
    // Get absolute address of feature in image
	unsigned char * base = (unsigned char*)(image->data + smpOffset + featureOffset);
	// Get sums in the feature blocks
    // There is no need to calculate the means because we care only about
    // rang of blocks not actual values 
    sumRegions3x3(base, stg->w, stg->h, image->widthStep, values);
	
	int countA = 0;
	int countB = 0;
	int valA = values[int(stg->A)];
	int valB = values[int(stg->B)];

	// calculate ranks
    int * data = values;
	for (int i = 0; i < 9; ++i, ++data)
	{
		if (valA > *data) ++countA;
		if (valB > *data) ++countB;
	}

    int lrd = countA - countB;

	// return weak hypothesis response
    return stg->alpha[lrd];
}

static inline float evalLRPStageSimple(TImage * image, unsigned smpOffset, TStage * stg)
{
	int values[9] = {0,0,0,0,0,0,0,0,0};
    int featureOffset = stg->y * image->widthStep + stg->x;
    // Get absolute address of feature in image
	unsigned char * base = (unsigned char*)(image->data + smpOffset + featureOffset);
	// Get sums in the feature blocks
    // There is no need to calculate the means because we care only about
    // rang of blocks not actual values 
    sumRegions3x3(base, stg->w, stg->h, image->widthStep, values);
	
	int countA = 0;
	int countB = 0;
	int valA = values[int(stg->A)];
	int valB = values[int(stg->B)];

	// calculate ranks
    int * data = values;
	for (int i = 0; i < 9; ++i, ++data)
	{
		if (valA > *data) ++countA;
		if (valB > *data) ++countB;
	}

    int lrp = 10 * countB + countA;

	// return weak hypothesis response
    return stg->alpha[lrp];
}

typedef float StageEvalFunc(TImage * image, unsigned smpOffset, TStage * stg);

static int evalLRClassifier(TImage * image, unsigned smpOffset, TClassifier * classifier, float * response, StageEvalFunc * eval)
{
    assert(eval != 0 && "LRD or LRP evaluation function is missing!");
    *response = 0.0;
    // go through all stages and accumulate response
    TStage * stages = (TStage*)(classifier->stage);
	for (TStage * stage = stages; stage < stages + classifier->evaluatedStageCount; ++stage)
    {
        ++weakClassifiers;
        //printf("%d: x=%d; y=%d; w=%d; h=%d; a=%d; b=%d; theta=%f; alpha=%d\n",
        //        stage-stages, stage->x, stage->y, stage->w, stage->h, stage->A, stage->B, stage->theta_b, stage->alpha - classifier->alpha);
        *response += eval(image, smpOffset, stage);
        // Test the waldboost threshold
        if (*response < stage->theta_b)
        {
            return 0;
        }
    }

    // All stages have been evaluated - test final stage threshold
    return (*response > classifier->threshold) ? 1 : 0;
}


unsigned scanImage(TImage * image, TClassifier * classifier, abonDetection * results, abonDetection * results_end, int hStep, int vStep)
{
    StageEvalFunc * eval = 0;
    switch (classifier->tp)
    {
    case LRD:
        eval = evalLRDStageSimple;
        break;
    case LRP:
        eval = evalLRPStageSimple;
        break;
    default:
        break;
    };

    if (!eval) return 0;

    // Set pointer to position for first detection
    abonDetection * iter = results;
    if(results == results_end)
        return 0;

    unsigned rowStep = vStep * image->widthStep;
    unsigned rowOffset = 0;
    for (unsigned y = 0; y < image->height-classifier->height; y+=vStep, rowOffset+=rowStep)
    {
        for (unsigned x = 0; x < image->width-classifier->width; x+=hStep)
        {
            // Evaluate classifier
            float response;
            int positive = evalLRClassifier(image, rowOffset+x, classifier, &response, eval);
            if (positive) // write position of the detection
            {
                *iter = abonDetection(x, y, classifier->width, classifier->height, response, 0.0f);
                ++iter;
                if(iter == results_end)
                    return iter-results;
            }
        }
    }

    return iter - results;
}


