//-----------------------------------------------------------------------------
// libsse.cpp
// Pyramid of images and auxiliary structures for detection.
// Jiri Havel, DCGM FIT BUT, Brno
// $Id: libsse.cpp 178 2009-10-08 12:21:51Z ijuranek $
//-----------------------------------------------------------------------------
#include "libsse.h"
#include "scale.h"
#include "cpuid.h"

#include <algorithm>
#include <climits>
#include <iostream>

using namespace std;

#ifdef _OPENMP
#include <omp.h>
#endif

#ifdef _OPENMP
#define LIBNAME "libsse+omp"
#else
#define LIBNAME "libsse"
#endif

#define LIBVERSION "$Id: libsse.cpp 178 2009-10-08 12:21:51Z ijuranek $"

const unsigned LEVELS_PER_OCTAVE = 5;
const unsigned INITIAL_DETECTION_COUNT = 512;

struct DetCmp
{
	bool operator()(const abonDetection &a, const abonDetection &b)
	{
		return b.response < a.response;
	}
};

extern "C" { // Plugin interface --------------------------------------------------

abonDetector *abonCreate()
{
	printf("["LIBNAME"]: "LIBVERSION"\n");

	//otestovat schopnosti procesoru
	if(!hasCPUID())
	{
		fprintf(stderr, "["LIBNAME"]: CPUID instruction not supported!\n");
		return NULL;
	}
	if(!hasSSE2())
	{
		fprintf(stderr, "["LIBNAME"]: SSE2 instruction set not supported!\n");
		return NULL;
	}

#ifndef _OPENMP
	return new(nothrow) SSEDetector;
#else
	return new(nothrow) OMPSSEDetector;
#endif
}

abonErrorCode sseSetClassifier(abonDetector *detector, TClassifier *classifier)
{
	if((classifier->tp != LRD) && (classifier->tp != LRP) && (classifier->tp != LRP3) && (classifier->tp != LBP))
	{
		fprintf(stderr, "["LIBNAME"]: ERROR: Can evaluate only LRD, LRP and LBP classifiers!\n");
		return aboneError;
	}
    for(unsigned i = 0; i < classifier->stageCount; ++i)
    {
        if(
            (((TStage*)classifier->stage)[i].w > 2) ||
            (((TStage*)classifier->stage)[i].h > 2)
            )
        {
		    fprintf(stderr, "["LIBNAME"]: ERROR: Can evaluate only classifier with kernels up to 2x2px!\n");
		    return aboneError;
        }
    }
	detector->classifier = classifier;
	return aboneOk;
}

abonErrorCode sseStart(abonDetector *det, abonCbGetImage *get, abonCbPutImage *put, void *param)
{
    try {
	    SSEDetector *detector = (SSEDetector *)det;
        if(detector == NULL) return aboneError;
	    detector->Init();
	    if(detector->verbosity) printf("["LIBNAME"]: Starting...\n");

        abonImageData *image;
	    while((image = get(param)) != NULL)
	    {
            assert(image->width == detector->width);
            assert(image->height == detector->height);
            assert(image->pitch == detector->pitch);

#if MAKE_STATS
            resetStats();
#endif

            //detect & measure time
		    double t0 = detector->timer == NULL ? 0.0 : timerTime(detector->timer);
		    detector->Build(image);
		    double t1 = detector->timer == NULL ? 0.0 : timerTime(detector->timer);
		    detector->Convolve();
		    double t2 = detector->timer == NULL ? 0.0 : timerTime(detector->timer);
		    Detections & detections = detector->Scan();
		    double t3 = detector->timer == NULL ? 0.0 : timerTime(detector->timer);

            //set timestamps
		    image->timestamps[abontsPyramid].start = t0;
		    image->timestamps[abontsPyramid].end = t1;
		    image->timestamps[abontsConvolution].start = t1;
		    image->timestamps[abontsConvolution].end = t2;
		    image->timestamps[abontsDetection].start = t2;
		    image->timestamps[abontsDetection].end = t3;

#if MAKE_STATS
            // write stats
            cout << _windows << "; " << _hypotheses << endl;
#endif

            //copy detections
            if(int(detections.size()) > (image->detectionsEnd - image->detections))
		    {
                //select only biggest responses
			    std::sort(detections.begin(), detections.end(), DetCmp());
                image->detectionCount = (image->detectionsEnd - image->detections);
		    }
		    else
		    {
			    image->detectionCount = detections.size();
		    }
		    std::copy(detections.begin(), detections.begin() + image->detectionCount, image->detections);

            //return image
		    abonErrorCode ec = put(param, image);
		    if(ec != aboneOk)
		    {
                fprintf(stderr, "["LIBNAME"]: ERROR: Can't put image!\n");
			    return ec;
		    }
	    }

        return aboneOk;
    } catch(bad_alloc & ex) {
        fprintf(stderr, "["LIBNAME"]: ERROR: %s!\n", ex.what());
        return aboneMemory;
    } catch(exception & ex) {
        fprintf(stderr, "["LIBNAME"]: ERROR: %s!\n", ex.what());
        return aboneError;
    } catch(abonErrorCode ec) {
        fprintf(stderr, "["LIBNAME"]: ERROR: %d!\n", ec);
        return ec;
    } catch(...) {
        return aboneError;
    }
}

abonErrorCode sseRelease(abonDetector *detector)
{
    if(detector == NULL) return aboneOk;
	if(detector->verbosity) printf("["LIBNAME"]: Releasing...\n");

    delete (SSEDetector *)detector;

    return aboneOk;
}

#ifdef _OPENMP

abonErrorCode ompsseStart(abonDetector *det, abonCbGetImage *get, abonCbPutImage *put, void *param)
{
    try {
	    OMPSSEDetector *detector = (OMPSSEDetector *)det;
	    detector->Init();
	    if(detector->verbosity) printf("["LIBNAME"]: Starting...\n");

        abonImageData *image;
	    while((image = get(param)) != NULL)
	    {
            assert(image->width == width);
            assert(image->height == height);
            assert(image->pitch == pitch);

#if MAKE_STATS
            resetStats();
#endif
            //detect & measure time
		    double t0 = detector->timer == NULL ? 0.0 : timerTime(detector->timer);
		    detector->Build(image);
		    double t1 = detector->timer == NULL ? 0.0 : timerTime(detector->timer);
		    Detections & detections = detector->ConvolveScan();
		    double t2 = detector->timer == NULL ? 0.0 : timerTime(detector->timer);

		    //set timestamps
		    image->timestamps[abontsPyramid].start = t0;
		    image->timestamps[abontsPyramid].end = t1;
		    image->timestamps[abontsConvolution].start = 0;
		    image->timestamps[abontsConvolution].end = 0;
		    image->timestamps[abontsDetection].start = t1;
		    image->timestamps[abontsDetection].end = t2;

#if MAKE_STATS
            cout << _windows << "; " << _hypotheses << endl;
#endif
		    //copy detections
            if(int(detections.size()) > (image->detectionsEnd - image->detections))
		    {
                //select only biggest responses
			    std::sort(detections.begin(), detections.end(), DetCmp());
                image->detectionCount = (image->detectionsEnd - image->detections);
		    }
		    else
		    {
			    image->detectionCount = detections.size();
		    }
		    std::copy(detections.begin(), detections.begin() + image->detectionCount, image->detections);

		    //return image
		    abonErrorCode ec = put(param, image);
		    if(ec != aboneOk)
		    {
			    return ec;
		    }
	    }

	    return aboneOk;
    } catch(bad_alloc & ex) {
        fprintf(stderr, "["LIBNAME"]: ERROR: %s!\n", ex.what());
        return aboneMemory;
    } catch(exception & ex) {
        fprintf(stderr, "["LIBNAME"]: ERROR: %s!\n", ex.what());
        return aboneError;
    } catch(abonErrorCode ec) {
        fprintf(stderr, "["LIBNAME"]: ERROR: %d!\n", ec);
        return ec;
    } catch(...) {
        return aboneError;
    }
}

abonErrorCode ompsseRelease(abonDetector *detector)
{
    if(detector == NULL) return aboneOk;
	if(detector->verbosity) printf("["LIBNAME"]: Releasing...\n");

    delete (OMPSSEDetector *)detector;

    return aboneOk;
}

#endif

} // Plugin interface --------------------------------------------------

// Pyramid level

SSEDetector::Level::Level()
{
	memset(this, 0, sizeof(*this));
}

void SSEDetector::Level::Init(
	const unsigned width,
	const unsigned height,
	const unsigned pitch,
	const TClassifier * classifier
)
{
	Kill();
	this->width = width;
	this->height = height;
	this->pitch = pitch;
	if((this->data = (char*)_mm_malloc((height+1)*pitch, 16)) == NULL)
    {
        throw bad_alloc();
    }
	convolutions = createConvolutions(width, height);
	xAddressTable = createColAddressTable(convolutions);
	yAddressTable = createRowAddressTable(convolutions);
	ranks = createClassifierRankTable(classifier, convolutions);
}

void SSEDetector::Level::Kill()
{
	_mm_free(data);
	releaseConvolutions(&convolutions);
	delete [] xAddressTable;
	delete [] yAddressTable;
	delete [] ranks;
	memset(this, 0, sizeof(*this));
}

void SSEDetector::Level::Convolve()
{
	convolve12all(width, height, pitch, data, *convolutions);
}

unsigned SSEDetector::Level::Scan(
	const TClassifier * classifier,
	const unsigned hstep, const unsigned vstep,
	Detections & detections)
{
	return scanConvolvedImage(convolutions, xAddressTable, yAddressTable, classifier, ranks, hstep, vstep, detections);
}

// Detector

SSEDetector::SSEDetector()
: abonDetector(abonDetectorInitializer)
{
	detections.reserve(INITIAL_DETECTION_COUNT);
	abonDetector::start = sseStart;
	abonDetector::setClassifier = sseSetClassifier;
	abonDetector::release = sseRelease;
}

SSEDetector::~SSEDetector() throw()
{
	Kill();
}

unsigned SSEDetector::getLevelCount(
	const unsigned w, const unsigned h, const TClassifier * classifier, const unsigned o)
{
	float width = w;
	float height = h;
	float octaveWidth = w;
	float octaveHeight = h;
	unsigned octaves = o == 0 ? UINT_MAX : o;
	unsigned levels = 0;
	while((octaves > 0) && (width >= classifier->width) && (height >= classifier->height))
	{
		++levels;
		if(levels%LEVELS_PER_OCTAVE == 0)
		{
			--octaves;
			octaveWidth  = width  = octaveWidth *0.5f;
			octaveHeight = height = octaveHeight*0.5f;
		}
		else
		{
			width  = width *0.875f;
			height = height*0.875f;
		}
	}
	return levels;
}

void SSEDetector::recalculateCorners(const unsigned level, Detections::iterator begin, Detections::iterator end)
{
	for(Detections::iterator i = begin; i != end; ++i)
	{
		float fx = float(width )/float(levels[level].w);
		float fy = float(height)/float(levels[level].h);
		i->x = int(i->x*fx);
		i->y = int(i->y*fy);
		i->width  = int(i->width *fx);
		i->height = int(i->height*fy);
	}
}

typedef AlignOp<4, 4> AlignConv;
typedef PairOp<AlignConv, Scale8to7> Align8to7;
typedef PairOp<Align8to7, ScaleHalf> AlignBoth;

void SSEDetector::Init()
{
	if(verbosity) printf("["LIBNAME"]: Initializing...\n");
	levels.resize(getLevelCount(width, height, classifier));

	std::pair<unsigned, unsigned> s = alignSizeForOp<AlignBoth>(width, height);
	levels[0].Init(s.first, s.second, s.first, classifier);
	levels[0].w = float(levels[0].width);
	levels[0].h = float(levels[0].height);
	for(unsigned i = 1; i < levels.size(); ++i)
	{
		float w, h;
		if(i%LEVELS_PER_OCTAVE == 0)
		{
			std::pair<unsigned, unsigned> ds = dstSizeOfOp<ScaleHalf>(
				levels[i-LEVELS_PER_OCTAVE].width,
				levels[i-LEVELS_PER_OCTAVE].height
			);
			s = alignSizeForOp<AlignBoth>(ds.first, ds.second);
			w = levels[i-LEVELS_PER_OCTAVE].w*0.5f;
			h = levels[i-LEVELS_PER_OCTAVE].h*0.5f;
		}
		else
		{
			std::pair<unsigned, unsigned> ds = dstSizeOfOp<Scale8to7>(
				levels[i-1].width, levels[i-1].height);
			s = i%LEVELS_PER_OCTAVE < LEVELS_PER_OCTAVE-1
				? alignSizeForOp<Align8to7>(ds.first, ds.second)
				: alignSizeForOp<AlignConv>(ds.first, ds.second);
			w = levels[i-1].w*0.875f;
			h = levels[i-1].h*0.875f;
		}
		levels[i].Init(s.first, s.second, s.first, classifier);
		levels[i].w = w;
		levels[i].h = h;
	}
	if(verbosity)
	{
		printf("["LIBNAME"]: Pyramid:\n");
		for(unsigned i = 0; i < levels.size(); ++i)
		{
            printf("["LIBNAME"]:  %fx%f (%dx%d_%d) 0x%x-0x%x %s\n",
                levels[i].w, levels[i].h,
                levels[i].width, levels[i].height, levels[i].pitch,
                levels[i].data, levels[i].data + levels[i].width*levels[i].pitch,
                i%LEVELS_PER_OCTAVE == 0 ? " octave" : "");
		}
	}
}

void SSEDetector::Kill()
{
	for(unsigned i = 0; i < levels.size(); ++i)
	{
		levels[i].Kill();
	}
	levels.clear();
}

void SSEDetector::Build(abonImageData *image)
{
	//copy first level
	for(unsigned r = 0; r < height; ++r)
	{
		memcpy(
			levels[0].data + r*levels[0].pitch,
			image->data + r*image->pitch, width);
		memset(
			levels[0].data + r*levels[0].pitch + width,
			0, levels[0].pitch - width);
	}
	//build pyramid
	for(unsigned i = 1; i < levels.size(); ++i)
	{
		if((i%LEVELS_PER_OCTAVE) == 0)
		{
			doOp<ScaleHalf>(
				levels[i-LEVELS_PER_OCTAVE].width,
				levels[i-LEVELS_PER_OCTAVE].height,
				levels[i-LEVELS_PER_OCTAVE].pitch,
				levels[i-LEVELS_PER_OCTAVE].data,
				levels[i].width, levels[i].height,
				levels[i].pitch, levels[i].data
			);
		}
		else
		{
			doOp<Scale8to7>(
				levels[i-1].width, levels[i-1].height,
				levels[i-1].pitch, levels[i-1].data,
				levels[i].width, levels[i].height,
				levels[i].pitch, levels[i].data
			);
		}
	}
}

void SSEDetector::Convolve()
{
	for(unsigned l = 0; l < levels.size(); ++l)
	{
		levels[l].Convolve();
	}
}

Detections & SSEDetector::Scan()
{
	detections.clear();

	for(unsigned l = 0; l < levels.size(); ++l)
	{
		unsigned previous = detections.size();

		//scan
		levels[l].Scan(classifier, hstep, vstep, detections);

		recalculateCorners(l, detections.begin()+previous, detections.end());
	}//for

	return detections;
}

// Parallel SSE detector

#ifdef _OPENMP

OMPSSEDetector::OMPSSEDetector()
: threadTemps(omp_get_num_procs()-1)
{
	for(unsigned i = 0; i < threadTemps.size(); ++i)
	{
		threadTemps[i].reserve(INITIAL_DETECTION_COUNT);
	}
	abonDetector::start = ompsseStart;
	abonDetector::release = ompsseRelease;
}

OMPSSEDetector::~OMPSSEDetector() throw()
{
}

Detections & OMPSSEDetector::ConvolveScan()
{
	#pragma omp parallel num_threads(threadTemps.size()+1)
	{
		unsigned thread_num = omp_get_thread_num();
		//for thread 0 use provided detections
		Detections & results = thread_num == 0 ? detections : threadTemps[thread_num-1];
		results.clear();

		#pragma omp for schedule(dynamic, 1)
		for(int l = 0; l < static_cast<int>(levels.size()); ++l)
		{
			unsigned previous = results.size();
			
			levels[l].Convolve();

			levels[l].Scan(classifier, hstep, vstep, results);

			recalculateCorners(l, results.begin()+previous, results.end());
		}//for
	}//parallel block

	//merge temporary detection lists
	for(unsigned i = 0; i < threadTemps.size(); ++i)
	{
		detections.insert(detections.end(), threadTemps[i].begin(), threadTemps[i].end());
	}

	return detections;
}

#endif
