/*
 * LRP/LRD/LBP evaluation engine - SIMD version
 *
 * Known issues:
 * - Image size dependency
 *   Problem:
 *     The classifier parameters are tuned for particular size of image
 *     TStage::ATab and TStage::BTab and also addressing tables should
 *     be provided according to the size.
 *   Solution:
 *     - Addressing tables will be within an convolution image (it makes more sense)
 *     - The classifier will hold constant vector of stages and its dynamic
 *       versions calculated for particular image size.
 * - Even though the LBP classifiers do not need ranks to be calculated,
 *   the tables are still created.
 *
 *
 *  Spravny postup pouziti
 *  - Znama velikost obrazu
 *  - createConvolutions
 *  - prepareAddressTables
 *  - loadClassifier
 *  - initClassifier
 *  
 *  - convolveImage
 *  - scanConvolvedImage
 *
 */

#include "lrd_engine.h"
#include "lrd_const.h"
#include "imagetools.h"
#include <iostream>
#include <pmmintrin.h>

using namespace std;


long long _windows;
long long _hypotheses;

void resetStats()
{
    _windows = 0;
    _hypotheses = 0;
}


Convolutions * createConvolutions(const unsigned width, const unsigned height)
{
    Convolutions * tmp = new Convolutions(4);
    Convolutions & conv = *tmp;
    conv[0] = new TConvolution(width, height, 1, 1);
    conv[1] = new TConvolution(width, height, 2, 1);
    conv[2] = new TConvolution(width, height, 1, 2);
    conv[3] = new TConvolution(width, height, 2, 2);
    return tmp;
}


void releaseConvolutions(Convolutions ** conv)
{
    if (conv && *conv)
    {
        Convolutions & c = **conv;
        // delete convolution images
        for (unsigned i = 0; i < c.size(); ++i)
        {
            if (c[i])
            {
                delete c[i];
            }
        }
        // delete the vector
        delete *conv;
        *conv = 0;
    }
}


int * createColAddressTable(const Convolutions * c)
{
    int * table = new int[(*c)[0]->width * 4];
    for (unsigned i = 0; i < (*c)[0]->width; ++i)
    {
        table[4 * i + 0] = (i & 0xFFFFFFFE) << 1;
        table[4 * i + 1] = (i & 0xFFFFFFFC);
        table[4 * i + 2] = (i & 0xFFFFFFFE) << 1;
        table[4 * i + 3] = (i & 0xFFFFFFFC);
    }
    return table;
}

    
int * createRowAddressTable(const Convolutions * c)
{
    int * table = new int[(*c)[0]->height * 4];
    for (unsigned i = 0; i < (*c)[0]->height; ++i)
    {
        table[4 * i + 0] = (i >> 1) * (*c)[0]->rowStep;
        table[4 * i + 1] = (i >> 1) * (*c)[1]->rowStep;
        table[4 * i + 2] = (i >> 2) * (*c)[2]->rowStep;
        table[4 * i + 3] = (i >> 2) * (*c)[3]->rowStep;
    }
    return table;
}

Rank * createClassifierRankTable(const TClassifier * classifier, const Convolutions * conv)
{
    Rank * ranks = new Rank[classifier->stageCount];
    for (unsigned s = 0; s < classifier->stageCount; ++s)
    {
        const TStage & stage = ((TStage*)(classifier->stage))[s];

        for (int t = 0; t < 4; ++t)
        {
            int & A = ranks[s][4 * t + 0];
            int & B = ranks[s][4 * t + 1];
            int & C = ranks[s][4 * t + 2];
            A = rankTable[t][int(stage.A)];
            B = rankTable[t][int(stage.B)];
            C = rankTable[t][int(stage.C)];
            if (A >= 8) A += (*conv)[stage.szType]->rowStep - 8;
            if (B >= 8) B += (*conv)[stage.szType]->rowStep - 8;
            if (C >= 8) C += (*conv)[stage.szType]->rowStep - 8;
        }
    }
    return ranks;
}

/////////////// LRD /////////////////////

/// SIMD evaluation
static inline float evalLRDStage(
        const Convolutions * const convolutions,
        const int * const addrX, const int * const addrY,
        int fx, int fy, int sampleModPos,
        const TStage * const stg, const Rank * const rank)
{
    // Get the convolution image;
    TConvolution * conv = (*convolutions)[stg->szType];
    
    // index to tables - depends on feature size, sample position and feature position relative to sample
    int tableIdx = (stg->szType << 8) | sampleModPos | stg->posType;
    
    // Get the block of the convolution (depends on feature modulo shift)
    int blockId = blockTable[tableIdx];
    
    // Get the mask type
    int maskType = maskTable[tableIdx];
    
    // Get address of the feature in image
    int dataOffset = addrX[4 * (fx + stg->x) + stg->szType] + addrY[4 * (fy+stg->y) + stg->szType];
    //signed char * data0 = conv->image + (blockId * conv->blockStep) + dataOffset;
    signed char * data0 = conv->block[blockId] + dataOffset;
    signed char * data1 = data0 + conv->rowStep;
    
    // Get the A nd B rank index according to the shift type
    int AOffset = (*rank)[4 * maskType + 0];
    int BOffset = (*rank)[4 * maskType + 1];

    register __m128i data = _mm_set_epi64(*(__m64*)(data1), *(__m64*)(data0));
    register __m128i zero = _mm_setzero_si128();

    union {
        __m128i q;
        signed short ss[8];
    } diff = { _mm_sub_epi16(
        _mm_sad_epu8( // countA
            _mm_and_si128(
                _mm_cmpgt_epi8(_mm_set1_epi8(*(data0+AOffset)), data),
                masks[maskType].q),
            zero),
        _mm_sad_epu8( // countB
            _mm_and_si128(
                _mm_cmpgt_epi8(_mm_set1_epi8(*(data0+BOffset)), data),
                masks[maskType].q),
            zero)
        )
    }; 

    int lrd = diff.ss[4] + diff.ss[0];
    _mm_empty();

    return stg->alpha[lrd]; // should be lrd+8 to make the index positive!!! The pointer was set to offset 8 to avoid this sum!!
}


static inline int evalLRDClassifier(
        const Convolutions * const convolutions, // The input image
        const int * const addrX, const int * const addrY, // Convolution Address tables
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const int x, const int y, // Position to evaluate
        float * const response) // Response value
{
    *response = 0.0f;
    
    // posType - yyxx0000
    int posType = (((y & 0x03) << 6) | (x & 0x03) << 4); // (position % 4) * 16
    
    for (unsigned s = 0; s < classifier->evaluatedStageCount; ++s)
    {
        const TStage * stg = (TStage*)(classifier->stage) + s;
        const Rank * rank = ranks + s;
        *response += evalLRDStage(convolutions, addrX, addrY, x, y, posType, stg, rank);
#ifndef DISABLE_WALDBOOST
        if (*response < stg->theta_b) // Negative result
        {
#if MAKE_STATS
            _hypotheses += s + 1;
#endif
            return 0;
        }
#endif
    }
    
#if MAKE_STATS
    _hypotheses += classifier->evaluatedStageCount;
#endif
    return (*response > classifier->threshold) ? 1 : 0;
}

static unsigned scanConvolvedImageLRD(
        const Convolutions * const convolutions, // The convolved image
        const int * const xtable, const int * const ytable, // Tables for addressing the convolutions
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const unsigned hstep, const unsigned vstep, // Scan step
        std::vector<abonDetection> & detections // Detection vector
        )
{
    unsigned originalSize = detections.size();

#if MAKE_STATS
    // Za kazdy zpracovany radek se k _windows pricte winsPerRow.
    // Na prvnim radku se nejdriv spocita skutecny pocet vyhodnoceni.
    // Je to lepsi nez to vyhodnocovat nejakym slozitym vzorcem
    int winsPerRow = 0;
    bool counted = false;
#endif
    
    for (unsigned y = 0; y < ((*convolutions)[0]->height - classifier->height); y += vstep)
    {
        for (unsigned x = 0; x < ((*convolutions)[0]->width - classifier->width); x += hstep)
        {
            float response;
            //int d = eval(convolutions, xtable, ytable, classifier, ranks, x, y, &response);
            int d = evalLRDClassifier(convolutions, xtable, ytable, classifier, ranks, x, y, &response);
            
#if MAKE_STATS
            if (!counted) ++winsPerRow;
#endif
            
            if (d) // Positive response
            {
                // create the detection record
                detections.push_back(abonDetection(x, y, classifier->width, classifier->height, response, 0.0f));
            }
        } // x

#if MAKE_STATS
        if (winsPerRow != 0) counted = true;
        _windows += winsPerRow;
#endif

    } // y
    
    return detections.size() - originalSize;
}


/////////////// LRD3 /////////////////////

/// SIMD evaluation
static inline float evalLRP3Stage(
        const Convolutions * const convolutions,
        const int * const addrX, const int * const addrY,
        int fx, int fy, int sampleModPos,
        const TStage * const stg, const Rank * const rank)
{
    // Get the convolution image;
    TConvolution * conv = (*convolutions)[stg->szType];
    
    // index to tables - depends on feature size, sample position and feature position relative to sample
    int tableIdx = (stg->szType << 8) | sampleModPos | stg->posType;
    
    // Get the block of the convolution (depends on feature modulo shift)
    int blockId = blockTable[tableIdx];
    
    // Get the mask type
    int maskType = maskTable[tableIdx];
    
    // Get address of the feature in image
    int dataOffset = addrX[4 * (fx + stg->x) + stg->szType] + addrY[4 * (fy+stg->y) + stg->szType];
    //signed char * data0 = conv->image + (blockId * conv->blockStep) + dataOffset;
    signed char * data0 = conv->block[blockId] + dataOffset;
    signed char * data1 = data0 + conv->rowStep;
    
    // Get the A nd B rank index according to the shift type
    int AOffset = (*rank)[4 * maskType + 0];
    int BOffset = (*rank)[4 * maskType + 1];
    int COffset = (*rank)[4 * maskType + 2];
    //int DOffset = (*rank)[4 * maskType + 3]; // only 3 rank pixels - D is omited

    register __m128i data = _mm_set_epi64(*(__m64*)(data1), *(__m64*)(data0));
    register __m128i zero = _mm_setzero_si128();

    union {
        __m128i q;
        signed short ss[8];
    } countA = {
        _mm_sad_epu8(
            _mm_and_si128(
                _mm_cmpgt_epi8(_mm_set1_epi8(*(data0+AOffset)), data),
                masks[maskType].q),
            zero) };

    union {
        __m128i q;
        signed short ss[8];
    } countB = {
        _mm_sad_epu8(
            _mm_and_si128(
                _mm_cmpgt_epi8(_mm_set1_epi8(*(data0+BOffset)), data),
                masks[maskType].q),
            zero) };

    union {
        __m128i q;
        signed short ss[8];
    } countC = {
        _mm_sad_epu8(
            _mm_and_si128(
                _mm_cmpgt_epi8(_mm_set1_epi8(*(data0+COffset)), data),
                masks[maskType].q),
            zero) };

    int lrp = (100 * (countC.ss[4] + countC.ss[0])) + (10 * (countB.ss[4] + countB.ss[0])) + (countA.ss[4] + countA.ss[0]);

    _mm_empty();

    return stg->alpha[lrp];

}


static inline int evalLRP3Classifier(
        const Convolutions * const convolutions, // The input image
        const int * const addrX, const int * const addrY, // Convolution Address tables
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const int x, const int y, // Position to evaluate
        float * const response) // Response value
{
    *response = 0.0f;
    
    // posType - yyxx0000
    int posType = (((y & 0x03) << 6) | (x & 0x03) << 4); // (position % 4) * 16
    
    for (unsigned s = 0; s < classifier->evaluatedStageCount; ++s)
    {
        const TStage * stg = (TStage*)(classifier->stage) + s;
        const Rank * rank = ranks + s;
        *response += evalLRP3Stage(convolutions, addrX, addrY, x, y, posType, stg, rank);
#ifndef DISABLE_WALDBOOST
        if (*response < stg->theta_b) // Negative result
        {
            return 0;
        }
#endif
    }
    
    return (*response > classifier->threshold) ? 1 : 0;
}

static unsigned scanConvolvedImageLRP3(
        const Convolutions * const convolutions, // The convolved image
        const int * const xtable, const int * const ytable, // Tables for addressing the convolutions
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const unsigned hstep, const unsigned vstep, // Scan step
        std::vector<abonDetection> & detections // Detection vector
        )
{
    unsigned originalSize = detections.size();
    
    for (unsigned y = 0; y < ((*convolutions)[0]->height - classifier->height); y += vstep)
    {
        for (unsigned x = 0; x < ((*convolutions)[0]->width - classifier->width); x += hstep)
        {
            float response;
            //int d = eval(convolutions, xtable, ytable, classifier, ranks, x, y, &response);
            int d = evalLRP3Classifier(convolutions, xtable, ytable, classifier, ranks, x, y, &response);
            
            if (d) // Positive response
            {
                // create the detection record
                detections.push_back(abonDetection(x, y, classifier->width, classifier->height, response, 0.0f));
            }
        } // x
    } // y
    
    return detections.size() - originalSize;
}

/////////////// LRP /////////////////////

/// SIMD evaluation
static inline float evalLRPStage(
        const Convolutions * const convolutions,
        const int * const addrX, const int * const addrY,
        int fx, int fy, int sampleModPos,
        const TStage * const stg, const Rank * const rank)
{
    // Get the convolution image;
    TConvolution * conv = (*convolutions)[stg->szType];
    
    // index to tables - depends on feature size, sample position and feature position relative to sample
    int tableIdx = (stg->szType << 8) | sampleModPos | stg->posType;
    
    // Get the block of the convolution (depends on feature modulo shift)
    int blockId = blockTable[tableIdx];
    
    // Get the mask type
    int maskType = maskTable[tableIdx];
    
    // Get address of the feature in image
    int dataOffset = addrX[4 * (fx + stg->x) + stg->szType] + addrY[4 * (fy+stg->y) + stg->szType];
    //signed char * data0 = conv->image + (blockId * conv->blockStep) + dataOffset;
    signed char * data0 = conv->block[blockId] + dataOffset;
    signed char * data1 = data0 + conv->rowStep;
    
    // Get the A nd B rank index according to the shift type
    int AOffset = (*rank)[4 * maskType + 0];
    int BOffset = (*rank)[4 * maskType + 1];

    register __m128i data = _mm_set_epi64(*(__m64*)(data1), *(__m64*)(data0));
    register __m128i zero = _mm_setzero_si128();

    union {
        __m128i q;
        signed short ss[8];
    } countA = {
        _mm_sad_epu8(
            _mm_and_si128(
                _mm_cmpgt_epi8(_mm_set1_epi8(*(data0+AOffset)), data),
                masks[maskType].q),
            zero) };

    union {
        __m128i q;
        signed short ss[8];
    } countB = {
        _mm_sad_epu8(
            _mm_and_si128(
                _mm_cmpgt_epi8(_mm_set1_epi8(*(data0+BOffset)), data),
                masks[maskType].q),
            zero) };

    int lrp = (10 * (countB.ss[4] + countB.ss[0])) + (countA.ss[4] + countA.ss[0]);
    _mm_empty();

    return stg->alpha[lrp];
}


static inline int evalLRPClassifier(
        const Convolutions * const convolutions, // The input image
        const int * const addrX, const int * const addrY, // Convolution Address tables
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const int x, const int y, // Position to evaluate
        float * const response) // Response value
{
    *response = 0.0f;
    
    // posType - yyxx0000
    int posType = (((y & 0x03) << 6) | (x & 0x03) << 4); // (position % 4) * 16
    
    for (unsigned s = 0; s < classifier->evaluatedStageCount; ++s)
    {
        const TStage * stg = (TStage*)(classifier->stage) + s;
        const Rank * rank = ranks + s;
        *response += evalLRPStage(convolutions, addrX, addrY, x, y, posType, stg, rank);
#ifndef DISABLE_WALDBOOST
        if (*response < stg->theta_b) // Negative result
        {
#if MAKE_STATS
            _hypotheses += s + 1;
#endif
            return 0;
        }
#endif
    }
    
#if MAKE_STATS
    _hypotheses += classifier->evaluatedStageCount;
#endif
    return (*response > classifier->threshold) ? 1 : 0;
}

static unsigned scanConvolvedImageLRP(
        const Convolutions * const convolutions, // The convolved image
        const int * const xtable, const int * const ytable, // Tables for addressing the convolutions
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const unsigned hstep, const unsigned vstep, // Scan step
        std::vector<abonDetection> & detections // Detection vector
        )
{
    unsigned originalSize = detections.size();
    
#if MAKE_STATS
    // Za kazdy zpracovany radek se k _windows pricte winsPerRow.
    // Na prvnim radku se nejdriv spocita skutecny pocet vyhodnoceni.
    // Je to lepsi nez to vyhodnocovat nejakym slozitym vzorcem
    int winsPerRow = 0;
    bool counted = false;
#endif
    for (unsigned y = 0; y < ((*convolutions)[0]->height - classifier->height); y += vstep)
    {
        for (unsigned x = 0; x < ((*convolutions)[0]->width - classifier->width); x += hstep)
        {
            float response;
            //int d = eval(convolutions, xtable, ytable, classifier, ranks, x, y, &response);
            int d = evalLRPClassifier(convolutions, xtable, ytable, classifier, ranks, x, y, &response);
            
#if MAKE_STATS
            if (!counted) ++winsPerRow;
#endif
            if (d) // Positive response
            {
                // create the detection record
                detections.push_back(abonDetection(x, y, classifier->width, classifier->height, response, 0.0f));
            }
        } // x
#if MAKE_STATS
        if (winsPerRow != 0) counted = true;
        _windows += winsPerRow;
#endif
    } // y
    return detections.size() - originalSize;
}
/////////////// LBP /////////////////////

/// SIMD evaluation
static inline float evalLBPStage(
        const Convolutions * const convolutions,
        const int * const addrX, const int * const addrY,
        int fx, int fy, int sampleModPos,
        const TStage * const stg)
{
    // Get the convolution image;
    TConvolution * conv = (*convolutions)[stg->szType];
    
    // index to tables - depends on feature size, sample position and feature position relative to sample
    int tableIdx = (stg->szType << 8) | sampleModPos | stg->posType;
    
    // Get the block of the convolution (depends on feature modulo shift)
    int blockId = blockTable[tableIdx];
    
    // Get the mask type
    int maskType = maskTable[tableIdx];
    
    // Get address of the feature in image
    int dataOffset = addrX[4 * (fx + stg->x) + stg->szType] + addrY[4 * (fy+stg->y) + stg->szType];
    //signed char * data0 = conv->image + (blockId * conv->blockStep) + dataOffset;
    signed char * data0 = conv->block[blockId] + dataOffset;
    signed char * data1 = data0 + conv->rowStep;

    // get the center pixel
    static const int centerOffset[4] = {3, 6, 1, 4};
    signed char * center = (maskType < 2) ? data0 : data1;
    center += centerOffset[maskType];
    
    register __m128i data = _mm_set_epi64(*(__m64*)(data1), *(__m64*)(data0));
    register __m128i zero = _mm_setzero_si128();

    // LBP evaluation
    // Comparison (pixel values to central pixel) result is used
    // to mask LBP weights. The masked vector is then summed up
    // producing the LBP value.
    union {
        __m128i q;
        signed short ss[8];
    } result = {
        _mm_sad_epu8(
            _mm_and_si128(
                lbpWeights[maskType].q,
                _mm_cmpgt_epi8(data, _mm_set1_epi8(*center))),
            zero)
    };

    int lbp = result.ss[4] + result.ss[0];
    _mm_empty();

    return stg->alpha[lbp];
}


static inline int evalLBPClassifier(
        const Convolutions * const convolutions, // The input image
        const int * const addrX, const int * const addrY, // Convolution Address tables
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const int x, const int y, // Position to evaluate
        float * const response) // Response value
{
    *response = 0.0f;
    
    // posType - yyxx0000
    int posType = (((y & 0x03) << 6) | (x & 0x03) << 4); // (position % 4) * 16
    
    for (unsigned s = 0; s < classifier->evaluatedStageCount; ++s)
    {
        const TStage * stg = (TStage*)(classifier->stage) + s;
        //const Rank * rank = ranks + s; // LBPs do not need ranks
        *response += evalLBPStage(convolutions, addrX, addrY, x, y, posType, stg);
#ifndef DISABLE_WALDBOOST
        if (*response < stg->theta_b) // Negative result
        {
#if MAKE_STATS
            _hypotheses += s + 1;
#endif
            return 0;
        }
#endif
    }
    
#if MAKE_STATS
    _hypotheses += classifier->evaluatedStageCount;
#endif
    return (*response > classifier->threshold) ? 1 : 0;
}

static unsigned scanConvolvedImageLBP(
        const Convolutions * const convolutions, // The convolved image
        const int * const xtable, const int * const ytable, // Tables for addressing the convolutions
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const unsigned hstep, const unsigned vstep, // Scan step
        std::vector<abonDetection> & detections // Detection vector
        )
{
    unsigned originalSize = detections.size();
    
#if MAKE_STATS
    // Za kazdy zpracovany radek se k _windows pricte winsPerRow.
    // Na prvnim radku se nejdriv spocita skutecny pocet vyhodnoceni.
    // Je to lepsi nez to vyhodnocovat nejakym slozitym vzorcem
    int winsPerRow = 0;
    bool counted = false;
#endif
    for (unsigned y = 0; y < ((*convolutions)[0]->height - classifier->height); y += vstep)
    {
        for (unsigned x = 0; x < ((*convolutions)[0]->width - classifier->width); x += hstep)
        {
            float response;
            //int d = eval(convolutions, xtable, ytable, classifier, ranks, x, y, &response);
            int d = evalLBPClassifier(convolutions, xtable, ytable, classifier, ranks, x, y, &response);
#if MAKE_STATS
            if (!counted) ++winsPerRow;
#endif
            
            if (d) // Positive response
            {
                // create the detection record
                detections.push_back(abonDetection(x, y, classifier->width, classifier->height, response, 0.0f));
            }
        } // x
#if MAKE_STATS
        if (winsPerRow != 0) counted = true;
        _windows += winsPerRow;
#endif
    } // y
    
    return detections.size() - originalSize;
}


////////////////////////////////////////

typedef unsigned ImageScanFunc(
        const Convolutions * const convolutions, // The convolved image
        const int * const xtable, const int * const ytable, // Tables for addressing the convolutions
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const unsigned hstep, const unsigned vstep, // Scan step
        std::vector<abonDetection> & detections // Detection vector
        );

unsigned scanConvolvedImage(
        const Convolutions * const convolutions, // The convolved image
        const int * const xtable, const int * const ytable, // Tables for addressing the convolutions
        const TClassifier * const classifier, const Rank * const ranks, // The classifier
        const unsigned hstep, const unsigned vstep, // Scan step
        std::vector<abonDetection> & detections // Detection vector
        )
{
    ImageScanFunc * eval = 0;

    // select evaluation function according
    // to classifier type
    switch (classifier->tp)
    {
    case LRD:
        eval = scanConvolvedImageLRD;
        break;
    case LRP:
        eval = scanConvolvedImageLRP;
        break;
    case LRP3:
        eval = scanConvolvedImageLRP3;
        break;
    case LBP:
        eval = scanConvolvedImageLBP;
        break;
    default:
        return 0;
        break;
    };

    return eval(convolutions, xtable, ytable, classifier, ranks, hstep, vstep, detections);
}

