//-----------------------------------------------------------------------------
// scale.h
// Fixed step image scaling, optimized via SSE intrinsics
// Jiri Havel, DCGM FIT BUT, Brno
// $Id: scale.h 87 2009-04-21 11:12:48Z ihavel $
//-----------------------------------------------------------------------------

#ifndef _SCALE_H_
#define _SCALE_H_

#include <cassert>

#include <utility>

#include <emmintrin.h>

/*
Fixed step scaling of 8b greyscale images, implemented via generic block transformations

doOp<op>(src, dst) performs operation specified by op on src image and stores resulting image to dst

op - block operation class

op::src{Width|Height} - size of input blocks
op::dst{Width|Height} - size of output blocks
op::rowOverlap - count of bytes written outside of output block
op::doOp - process one block, used by doOp

src - source image

Must be aligned on input block size, do that by:
alignForOp<op>(requestedSize) returns nearest greater aligned size
If src is used for more operations, combine them by PairOp<op1, op2>
Resulting alignment is lest common multiple of both. PairOp can be nested.
User alignment can be added by AlignOp<width, height>
PairOp and AlignOp can't be used in doOp.

dst - destination image

Must have at least size of output block count times output block size
dstSizeOfOp<op>(srcSize) returns minimum size, required for output
It is safe to align this size further by alignForOp

Provided operations :

Scale8to7 - scale down by factor 7/8, using bilinear filtering
ScaleHalf - scale down by averaging four pixels

Both are optimized by SSE intrinsics and require at least SSE2,
but are five times faster than cvResize with CV_INTER_LINEAR
*/

// Greatest common divisor

template <int M, int N>
struct Gcd
{
    enum { value = Gcd<N, M%N>::value };
};
 
template <int N>
struct Gcd<N, 0>
{
    enum { value = N };
};

// Least common multiple

template <int M, int N>
struct Lcm
{
    enum { value = M*N/Gcd<M, N>::value };
};

// User defined alignment

template <int W, int H>
struct AlignOp
{
    enum
    {
        srcWidth  = W,
        srcHeight = H
    };
};

// Combination of two operation for alignment

template <class A, class B>
struct PairOp
{
    enum
    {
        srcWidth  = Lcm<A::srcWidth , B::srcWidth >::value,
        srcHeight = Lcm<A::srcHeight, B::srcHeight>::value
    };
};

// Operation for combining W*H blocks to one 

template<class Op, unsigned W, unsigned H>
struct UnrollOp
{
    enum
    {
        srcWidth   = W*Op::srcWidth,
        srcHeight  = H*Op::srcHeight,
        dstWidth   = W*Op::dstWidth,
        dstHeight  = H*Op::dstHeight,
        rowOverlap =   Op::rowOverlap
    };

    static inline void doBlock(const char * src, const unsigned srcPitch, char * dst, const unsigned dstPitch)
    {
        //process upper left block
        Op::doBlock(src, srcPitch, dst, dstPitch);
        //finish first row
        UnrollOp<Op, W-1, 1>::doBlock(src+Op::srcWidth, srcPitch, dst+Op::dstWidth, dstPitch);
        //process remaining rows
        UnrollOp<Op, W, H-1>::doBlock(src+srcPitch, srcPitch, dst+dstPitch, dstPitch);
    }
};

template<class Op>
struct UnrollOp<Op, 0, 0>
{
    static inline void doBlock(const char * /*src*/, const unsigned /*srcPitch*/, char * /*dst*/, const unsigned /*dstPitch*/)
    {
    }
};

// Scale down blocks 8x8 to 7x7 with bilinear filtering

class Scale8to7
{
private :
    // Coefficients are static variable, thus are calculated at runtime before main
    // It is not safe to use Scale8to7 before main
    struct Coefficients
    {
        union
        {
            __m128i v;
            short a[8];//only 7 coefs are used
        } a, b;
        Coefficients();
    };
    static const Coefficients c;

    // Scale line from 8 to 7 pixels
    // Every two adjacent pixels are multiplied by coefficients a and b and added together
    static inline __m128i doLine(const char * src)
    {
        //load 8 8b values and unpack to 16b
        const __m128i a = _mm_unpacklo_epi8(
            _mm_loadl_epi64((__m128i*)src),
            _mm_setzero_si128()
        );
        return _mm_add_epi16(//add together
            _mm_mullo_epi16(//multiply left pixel by coef a
                a,
                c.a.v//a coefs as __m128i
            ),
            _mm_mullo_epi16(//multiply right pixel by coef b
                _mm_srli_si128(//get to right pixel by 16b shift
                    a,
                    2
                ),
                c.b.v//b coefs as __m128i
            )
        );
    }
public :
    enum
    {
        srcWidth   = 8,
        srcHeight  = 8,
        dstWidth   = 7,
        dstHeight  = 7,
        rowOverlap = 1
    };

    // Scale block the same style as line
    // Every two lines are multiplied by coefs and added
    static inline void doBlock(const char * src, const unsigned srcPitch, char * dst, const unsigned dstPitch)
    {
        __m128i a = doLine(src);//scale first line
        for(int r = 0; r < dstHeight; ++r)
        {
            const __m128i b = doLine(src + (r+1)*srcPitch);//scale next line
            _mm_storel_epi64(//store 8B, one will be overwritten
                (__m128i*)(dst + r*dstPitch),
                _mm_packs_epi16(//pack to 8b
                    _mm_srai_epi16(//shift right by 8b(4 for line multiply, 4 for this)
                        _mm_add_epi16(//add lines
                            _mm_mullo_epi16(//mul a line by coefs
                                a,
                                _mm_set1_epi16(c.a.a[r])//a coef, depends on row, is same for whole line
                            ),
                            _mm_mullo_epi16(//mul b line by coefs
                                b,
                                _mm_set1_epi16(c.b.a[r])//b coef, similar to a
                            )
                        ),
                        8
                    ),
                    _mm_setzero_si128()
                )
            );
            a = b;//line b will act as a for next row
        }
    }
};

// Scale down two 16px rows by averaging 2x2px blocks

struct ScaleHalf
{
    enum
    {
        srcWidth   = 16,
        srcHeight  = 2,
        dstWidth   = 8,
        dstHeight  = 1,
        rowOverlap = 0
    };

    // dst pitch isn't used because only one row is written
    static inline void doBlock(const char * src, const unsigned srcPitch, char * dst, const unsigned /*dstPitch*/)
    {
        //load two lines and average columns to one 16px line
        const __m128i c = _mm_avg_epu8(
            _mm_load_si128((__m128i*)(src           )),
            _mm_load_si128((__m128i*)(src + srcPitch))
        );
        _mm_storel_epi64(
            (__m128i*)dst,
            _mm_packus_epi16(//pack to 8 8b pixels
                _mm_srli_epi16(//aritmetic shift right by 8b -> top 8b are zeroes so we can saturate
                    _mm_avg_epu8(//average neighbour pixels, right values are in top 8b of 16b pixels
                        c,
                        _mm_slli_si128(//shift left to match neighbour pixels
                            c,
                            1
                        )
                    ),
                    8
                ),
                _mm_setzero_si128()
            )
        );
    }
};

template <typename T>
inline T divAndRoundUp(const T num, const T denom)
{
    return (num+denom-1)/denom;
}

// Convert size to source block count, round up

template <class Op>
inline std::pair<unsigned, unsigned> sizeToSrcBlocksOfOp(const unsigned width, const unsigned height)
{
    return std::make_pair(
        divAndRoundUp(width , static_cast<unsigned>(Op::srcWidth)),
        divAndRoundUp(height, static_cast<unsigned>(Op::srcHeight))
    );
}

template <class Op>
inline std::pair<unsigned, unsigned> dstSizeOfOp(const unsigned width, const unsigned height)
{
    std::pair<unsigned, unsigned> blocks = sizeToSrcBlocksOfOp<Op>(width, height);
    return std::make_pair(
        blocks.first *Op::dstWidth + Op::rowOverlap,
        blocks.second*Op::dstHeight
    );
}

template <typename T>
inline T alignValue(const T val, const T alignment)
{
    return (val+alignment-1)/alignment*alignment;
}

inline std::pair<unsigned, unsigned> alignSize(
    const unsigned width         , const unsigned height,
    const unsigned widthAlignment, const unsigned heightAlignment)
{
    return std::make_pair(
        alignValue(width , widthAlignment),
        alignValue(height, heightAlignment)
    );
}

template <class Op>
inline std::pair<unsigned, unsigned> alignSizeForOp(const unsigned width, const unsigned height)
{
    return alignSize(width, height, Op::srcWidth, Op::srcHeight);
}

// Block operation

template <class Op>
inline void doOp(
    const unsigned srcWidth, const unsigned srcHeight, const unsigned srcPitch, const char * srcData,
    const unsigned dstWidth, const unsigned dstHeight, const unsigned dstPitch, char * dstData)
{
    //get block count
    std::pair<unsigned, unsigned> blocks = sizeToSrcBlocksOfOp<Op>(srcWidth, srcHeight);
    //check sizes
    assert(blocks.first *Op::srcWidth                  <= srcPitch);
    assert(blocks.first *Op::dstWidth + Op::rowOverlap <= dstPitch);
    assert(blocks.second*Op::srcHeight                 <= srcHeight);
    assert(blocks.second*Op::dstHeight                 <= dstHeight);
    //for every row of blocks
    for(unsigned ri = 0; ri < blocks.second; ++ri)
    {
        //calculate pointers to first row of block
        const char * srcRow = srcData + ri*Op::srcHeight*srcPitch;
        char * dstRow = dstData + ri*Op::dstHeight*dstPitch;
        //for every column of blocks
        for(unsigned ci = 0; ci < blocks.first; ++ci)
        {
            //process block
            Op::doBlock(srcRow + ci*Op::srcWidth, srcPitch, dstRow + ci*Op::dstWidth, dstPitch);
        }
    }
}

#endif
